@@ -41,6 +41,9 @@ void _IO_ldtostr(long double *, char *, int, int, char);
  /* The exponent of 1.0 */
  #define EXONE (0x3fff)
 
+ /* Maximum exponent digits - base 10 */
+ #define MAX_EXP_DIGITS 5
+
 /* Control structure for long doublue conversion including rounding precision values.
  * rndprc can be set to 80 (if NE=6), 64, 56, 53, or 24 bits.
  */
@@ -1859,7 +1862,7 @@ if((yy[NE-1] & 0x7fff) == 0 && (yy[NE-2] & 0x8000) == 0)
 #ifdef INFINITY
 /* Point to the exponent field.  */
 p = &yy[NE-1];
-if( *p == 0x7fff )
+if( (*p & 0x7fff) == 0x7fff )
 	{
 #ifdef NANS
 #ifdef IBMPC
@@ -2703,11 +2706,14 @@ _ldtoa_r (struct _reent *ptr, long double d, int mode, int ndigits, int *decpt,
 {
 unsigned short e[NI];
 char *s, *p;
-int k;
+int i, j, k;
+int orig_ndigits;
 LDPARMS rnd;
 LDPARMS *ldp = &rnd;
 char *outstr;
+char outbuf[NDEC + MAX_EXP_DIGITS + 10];
 
+orig_ndigits = ndigits;
 rnd.rlast = -1;
 rnd.rndprc = NBITS;
 
@@ -2744,18 +2750,14 @@ if( mode != 3 )
    For now, just ask for 20 digits which is enough but sometimes too many.  */
 if( mode == 0 )
         ndigits = 20;
+
 /* This sanity limit must agree with the corresponding one in etoasc, to
    keep straight the returned value of outexpon.  */
 if( ndigits > NDEC )
         ndigits = NDEC;
 
-/* reentrancy addition to use mprec storage pool */
-_REENT_MP_RESULT(ptr) = Balloc (ptr, 3);
-_REENT_MP_RESULT_K(ptr) = 3;
-outstr = (char *)_REENT_MP_RESULT(ptr);
-
-etoasc( e, outstr, ndigits, mode, ldp );
-s =  outstr;
+etoasc( e, outbuf, ndigits, mode, ldp );
+s =  outbuf;
 if( eisinf(e) || eisnan(e) )
         {
         *decpt = 9999;
@@ -2766,7 +2768,7 @@ if( eisinf(e) || eisnan(e) )
 /* Transform the string returned by etoasc into what the caller wants.  */
 
 /* Look for decimal point and delete it from the string. */
-s = outstr;
+s = outbuf;
 while( *s != '\0' )
         {
         if( *s == '.' )
@@ -2787,19 +2789,19 @@ while( *s != '\0' )
 nodecpt:
 
 /* Back up over the exponent field. */
-while( *s != 'E' && s > outstr)
+while( *s != 'E' && s > outbuf)
         --s;
 *s = '\0';
 
 stripspaces:
 
 /* Strip leading spaces and sign. */
-p = outstr;
+p = outbuf;
 while( *p == ' ' || *p == '-')
         ++p;
 
 /* Find new end of string.  */
-s = outstr;
+s = outbuf;
 while( (*s++ = *p++) != '\0' )
         ;
 --s;
@@ -2812,20 +2814,38 @@ else if( ndigits > ldp->outexpon )
 else
         k = ldp->outexpon;
 
-while( *(s-1) == '0' && ((s - outstr) > k))
+while( *(s-1) == '0' && ((s - outbuf) > k))
         *(--s) = '\0';
 
 /* In f format, flush small off-scale values to zero.
    Rounding has been taken care of by etoasc. */
 if( mode == 3 && ((ndigits + ldp->outexpon) < 0))
         {
-        s = outstr;
+        s = outbuf;
         *s = '\0';
         *decpt = 0;
         }
 
+/* reentrancy addition to use mprec storage pool */
+/* we want to have enough space to hold the formatted result */
+
+if (mode == 3) /* f format, account for sign + dec digits + decpt + frac */
+  i = *decpt + orig_ndigits + 3;
+else /* account for sign + max precision digs + E + exp sign + exponent */
+  i = orig_ndigits + MAX_EXP_DIGITS + 4;
+
+j = sizeof (__ULong);
+for (_REENT_MP_RESULT_K(ptr) = 0; sizeof (_Bigint) - sizeof (__ULong) + j <= i; j <<= 1)
+  _REENT_MP_RESULT_K(ptr)++;
+_REENT_MP_RESULT(ptr) = Balloc (ptr, _REENT_MP_RESULT_K(ptr));
+
+/* Copy from internal temporary buffer to permanent buffer.  */
+outstr = (char *)_REENT_MP_RESULT(ptr);
+strcpy (outstr, outbuf);
+
 if( rve )
-        *rve = s;
+        *rve = outstr + (s - outbuf);
+
 return outstr;
 }
 
