@@ -1,6 +1,6 @@
 /* dcrt0.cc -- essentially the main() for the Cygwin dll
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -18,10 +18,8 @@ details. */
 #include <wingdi.h>
 #include <winuser.h>
 #include <errno.h>
-#include "sync.h"
 #include "sigproc.h"
 #include "pinfo.h"
-#include "heap.h"
 #include "cygerrno.h"
 #define NEED_VFORK
 #include "perprocess.h"
@@ -35,13 +33,15 @@ details. */
 #include "shared_info.h"
 #include "cygwin_version.h"
 #include "dll_init.h"
+#include "cygthread.h"
+#include "sync.h"
 
 #define MAX_AT_FILE_LEVEL 10
 
 #define PREMAIN_LEN (sizeof (user_data->premain) / sizeof (user_data->premain[0]))
 
-HANDLE NO_COPY hMainProc = NULL;
-HANDLE NO_COPY hMainThread = NULL;
+HANDLE NO_COPY hMainProc;
+HANDLE NO_COPY hMainThread;
 
 sigthread NO_COPY mainthread;		// ID of the main thread
 
@@ -75,13 +75,10 @@ MTinterface _mtinterf;
 bool NO_COPY _cygwin_testing;
 unsigned NO_COPY _cygwin_testing_magic;
 
+char NO_COPY almost_null[1];
+
 extern "C"
 {
-  void *export_malloc (unsigned int);
-  void export_free (void *);
-  void *export_realloc (void *, unsigned int);
-  void *export_calloc (unsigned int, unsigned int);
-
   /* This is an exported copy of environ which can be used by DLLs
      which use cygwin.dll.  */
   char **__cygwin_environ;
@@ -94,12 +91,12 @@ extern "C"
    /* dll_major */ CYGWIN_VERSION_DLL_MAJOR,
    /* dll_major */ CYGWIN_VERSION_DLL_MINOR,
    /* impure_ptr_ptr */ NULL, /* envptr */ NULL,
-   /* malloc */ export_malloc, /* free */ export_free,
-   /* realloc */ export_realloc,
+   /* malloc */ malloc, /* free */ free,
+   /* realloc */ realloc,
    /* fmode_ptr */ NULL, /* main */ NULL, /* ctors */ NULL,
    /* dtors */ NULL, /* data_start */ NULL, /* data_end */ NULL,
    /* bss_start */ NULL, /* bss_end */ NULL,
-   /* calloc */ export_calloc,
+   /* calloc */ calloc,
    /* premain */ {NULL, NULL, NULL, NULL},
    /* run_ctors_p */ 0,
    /* unused */ {0, 0, 0, 0, 0, 0, 0},
@@ -157,10 +154,10 @@ do_global_ctors (void (**in_pfunc)(), int force)
 }
 
 /*
- * Replaces -@file in the command line with the contents of the file.
- * There may be multiple -@file's in a single command line
- * A \-@file is replaced with -@file so that echo \-@foo would print
- * -@foo and not the contents of foo.
+ * Replaces @file in the command line with the contents of the file.
+ * There may be multiple @file's in a single command line
+ * A \@file is replaced with @file so that echo \@foo would print
+ * @foo and not the contents of foo.
  */
 static int __stdcall
 insert_file (char *name, char *&cmd)
@@ -337,7 +334,7 @@ globify (char *word, char **&argv, int &argc, int &argvlen)
   char **av = argv + n;
   while (*gv)
     {
-      debug_printf ("argv[%d] = '%s'\n", n++, *gv);
+      debug_printf ("argv[%d] = '%s'", n++, *gv);
       *av++ = *gv++;
     }
 
@@ -407,7 +404,7 @@ build_argv (char *cmd, char **&argv, int &argc, int winshell)
       /* Add word to argv file after (optional) wildcard expansion. */
       if (!winshell || !argc || !globify (word, argv, argc, argvlen))
 	{
-	  debug_printf ("argv[%d] = '%s'\n", argc, word);
+	  debug_printf ("argv[%d] = '%s'", argc, word);
 	  argv[argc++] = word;
 	}
     }
@@ -451,13 +448,11 @@ check_sanity_and_sync (per_process *p)
     signal_shift_subtract = 0;
 }
 
-static NO_COPY STARTUPINFO si;
-# define fork_info ((struct child_info_fork *)(si.lpReserved2))
-# define spawn_info ((struct child_info_spawn *)(si.lpReserved2))
-child_info_fork NO_COPY *child_proc_info = NULL;
-static MEMORY_BASIC_INFORMATION sm;
+child_info NO_COPY *child_proc_info = NULL;
+static MEMORY_BASIC_INFORMATION NO_COPY sm;
 
-#define CYGWIN_GUARD ((wincap.has_page_guard ()) ? PAGE_GUARD : PAGE_NOACCESS)
+#define CYGWIN_GUARD ((wincap.has_page_guard ()) ? \
+		     PAGE_EXECUTE_READWRITE|PAGE_GUARD : PAGE_NOACCESS)
 
 // __inline__ void
 extern void
@@ -498,7 +493,7 @@ alloc_stack_hard_way (child_info_fork *ci, volatile char *b)
     {
       m.BaseAddress = (LPVOID)((DWORD)m.BaseAddress - 1);
       if (!VirtualAlloc ((LPVOID) m.BaseAddress, 1, MEM_COMMIT,
-			 PAGE_EXECUTE_READWRITE|CYGWIN_GUARD))
+			 CYGWIN_GUARD))
 	api_fatal ("fork: couldn't allocate new stack guard page %p, %E",
 		   m.BaseAddress);
     }
@@ -555,12 +550,12 @@ dll_crt0_1 ()
   /* Initialize SIGSEGV handling, etc. */
   init_exceptions (&cygwin_except_entry);
 
-  do_global_ctors (&__CTOR_LIST__, 1);
-
   /* Set the os_being_run global. */
   wincap.init ();
   check_sanity_and_sync (user_data);
 
+  do_global_ctors (&__CTOR_LIST__, 1);
+
   /* Nasty static stuff needed by newlib -- point to a local copy of
      the reent stuff.
      Note: this MUST be done here (before the forkee code) as the
@@ -572,43 +567,46 @@ dll_crt0_1 ()
   user_data->resourcelocks->Init ();
   user_data->threadinterface->Init (user_data->forkee);
 
-  threadname_init ();
-  debug_init ();
-  (void) getpagesize ();	/* initialize page size constant */
-
-  regthread ("main", GetCurrentThreadId ());
   mainthread.init ("mainthread"); // For use in determining if signals
 				  //  should be blocked.
 
+  winpids::init ();
+
   int envc = 0;
   char **envp = NULL;
 
-  if (child_proc_info)
+  if (!child_proc_info)
+    memory_init ();
+  else
     {
+      bool close_ppid_handle = false;
+      bool close_hexec_proc = false;
       switch (child_proc_info->type)
 	{
 	  case _PROC_FORK:
-	    cygheap_fixup_in_child (child_proc_info, 0);
 	    alloc_stack (fork_info);
+	    cygheap_fixup_in_child (0);
+	    memory_init ();
 	    set_myself (mypid);
-	    ProtectHandle (child_proc_info->forker_finished);
+	    close_ppid_handle = !!child_proc_info->pppid_handle;
 	    break;
 	  case _PROC_SPAWN:
-	    if (spawn_info->hexec_proc)
-	      CloseHandle (spawn_info->hexec_proc);
+	    /* Have to delay closes until after cygheap is setup */
+	    close_hexec_proc = !!spawn_info->hexec_proc;
+	    close_ppid_handle = !!child_proc_info->pppid_handle;
 	    goto around;
 	  case _PROC_EXEC:
 	    hexec_proc = spawn_info->hexec_proc;
 	  around:
 	    HANDLE h;
-	    cygheap_fixup_in_child (spawn_info, 1);
+	    cygheap_fixup_in_child (1);
+	    memory_init ();
 	    if (!spawn_info->moreinfo->myself_pinfo ||
 		!DuplicateHandle (hMainProc, spawn_info->moreinfo->myself_pinfo,
 				  hMainProc, &h, 0, 0,
 				  DUPLICATE_SAME_ACCESS | DUPLICATE_CLOSE_SOURCE))
 	      h = NULL;
 	    set_myself (mypid, h);
-	    myself->uid = spawn_info->moreinfo->uid;
 	    __argc = spawn_info->moreinfo->argc;
 	    __argv = spawn_info->moreinfo->argv;
 	    envp = spawn_info->moreinfo->envp;
@@ -621,19 +619,24 @@ dll_crt0_1 ()
 		old_title = strcpy (title_buf, spawn_info->moreinfo->old_title);
 		cfree (spawn_info->moreinfo->old_title);
 	      }
-	    if (child_proc_info->subproc_ready)
-	      ProtectHandle (child_proc_info->subproc_ready);
-	    if (myself->uid == ILLEGAL_UID)
-	      cygheap->user.set_sid (NULL);
 	    break;
 	}
+      if (close_hexec_proc)
+	CloseHandle (spawn_info->hexec_proc);
+      if (close_ppid_handle)
+	CloseHandle (child_proc_info->pppid_handle);
     }
+
   ProtectHandle (hMainProc);
   ProtectHandle (hMainThread);
+  cygthread::init ();
+
+  pthread::initMainThread (!user_data->forkee);
+
+  /* Initialize debug muto, if DLL is built with --enable-debugging.
+     Need to do this before any helper threads start. */
+  debug_init ();
 
-  /* Initialize the cygwin subsystem if this is the first process,
-     or attach to shared data structures if it's already running. */
-  memory_init ();
   cygheap->fdtab.vfork_child_fixup ();
 
   (void) SetErrorMode (SEM_FAILCRITICALERRORS);
@@ -679,7 +682,10 @@ dll_crt0_1 ()
   /* Allocate cygheap->fdtab */
   dtable_init ();
 
-/* Initialize uid, gid. */
+  /* Init global well known SID objects */
+  cygsid::init ();
+
+  /* Initialize user info. */
   uinfo_init ();
 
   /* Initialize signal/subprocess handling. */
@@ -707,9 +713,6 @@ dll_crt0_1 ()
 	{
 	  char *new_argv0 = (char *) alloca (MAX_PATH);
 	  cygwin_conv_to_posix_path (__argv[0], new_argv0);
-	  char *p = strchr (new_argv0, '\0') - 4;
-	  if (p > new_argv0 && strcasematch (p, ".exe"))
-	    *p = '\0';
 	  __argv[0] = new_argv0;
 	}
     }
@@ -722,7 +725,16 @@ dll_crt0_1 ()
   cygheap->fdtab.stdio_init ();
 
   /* Set up __progname for getopt error call. */
-  __progname = __argv[0];
+  if (__argv[0] && (__progname = strrchr (__argv[0], '/')))
+    ++__progname;
+  else
+    __progname = __argv[0];
+  if (__progname)
+    {
+      char *cp = strchr (__progname, '\0') - 4;
+      if (cp > __progname && strcasematch (cp, ".exe"))
+	*cp = '\0';
+    }
 
   /* Set new console title if appropriate. */
 
@@ -759,7 +771,7 @@ dll_crt0_1 ()
   /* Flush signals and ensure that signal thread is up and running. Can't
      do this for noncygwin case since the signal thread is blocked due to
      LoadLibrary serialization. */
-  sig_send (NULL, __SIGFLUSH);
+  wait_for_sigthread ();
 
   set_errno (0);
 
@@ -769,22 +781,51 @@ dll_crt0_1 ()
     exit (user_data->main (__argc, __argv, *user_data->envptr));
 }
 
+#ifdef DEBUGGING
+void
+break_here ()
+{
+  debug_printf ("break here");
+}
+#endif
+
 void
 initial_env ()
 {
+  DWORD len;
   char buf[MAX_PATH + 1];
 #ifdef DEBUGGING
   if (GetEnvironmentVariable ("CYGWIN_SLEEP", buf, sizeof (buf) - 1))
     {
-      console_printf ("Sleeping %d, pid %u\n", atoi (buf), GetCurrentProcessId ());
-      Sleep (atoi (buf));
+      DWORD ms = atoi (buf);
+      buf[0] = '\0';
+      len = GetModuleFileName (NULL, buf, MAX_PATH);
+      console_printf ("Sleeping %d, pid %u %s\n", ms, GetCurrentProcessId (), buf);
+      Sleep (ms);
+    }
+  if (GetEnvironmentVariable ("CYGWIN_DEBUG", buf, sizeof (buf) - 1))
+    {
+      char buf1[MAX_PATH + 1];
+      len = GetModuleFileName (NULL, buf1, MAX_PATH);
+      strlwr (buf1);
+      strlwr (buf);
+      char *p = strchr (buf, ':');
+      if (!p)
+	p = (char *) "gdb.exe -nw";
+      else
+	*p++ = '\0';
+      if (strstr (buf1, buf))
+	{
+	  error_start_init (p);
+	  try_to_debug ();
+	  break_here ();
+	}
     }
 #endif
 
   if (GetEnvironmentVariable ("CYGWIN_TESTING", buf, sizeof (buf) - 1))
     {
       _cygwin_testing = 1;
-      DWORD len;
       if ((len = GetModuleFileName (cygwin_hmodule, buf, MAX_PATH))
 	  && len > sizeof ("new-cygwin1.dll")
 	  && strcasematch (buf + len - sizeof ("new-cygwin1.dll"),
@@ -805,6 +846,7 @@ _dll_crt0 ()
   DECLARE_TLS_STORAGE;
   initial_env ();
   char zeros[sizeof (fork_info->zero)] = {0};
+  static NO_COPY STARTUPINFO si;
 #ifdef DEBUGGING
   strace.microseconds ();
 #endif
@@ -822,42 +864,47 @@ _dll_crt0 ()
 		   &hMainThread, 0, false, DUPLICATE_SAME_ACCESS);
 
   GetStartupInfo (&si);
-  if (si.cbReserved2 >= EXEC_MAGIC_SIZE &&
-      memcmp (fork_info->zero, zeros, sizeof (zeros)) == 0)
+  child_proc_info = (child_info *) si.lpReserved2;
+  if (si.cbReserved2 < EXEC_MAGIC_SIZE || !child_proc_info
+      || memcmp (child_proc_info->zero, zeros, sizeof (zeros)) != 0)
+    child_proc_info = NULL;
+  else
     {
-      if ((fork_info->intro & OPROC_MAGIC_MASK) == OPROC_MAGIC_GENERIC)
-	multiple_cygwin_problem ("proc", fork_info->intro, 0);
-      else if (fork_info->intro == PROC_MAGIC_GENERIC
-	       && fork_info->magic != CHILD_INFO_MAGIC)
-	multiple_cygwin_problem ("proc", fork_info->magic, CHILD_INFO_MAGIC);
+      if ((child_proc_info->intro & OPROC_MAGIC_MASK) == OPROC_MAGIC_GENERIC)
+	multiple_cygwin_problem ("proc", child_proc_info->intro, 0);
+      else if (child_proc_info->intro == PROC_MAGIC_GENERIC
+	       && child_proc_info->magic != CHILD_INFO_MAGIC)
+	multiple_cygwin_problem ("proc", child_proc_info->magic,
+				 CHILD_INFO_MAGIC);
+      else if (child_proc_info->cygheap != (void *) &_cygheap_start)
+	multiple_cygwin_problem ("cygheap", (DWORD) child_proc_info->cygheap,
+				 (DWORD) &_cygheap_start);
       unsigned should_be_cb = 0;
-      switch (fork_info->type)
+      switch (child_proc_info->type)
 	{
 	  case _PROC_FORK:
-	    user_data->forkee = fork_info->cygpid;
+	    user_data->forkee = child_proc_info->cygpid;
 	    should_be_cb = sizeof (child_info_fork);
 	    /* fall through */;
 	  case _PROC_SPAWN:
-	    if (fork_info->pppid_handle)
-	      CloseHandle (fork_info->pppid_handle);
-	    /* fall through */;
 	  case _PROC_EXEC:
 	    if (!should_be_cb)
 	      should_be_cb = sizeof (child_info);
-	    if (should_be_cb != fork_info->cb)
-	      multiple_cygwin_problem ("proc size", fork_info->cb, should_be_cb);
-	    else if (sizeof (fhandler_union) != fork_info->fhandler_union_cb)
-	      multiple_cygwin_problem ("fhandler size", fork_info->fhandler_union_cb, sizeof (fhandler_union));
+	    if (should_be_cb != child_proc_info->cb)
+	      multiple_cygwin_problem ("proc size", child_proc_info->cb, should_be_cb);
+	    else if (sizeof (fhandler_union) != child_proc_info->fhandler_union_cb)
+	      multiple_cygwin_problem ("fhandler size", child_proc_info->fhandler_union_cb, sizeof (fhandler_union));
 	    else
 	      {
-		child_proc_info = fork_info;
 		cygwin_mount_h = child_proc_info->mount_h;
 		mypid = child_proc_info->cygpid;
 		break;
 	      }
 	  default:
-	    system_printf ("unknown exec type %d", fork_info->type);
-	    fork_info = NULL;
+	    system_printf ("unknown exec type %d", child_proc_info->type);
+	    /* intentionally fall through */
+	  case _PROC_WHOOPS:
+	    child_proc_info = NULL;
 	    break;
 	}
     }
@@ -905,26 +952,39 @@ __main (void)
   do_global_ctors (user_data->ctors, FALSE);
 }
 
-enum
+enum exit_states
   {
-    ES_SIGNAL = 1,
-    ES_CLOSEALL = 2,
-    ES_SIGPROCTERMINATE = 3
+    ES_NOT_EXITING = 0,
+    ES_THREADTERM,
+    ES_SIGNAL,
+    ES_CLOSEALL,
+    ES_SIGPROCTERMINATE,
+    ES_TITLE,
+    ES_HUP_PGRP,
+    ES_HUP_SID,
+    ES_TTY_TERMINATE,
+    ES_EVENTS_TERMINATE
   };
 
+exit_states NO_COPY exit_state;
+extern CRITICAL_SECTION exit_lock;
+
 extern "C" void __stdcall
 do_exit (int status)
 {
+  EnterCriticalSection (&exit_lock);
   UINT n = (UINT) status;
-  static int NO_COPY exit_state = 0;
 
   syscall_printf ("do_exit (%d)", n);
 
   vfork_save *vf = vfork_storage.val ();
   if (vf != NULL && vf->pid < 0)
+    vf->restore_exit (status);
+
+  if (exit_state < ES_THREADTERM)
     {
-      vf->pid = status < 0 ? status : -status;
-      longjmp (vf->j, 1);
+      exit_state = ES_THREADTERM;
+      cygthread::terminate ();
     }
 
   if (exit_state < ES_SIGNAL)
@@ -951,16 +1011,18 @@ do_exit (int status)
       sigproc_terminate ();
     }
 
-  if (n & EXIT_REPARENTING)
-    n &= ~EXIT_REPARENTING;
-  else
+  myself->stopsig = 0;
+  if (exit_state < ES_TITLE)
     {
-      myself->stopsig = 0;
-
+      exit_state = ES_TITLE;
       /* restore console title */
       if (old_title && display_title)
 	set_console_title (old_title);
+    }
 
+  if (exit_state < ES_HUP_PGRP)
+    {
+      exit_state = ES_HUP_PGRP;
       /* Kill orphaned children on group leader exit */
       if (myself->has_pgid_children && myself->pid == myself->pgid)
 	{
@@ -968,7 +1030,11 @@ do_exit (int status)
 			  myself->pid, myself->pgid);
 	  kill_pgrp (myself->pgid, -SIGHUP);
 	}
+    }
 
+  if (exit_state < ES_HUP_SID)
+    {
+      exit_state = ES_HUP_SID;
       /* Kill the foreground process group on session leader exit */
       if (getpgrp () > 0 && myself->pid == myself->sid && real_tty_attached (myself))
 	{
@@ -981,17 +1047,46 @@ do_exit (int status)
 	    tp->kill_pgrp (SIGHUP);
 	}
 
+    }
+
+  if (exit_state < ES_TTY_TERMINATE)
+    {
+      exit_state = ES_TTY_TERMINATE;
       tty_terminate ();
     }
 
-  window_terminate ();
-  events_terminate ();
-  shared_terminate ();
+  if (exit_state < ES_EVENTS_TERMINATE)
+    {
+      exit_state = ES_EVENTS_TERMINATE;
+      events_terminate ();
+    }
 
   minimal_printf ("winpid %d, exit %d", GetCurrentProcessId (), n);
   myself->exit (n);
 }
 
+static muto *atexit_lock;
+
+extern "C" int
+cygwin_atexit (void (*function)(void))
+{
+  int res;
+  if (!atexit_lock)
+    new_muto (atexit_lock);
+  atexit_lock->acquire ();
+  res = atexit (function);
+  atexit_lock->release ();
+  return res;
+}
+
+extern "C" void
+cygwin_exit (int n)
+{
+  if (atexit_lock)
+    atexit_lock->acquire ();
+  exit (n);
+}
+
 extern "C" void
 _exit (int n)
 {
@@ -1016,10 +1111,10 @@ __api_fatal (const char *fmt, ...)
      a serious error. */
   if (GetFileType (GetStdHandle (STD_ERROR_HANDLE)) != FILE_TYPE_CHAR)
     {
-      HANDLE h = CreateFileA ("CONOUT$", GENERIC_READ|GENERIC_WRITE,
-			      FILE_SHARE_WRITE | FILE_SHARE_WRITE, &sec_none,
-			      OPEN_EXISTING, 0, 0);
-      if (h)
+      HANDLE h = CreateFile ("CONOUT$", GENERIC_READ | GENERIC_WRITE,
+			     FILE_SHARE_WRITE | FILE_SHARE_WRITE,
+			     &sec_none, OPEN_EXISTING, 0, 0);
+      if (h != INVALID_HANDLE_VALUE)
 	(void) WriteFile (h, buf, len, &done, 0);
     }
 
@@ -1035,9 +1130,9 @@ __api_fatal (const char *fmt, ...)
 void
 multiple_cygwin_problem (const char *what, unsigned magic_version, unsigned version)
 {
-  if (_cygwin_testing && strstr (what, "proc"))
+  if (_cygwin_testing && (strstr (what, "proc") || strstr (what, "cygheap")))
     {
-      fork_info = NULL;
+      child_proc_info->type = _PROC_WHOOPS;
       return;
     }
 
