@@ -1,6 +1,6 @@
 /* dtable.cc: file descriptor support.
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -16,14 +16,13 @@ details. */
 #include <stdlib.h>
 #include <stdio.h>
 #include <unistd.h>
-#include <fcntl.h>
 #include <sys/cygwin.h>
 #include <assert.h>
+#include <ntdef.h>
+#include <winnls.h>
 
 #define USE_SYS_TYPES_FD_SET
 #include <winsock.h>
-#include "sync.h"
-#include "sigproc.h"
 #include "pinfo.h"
 #include "cygerrno.h"
 #include "perprocess.h"
@@ -32,10 +31,15 @@ details. */
 #include "path.h"
 #include "dtable.h"
 #include "cygheap.h"
+#include "ntdll.h"
 
 static const NO_COPY DWORD std_consts[] = {STD_INPUT_HANDLE, STD_OUTPUT_HANDLE,
 					   STD_ERROR_HANDLE};
 
+static const char *handle_to_fn (HANDLE, char *);
+
+static const char NO_COPY unknown_file[] = "some disk file";
+
 /* Set aside space for the table of fds */
 void
 dtable_init (void)
@@ -53,14 +57,6 @@ set_std_handle (int fd)
     SetStdHandle (std_consts[fd], cygheap->fdtab[fd]->get_output_handle ());
 }
 
-void
-dtable::dec_console_fds ()
-{
-  if (console_fds > 0 && !--console_fds &&
-      myself->ctty != TTY_CONSOLE && !check_pty_fds())
-    FreeConsole ();
-}
-
 int
 dtable::extend (int howmuch)
 {
@@ -93,7 +89,7 @@ dtable::extend (int howmuch)
 void
 dtable::get_debugger_info ()
 {
-  if (IsDebuggerPresent ())
+  if (being_debugged ())
     {
       char std[3][sizeof ("/dev/ttyNNNN")];
       std[0][0] = std[1][0] = std [2][0] = '\0';
@@ -135,7 +131,7 @@ dtable::stdio_init ()
   HANDLE out = GetStdHandle (STD_OUTPUT_HANDLE);
   HANDLE err = GetStdHandle (STD_ERROR_HANDLE);
 
-  init_std_file_from_handle (0, in, GENERIC_READ);
+  init_std_file_from_handle (0, in);
 
   /* STD_ERROR_HANDLE has been observed to be the same as
      STD_OUTPUT_HANDLE.  We need separate handles (e.g. using pipes
@@ -153,8 +149,8 @@ dtable::stdio_init ()
 	}
     }
 
-  init_std_file_from_handle (1, out, GENERIC_WRITE);
-  init_std_file_from_handle (2, err, GENERIC_WRITE);
+  init_std_file_from_handle (1, out);
+  init_std_file_from_handle (2, err);
   /* Assign the console as the controlling tty for this process if we actually
      have a console and no other controlling tty has been assigned. */
   if (myself->ctty < 0 && GetConsoleCP () > 0)
@@ -165,10 +161,9 @@ int
 dtable::find_unused_handle (int start)
 {
   AssertResourceOwner (LOCK_FD_LIST, READ_LOCK);
-
   do
     {
-      for (int i = start; i < (int) size; i++)
+      for (size_t i = start; i < size; i++)
 	/* See if open -- no need for overhead of not_open */
 	if (fds[i] == NULL)
 	  return i;
@@ -187,17 +182,13 @@ dtable::release (int fd)
 	case FH_SOCKET:
 	  dec_need_fixup_before ();
 	  break;
-	case FH_CONSOLE:
-	  dec_console_fds ();
-	  break;
 	}
       delete fds[fd];
       fds[fd] = NULL;
     }
 }
 
-extern "C"
-int
+extern "C" int
 cygwin_attach_handle_to_fd (char *name, int fd, HANDLE handle, mode_t bin,
 			    DWORD myaccess)
 {
@@ -206,93 +197,119 @@ cygwin_attach_handle_to_fd (char *name, int fd, HANDLE handle, mode_t bin,
   path_conv pc;
   fhandler_base *res = cygheap->fdtab.build_fhandler_from_name (fd, name, handle,
 								pc);
-  res->init (handle, myaccess, bin);
+  res->init (handle, myaccess, bin ?: pc.binmode ());
   return fd;
 }
 
 void
-dtable::init_std_file_from_handle (int fd, HANDLE handle, DWORD myaccess)
+dtable::init_std_file_from_handle (int fd, HANDLE handle)
 {
-  int bin;
   const char *name;
   CONSOLE_SCREEN_BUFFER_INFO buf;
   struct sockaddr sa;
   int sal = sizeof (sa);
   DCB dcb;
+  unsigned bin = O_BINARY;
 
   first_fd_for_open = 0;
 
   if (!not_open (fd))
     return;
 
-  if (!handle || handle == INVALID_HANDLE_VALUE)
-    {
-      fds[fd] = NULL;
-      return;
-    }
-
-  if (__fmode)
-    bin = __fmode;
+  SetLastError (0);
+  DWORD ft = GetFileType (handle);
+  if (ft == FILE_TYPE_UNKNOWN && GetLastError () == ERROR_INVALID_HANDLE)
+    name = NULL;
   else
-    bin = binmode ?: 0;
+    {
+      /* See if we can consoleify it */
+      if (GetConsoleScreenBufferInfo (handle, &buf))
+	{
+	  if (ISSTATE (myself, PID_USETTY))
+	    name = "/dev/tty";
+	  else
+	    name = "/dev/conout";
+	}
+      else if (GetNumberOfConsoleInputEvents (handle, (DWORD *) &buf))
+	{
+	  if (ISSTATE (myself, PID_USETTY))
+	    name = "/dev/tty";
+	  else
+	    name = "/dev/conin";
+	}
+      else if (ft == FILE_TYPE_PIPE)
+	{
+	  if (fd == 0)
+	    name = "/dev/piper";
+	  else
+	    name = "/dev/pipew";
+	}
+      else if (wsock_started && getpeername ((SOCKET) handle, &sa, &sal) == 0)
+	name = "/dev/socket";
+      else if (GetCommState (handle, &dcb))
+	name = "/dev/ttyS0"; // FIXME - determine correct device
+      else
+	{
+	  name = handle_to_fn (handle, (char *) alloca (MAX_PATH + 100));
+	  bin = 0;
+	}
+    }
 
-  /* See if we can consoleify it  - if it is a console,
-   don't open it in binary.  That will screw up our crlfs*/
-  if (GetConsoleScreenBufferInfo (handle, &buf))
-    {
-      if (ISSTATE (myself, PID_USETTY))
-	name = "/dev/tty";
-      else
-	name = "/dev/conout";
-      bin = 0;
-    }
-  else if (GetNumberOfConsoleInputEvents (handle, (DWORD *) &buf))
-    {
-      if (ISSTATE (myself, PID_USETTY))
-	name = "/dev/tty";
-      else
-	name = "/dev/conin";
-      bin = 0;
-    }
-  else if (GetFileType (handle) == FILE_TYPE_PIPE)
-    {
-      if (fd == 0)
-	name = "/dev/piper";
-      else
-	name = "/dev/pipew";
-      if (bin == 0)
-	bin = O_BINARY;
-    }
-  else if (wsock_started && getpeername ((SOCKET) handle, &sa, &sal) == 0)
-    name = "/dev/socket";
-  else if (GetCommState (handle, &dcb))
-    name = "/dev/ttyS0"; // FIXME - determine correct device
+  if (!name)
+    fds[fd] = NULL;
   else
-    name = "unknown disk file";
+    {
+      path_conv pc;
+      fhandler_base *fh = build_fhandler_from_name (fd, name, handle, pc);
+      if (!bin)
+	{
+	  bin = fh->get_default_fmode (O_RDWR);
+	  if (!bin && name != unknown_file)
+	    bin = pc.binmode ();
+	}
 
-  path_conv pc;
-  build_fhandler_from_name (fd, name, handle, pc)->init (handle, myaccess, bin);
-  set_std_handle (fd);
-  paranoid_printf ("fd %d, handle %p", fd, handle);
+      fh->init (handle, GENERIC_READ | GENERIC_WRITE, bin);
+      set_std_handle (fd);
+      paranoid_printf ("fd %d, handle %p", fd, handle);
+    }
 }
 
 fhandler_base *
 dtable::build_fhandler_from_name (int fd, const char *name, HANDLE handle,
 				  path_conv& pc, unsigned opt, suffix_info *si)
 {
-  pc.check (name, opt | PC_NULLEMPTY | PC_FULL, si);
+  pc.check (name, opt | PC_NULLEMPTY | PC_FULL | PC_POSIX, si);
   if (pc.error)
     {
       set_errno (pc.error);
       return NULL;
     }
 
-  return build_fhandler (fd, pc.get_devn (), name, pc, pc.get_unitn ());
+  if (!pc.exists () && handle)
+    pc.fillin (handle);
+
+  char *posix_path = pc.return_and_clear_normalized_path ();
+  fhandler_base *fh = build_fhandler (fd, pc.get_devn (), posix_path,
+				      pc, pc.get_unitn ());
+  if (pc.issocket ()) /* Only true for files pretending an AF_LOCAL socket. */
+    {
+      fhandler_socket * fhs = (fhandler_socket *) fh;
+      fhs->set_addr_family (AF_LOCAL);
+      fhs->set_sun_path (posix_path);
+    }
+  return fh;
+}
+
+fhandler_base *
+dtable::build_fhandler (int fd, DWORD dev, const char *unix_name,
+			const char *win32_name, int unit)
+{
+  return build_fhandler (fd, dev, cstrdup (unix_name), win32_name, unit);
 }
 
 #define cnew(name) new ((void *) ccalloc (HEAP_FHANDLER, 1, sizeof (name))) name
 fhandler_base *
-dtable::build_fhandler (int fd, DWORD dev, const char *unix_name,
+dtable::build_fhandler (int fd, DWORD dev, char *unix_name,
 			const char *win32_name, int unit)
 {
   fhandler_base *fh;
@@ -306,8 +323,7 @@ dtable::build_fhandler (int fd, DWORD dev, const char *unix_name,
       case FH_CONSOLE:
       case FH_CONIN:
       case FH_CONOUT:
-	if ((fh = cnew (fhandler_console) ()))
-	  inc_console_fds ();
+	fh = cnew (fhandler_console) ();
 	break;
       case FH_PTYM:
 	fh = cnew (fhandler_pty_master) ();
@@ -330,7 +346,7 @@ dtable::build_fhandler (int fd, DWORD dev, const char *unix_name,
 	fh = cnew (fhandler_pipe) (dev);
 	break;
       case FH_SOCKET:
-	if ((fh = cnew (fhandler_socket) ()))
+	if ((fh = cnew (fhandler_socket) (unit)))
 	  inc_need_fixup_before ();
 	break;
       case FH_DISK:
@@ -363,6 +379,15 @@ dtable::build_fhandler (int fd, DWORD dev, const char *unix_name,
       case FH_OSS_DSP:
 	fh = cnew (fhandler_dev_dsp) ();
 	break;
+      case FH_PROC:
+	fh = cnew (fhandler_proc) ();
+	break;
+      case FH_REGISTRY:
+	fh = cnew (fhandler_registry) ();
+	break;
+      case FH_PROCESS:
+	fh = cnew (fhandler_process) ();
+	break;
       default:
 	system_printf ("internal error -- unknown device - %p", dev);
 	fh = NULL;
@@ -389,7 +414,7 @@ dtable::build_fhandler (int fd, DWORD dev, const char *unix_name,
 fhandler_base *
 dtable::dup_worker (fhandler_base *oldfh)
 {
-  fhandler_base *newfh = build_fhandler (-1, oldfh->get_device (), NULL);
+  fhandler_base *newfh = build_fhandler (-1, oldfh->get_device ());
   *newfh = *oldfh;
   newfh->set_io_handle (NULL);
   if (oldfh->dup (newfh))
@@ -413,6 +438,7 @@ dtable::dup2 (int oldfd, int newfd)
 
   MALLOC_CHECK;
   debug_printf ("dup2 (%d, %d)", oldfd, newfd);
+  SetResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "dup");
 
   if (not_open (oldfd))
     {
@@ -421,6 +447,13 @@ dtable::dup2 (int oldfd, int newfd)
       goto done;
     }
 
+  if (newfd < 0)
+    {
+      syscall_printf ("new fd out of bounds: %d", newfd);
+      set_errno (EBADF);
+      goto done;
+    }
+
   if (newfd == oldfd)
     {
       res = 0;
@@ -433,48 +466,33 @@ dtable::dup2 (int oldfd, int newfd)
       goto done;
     }
 
-  debug_printf ("newfh->io_handle %p, oldfh->io_handle %p", newfh->get_io_handle (), fds[oldfd]->get_io_handle ());
-  SetResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "dup");
+  debug_printf ("newfh->io_handle %p, oldfh->io_handle %p",
+		newfh->get_io_handle (), fds[oldfd]->get_io_handle ());
 
-  if (newfd < 0)
+  if (!not_open (newfd))
+    close (newfd);
+  else if ((size_t) newfd < size)
+    /* nothing to do */;
+  else if (find_unused_handle (newfd) < 0)
     {
-      syscall_printf ("new fd out of bounds: %d", newfd);
-      set_errno (EBADF);
+      newfh->close ();
+      res = -1;
       goto done;
     }
 
-  if ((size_t) newfd >= size)
-   {
-     int inc_size = NOFILE_INCR * ((newfd + NOFILE_INCR - 1) / NOFILE_INCR) -
-		    size;
-     extend (inc_size);
-   }
-
-  if (!not_open (newfd))
-    _close (newfd);
   fds[newfd] = newfh;
 
-  ReleaseResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "dup");
-  MALLOC_CHECK;
-
   if ((res = newfd) <= 2)
     set_std_handle (res);
 
-  MALLOC_CHECK;
 done:
+  MALLOC_CHECK;
+  ReleaseResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "dup");
   syscall_printf ("%d = dup2 (%d, %d)", res, oldfd, newfd);
 
   return res;
 }
 
-void
-dtable::reset_unix_path_name (int fd, const char *name)
-{
-  SetResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "reset_unix_name");
-  fds[fd]->reset_unix_path_name (name);
-  ReleaseResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "reset_unix_name");
-}
-
 select_record *
 dtable::select_read (int fd, select_record *s)
 {
@@ -556,6 +574,17 @@ dtable::fixup_before_exec (DWORD target_proc_id)
   ReleaseResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "fixup_before_exec");
 }
 
+void
+dtable::set_file_pointers_for_exec ()
+{
+  SetResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "set_file_pointers_for_exec");
+  fhandler_base *fh;
+  for (size_t i = 0; i < size; i++)
+    if ((fh = fds[i]) != NULL && fh->get_flags () & O_APPEND)
+      SetFilePointer (fh->get_handle (), 0, 0, FILE_END);
+  ReleaseResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "fixup_before_exec");
+}
+
 void
 dtable::fixup_after_exec (HANDLE parent)
 {
@@ -606,7 +635,7 @@ dtable::vfork_child_dup ()
   int res = 1;
 
   /* Remove impersonation */
-  if (cygheap->user.impersonated && cygheap->user.token != INVALID_HANDLE_VALUE)
+  if (cygheap->user.issetuid ())
     RevertToSelf ();
 
   for (size_t i = 0; i < size; i++)
@@ -621,14 +650,14 @@ dtable::vfork_child_dup ()
 	goto out;
       }
 
-  /* Restore impersonation */
-  if (cygheap->user.impersonated && cygheap->user.token != INVALID_HANDLE_VALUE)
-    ImpersonateLoggedOnUser (cygheap->user.token);
-
   fds_on_hold = fds;
   fds = newtable;
 
 out:
+  /* Restore impersonation */
+  if (cygheap->user.issetuid ())
+    ImpersonateLoggedOnUser (cygheap->user.token);
+
   ReleaseResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "dup");
   return 1;
 }
@@ -640,7 +669,6 @@ dtable::vfork_parent_restore ()
 
   close_all_files ();
   fhandler_base **deleteme = fds;
-  assert (fds_on_hold != NULL);
   fds = fds_on_hold;
   fds_on_hold = NULL;
   cfree (deleteme);
@@ -678,3 +706,102 @@ dtable::vfork_child_fixup ()
 
   return;
 }
+
+#define DEVICE_PREFIX "\\device\\"
+#define DEVICE_PREFIX_LEN sizeof (DEVICE_PREFIX) - 1
+#define REMOTE "\\Device\\LanmanRedirector\\"
+#define REMOTE_LEN sizeof (REMOTE) - 1
+
+static const char *
+handle_to_fn (HANDLE h, char *posix_fn)
+{
+  OBJECT_NAME_INFORMATION *ntfn;
+  char fnbuf[32768];
+
+  memset (fnbuf, 0, sizeof (fnbuf));
+  ntfn = (OBJECT_NAME_INFORMATION *) fnbuf;
+  ntfn->Name.MaximumLength = sizeof (fnbuf) - sizeof (*ntfn);
+  ntfn->Name.Buffer = (WCHAR *) (ntfn + 1);
+
+  DWORD res = NtQueryObject (h, ObjectNameInformation, ntfn, sizeof (fnbuf), NULL);
+
+  if (res)
+    {
+      strcpy (posix_fn, unknown_file);
+      debug_printf ("NtQueryObject failed");
+      return unknown_file;
+    }
+
+  // NT seems to do this on an unopened file
+  if (!ntfn->Name.Buffer)
+    {
+      debug_printf ("nt->Name.Buffer == NULL");
+      return NULL;
+    }
+
+  ntfn->Name.Buffer[ntfn->Name.Length / sizeof (WCHAR)] = 0;
+
+  char win32_fn[MAX_PATH + 100];
+  sys_wcstombs (win32_fn, ntfn->Name.Buffer, ntfn->Name.Length);
+  debug_printf ("nt name '%s'", win32_fn);
+  if (!strncasematch (win32_fn, DEVICE_PREFIX, DEVICE_PREFIX_LEN)
+      || !QueryDosDevice (NULL, fnbuf, sizeof (fnbuf)))
+    return strcpy (posix_fn, win32_fn);
+
+  char *p = strchr (win32_fn + DEVICE_PREFIX_LEN, '\\');
+  if (!p)
+    p = strchr (win32_fn + DEVICE_PREFIX_LEN, '\0');
+
+  int n = p - win32_fn;
+  int maxmatchlen = 0;
+  char *maxmatchdos = NULL;
+  for (char *s = fnbuf; *s; s = strchr (s, '\0') + 1)
+    {
+      char device[MAX_PATH + 10];
+      device[MAX_PATH + 9] = '\0';
+      if (strchr (s, ':') == NULL)
+	continue;
+      if (!QueryDosDevice (s, device, sizeof (device) - 1))
+	continue;
+      char *q = strrchr (device, ';');
+      if (q)
+	{
+	  char *r = strchr (q, '\\');
+	  if (r)
+	    strcpy (q, r + 1);
+	}
+      int devlen = strlen (device);
+      if (device[devlen - 1] == '\\')
+	device[--devlen] = '\0';
+      if (devlen < maxmatchlen)
+	continue;
+      if (!strncasematch (device, win32_fn, devlen) ||
+	  (win32_fn[devlen] != '\0' && win32_fn[devlen] != '\\'))
+	continue;
+      maxmatchlen = devlen;
+      maxmatchdos = s;
+      debug_printf ("current match '%s'", device);
+    }
+
+  char *w32 = win32_fn;
+  if (maxmatchlen)
+    {
+      n = strlen (maxmatchdos);
+      if (maxmatchdos[n - 1] == '\\')
+	n--;
+      w32 += maxmatchlen - n;
+      memcpy (w32, maxmatchdos, n);
+      w32[n] = '\\';
+    }
+  else if (strncasematch (w32, REMOTE, REMOTE_LEN))
+    {
+      w32 += REMOTE_LEN - 2;
+      *w32 = '\\';
+      debug_printf ("remote drive");
+    }
+
+
+  debug_printf ("derived path '%s'", w32);
+  cygwin_conv_to_full_posix_path (w32, posix_fn);
+  return posix_fn;
+}
