@@ -1,6 +1,6 @@
 /* exceptions.cc
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -148,11 +148,13 @@ error_start_init (const char *buf)
       return;
     }
 
-  char myself_posix_name[MAX_PATH];
+  char pgm[MAX_PATH + 1];
+  if (!GetModuleFileName (NULL, pgm, MAX_PATH))
+    strcpy (pgm, "cygwin1.dll");
+  for (char *p = strchr (pgm, '\\'); p; p = strchr (p, '\\'))
+    *p = '/';
 
-  /* FIXME: gdb cannot use win32 paths, but what if debugger isn't gdb? */
-  cygwin_conv_to_posix_path (myself->progname, myself_posix_name);
-  __small_sprintf (debugger_command, "%s %s", buf, myself_posix_name);
+  __small_sprintf (debugger_command, "%s %s", buf, pgm);
 }
 
 static void
@@ -238,7 +240,7 @@ public:
   void init (DWORD, bool, bool); /* Called the first time that stack info is needed */
 
   /* Postfix ++ iterates over the stack, returning zero when nothing is left. */
-  int operator ++(int) { return this->walk (); }
+  int operator ++(int) { return walk (); }
 };
 
 /* The number of parameters used in STACKFRAME */
@@ -336,14 +338,13 @@ extern "C" int
 try_to_debug (bool waitloop)
 {
   debug_printf ("debugger_command '%s'", debugger_command);
-  if (*debugger_command == '\0')
+  if (*debugger_command == '\0' || being_debugged ())
     return 0;
 
   __small_sprintf (strchr (debugger_command, '\0'), " %u", GetCurrentProcessId ());
 
-  BOOL dbg;
-
-  SetThreadPriority (hMainThread, THREAD_PRIORITY_HIGHEST);
+  LONG prio = GetThreadPriority (GetCurrentThread ());
+  SetThreadPriority (GetCurrentThread (), THREAD_PRIORITY_HIGHEST);
   PROCESS_INFORMATION pi = {NULL, 0, 0, 0};
 
   STARTUPINFO si = {0, NULL, NULL, NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, NULL, NULL, NULL};
@@ -375,6 +376,7 @@ try_to_debug (bool waitloop)
 	}
     }
 
+  BOOL dbg;
   dbg = CreateProcess (NULL,
 		       debugger_command,
 		       NULL,
@@ -392,11 +394,12 @@ try_to_debug (bool waitloop)
     {
       if (!waitloop)
 	return 1;
-      SetThreadPriority (hMainThread, THREAD_PRIORITY_IDLE);
-      while (!IsDebuggerPresent ())
-	/* spin */;
-      Sleep (4000);
+      SetThreadPriority (GetCurrentThread (), THREAD_PRIORITY_IDLE);
+      while (!being_debugged ())
+	Sleep (0);
+      Sleep (2000);
       small_printf ("*** continuing from debugger call\n");
+      SetThreadPriority (GetCurrentThread (), prio);
     }
 
   /* FIXME: need to know handles of all running threads to
@@ -567,6 +570,7 @@ stack (void)
 int __stdcall
 handle_sigsuspend (sigset_t tempmask)
 {
+  sig_dispatch_pending (0);
   sigframe thisframe (mainthread);
   sigset_t oldmask = myself->getsigmask ();	// Remember for restoration
 
@@ -574,7 +578,8 @@ handle_sigsuspend (sigset_t tempmask)
 				//  interested in through.
   sigproc_printf ("old mask %x, new mask %x", oldmask, tempmask);
 
-  WaitForSingleObject (signal_arrived, INFINITE);
+  pthread_testcancel ();
+  pthread::cancelable_wait (signal_arrived, INFINITE);
 
   set_sig_errno (EINTR);	// Per POSIX
 
@@ -614,7 +619,9 @@ sig_handle_tty_stop (int sig)
     }
   sigproc_printf ("process %d stopped by signal %d, myself->ppid_handle %p",
 		  myself->pid, sig, myself->ppid_handle);
-  SuspendThread (hMainThread);
+  if (WaitForSingleObject (sigCONT, INFINITE) != WAIT_OBJECT_0)
+    api_fatal ("WaitSingleObject failed, %E");
+  (void) ResetEvent (sigCONT);
   return;
 }
 }
@@ -634,7 +641,7 @@ interruptible (DWORD pc, int testvalid = 0)
 
 # define h ((HMODULE) m.AllocationBase)
   /* Apparently Windows 95 can sometimes return bogus addresses from
-     GetThreadContext.  These resolve to an allocation base == 0.
+     GetThreadContext.  These resolve to a strange allocation base.
      These should *never* be treated as interruptible. */
   if (!h || m.State != MEM_COMMIT)
     res = 0;
@@ -663,7 +670,7 @@ sigthread::get_winapi_lock (int test)
   /* Need to do a busy loop because we can't block or a potential SuspendThread
      will hang. */
   while (InterlockedExchange (&winapi_lock, 1))
-    Sleep (1);
+    low_priority_sleep (0);
   return 1;
 }
 
@@ -685,7 +692,8 @@ interrupt_setup (int sig, void *handler, DWORD retaddr, DWORD *retaddr_on_stack,
   sigsave.retaddr = retaddr;
   sigsave.retaddr_on_stack = retaddr_on_stack;
   /* FIXME: Not multi-thread aware */
-  sigsave.newmask = myself->getsigmask () | siga.sa_mask | SIGTOMASK (sig);
+  sigsave.oldmask = myself->getsigmask ();
+  sigsave.newmask = sigsave.oldmask | siga.sa_mask | SIGTOMASK (sig);
   sigsave.sa_flags = siga.sa_flags;
   sigsave.func = (void (*)(int)) handler;
   sigsave.sig = sig;
@@ -714,15 +722,16 @@ interrupt_now (CONTEXT *ctx, int sig, void *handler, struct sigaction& siga)
 void __stdcall
 signal_fixup_after_fork ()
 {
-  if (!sigsave.sig)
-    return;
-
-  sigsave.sig = 0;
-  if (sigsave.retaddr_on_stack)
+  if (sigsave.sig)
     {
-      *sigsave.retaddr_on_stack = sigsave.retaddr;
-      set_process_mask (sigsave.oldmask);
+      sigsave.sig = 0;
+      if (sigsave.retaddr_on_stack)
+	{
+	  *sigsave.retaddr_on_stack = sigsave.retaddr;
+	  set_process_mask (sigsave.oldmask);
+	}
     }
+  sigproc_init ();
 }
 
 void __stdcall
@@ -731,9 +740,9 @@ signal_fixup_after_exec (bool isspawn)
   /* Set up child's signal handlers */
   for (int i = 0; i < NSIG; i++)
     {
-      myself->getsig(i).sa_mask = 0;
-      if (myself->getsig(i).sa_handler != SIG_IGN || isspawn)
-	myself->getsig(i).sa_handler = SIG_DFL;
+      myself->getsig (i).sa_mask = 0;
+      if (myself->getsig (i).sa_handler != SIG_IGN || isspawn)
+	myself->getsig (i).sa_handler = SIG_DFL;
     }
 }
 
@@ -760,7 +769,7 @@ interrupt_on_return (sigthread *th, int sig, void *handler, struct sigaction& si
 	return 1;
       }
 
-  sigproc_printf ("couldn't find a stack frame, i %d\n", i);
+  sigproc_printf ("couldn't find a stack frame, i %d", i);
   return 0;
 }
 
@@ -777,26 +786,33 @@ static int
 setup_handler (int sig, void *handler, struct sigaction& siga)
 {
   CONTEXT cx;
-  bool interrupted = 0;
-  HANDLE hth = NULL;
-  int res;
+  bool interrupted = false;
   sigthread *th = NULL;		// Initialization needed to shut up gcc
+  int prio = INFINITE;
 
   if (sigsave.sig)
     goto set_pending;
 
-  for (int i = 0; !interrupted && i < CALL_HANDLER_RETRY; i++)
+  for (int i = 0; i < CALL_HANDLER_RETRY; i++)
     {
+      DWORD res;
+      HANDLE hth;
+
       EnterCriticalSection (&mainthread.lock);
       if (mainthread.frame)
-	th = &mainthread;
+	{
+	  hth = NULL;
+	  th = &mainthread;
+	}
       else
 	{
 	  LeaveCriticalSection (&mainthread.lock);
 
-	  th = NULL;
+	  if (!mainthread.get_winapi_lock (1))
+	    continue;
 
 	  hth = myself->getthread2signal ();
+	  th = NULL;
 
 	  /* Suspend the thread which will receive the signal.  But first ensure that
 	     this thread doesn't have any mutos.  (FIXME: Someday we should just grab
@@ -806,20 +822,29 @@ setup_handler (int sig, void *handler, struct sigaction& siga)
 	     If one of these conditions is not true we loop for a fixed number of times
 	     since we don't want to stall the signal handler.  FIXME: Will this result in
 	     noticeable delays?
-	     If the thread is already suspended (which can occur when a program is stopped) then
-	     just queue the signal. */
+	     If the thread is already suspended (which can occur when a program has called
+	     SuspendThread on itself then just queue the signal. */
 
-	  if (!mainthread.get_winapi_lock (1))
-	    continue;
+	  EnterCriticalSection (&mainthread.lock);
 	  sigproc_printf ("suspending mainthread");
 	  res = SuspendThread (hth);
-	  mainthread.release_winapi_lock ();
-	  if (mainthread.frame)
-	    goto resume_thread;	/* In case the main thread *just* set the frame */
+	  /* Just release the lock now since we hav suspended the main thread and it
+	     definitely can't be grabbing it now.  This will have to change, of course,
+	     if/when we can send signals to other than the main thread. */
+	  LeaveCriticalSection (&mainthread.lock);
 
 	  /* Just set pending if thread is already suspended */
 	  if (res)
-	    goto set_pending;
+	    {
+	      (void) ResumeThread (hth);
+	      break;
+	    }
+
+	  mainthread.release_winapi_lock ();
+	  if (mainthread.frame)
+	    goto resume_thread;	/* We just got the frame.  What are the odds?
+				   Just loop and we'll hopefully pick it up on
+				   the next pass through. */
 
 	  muto *m;
 	  /* FIXME: Make multi-thread aware */
@@ -830,31 +855,26 @@ setup_handler (int sig, void *handler, struct sigaction& siga)
 		goto resume_thread;
 	      }
 
-	  EnterCriticalSection (&mainthread.lock);
 	  if (mainthread.frame)
-	    {
-	      th = &mainthread;
-	      goto try_to_interrupt;
-	    }
-
-	  LeaveCriticalSection (&mainthread.lock);
-
-	  cx.ContextFlags = CONTEXT_CONTROL | CONTEXT_INTEGER;
-	  if (!GetThreadContext (hth, &cx))
-	    system_printf ("couldn't get context of main thread, %E");
-	  else if (!interruptible (cx.Eip, 1))
-	    sigproc_printf ("suspended thread in a strange state pc %p, sp %p",
-			    cx.Eip, cx.Esp);
+	    th = &mainthread;
 	  else
-	    goto try_to_interrupt;
+	    {
+	      cx.ContextFlags = CONTEXT_CONTROL | CONTEXT_INTEGER;
+	      if (!GetThreadContext (hth, &cx))
+		{
+		  system_printf ("couldn't get context of main thread, %E");
+		  goto resume_thread;
+		}
+	    }
+	}
 
-	resume_thread:
-	  ResumeThread (hth);
-	  Sleep (0);
-	  continue;
+      if ((DWORD) prio != INFINITE)
+	{
+	  /* Reset the priority so we can finish this off quickly. */
+	  SetThreadPriority (GetCurrentThread (), WAIT_SIG_PRIORITY);
+	  prio = INFINITE;
 	}
 
-    try_to_interrupt:
       if (th)
 	{
 	  interrupted = interrupt_on_return (th, sig, handler, siga);
@@ -862,30 +882,35 @@ setup_handler (int sig, void *handler, struct sigaction& siga)
 	}
       else if (interruptible (cx.Eip))
 	interrupted = interrupt_now (&cx, sig, handler, siga);
-      else
+
+    resume_thread:
+      if (hth)
+	res = ResumeThread (hth);
+
+      if (interrupted)
 	break;
+
+      if ((DWORD) prio != INFINITE && !mainthread.frame)
+	prio = low_priority_sleep (SLEEP_0_STAY_LOW);
+      sigproc_printf ("couldn't interrupt.  trying again.");
     }
 
  set_pending:
   if (interrupted)
-    res = 1;
+    {
+      if ((DWORD) prio != INFINITE)
+	SetThreadPriority (GetCurrentThread (), WAIT_SIG_PRIORITY);
+      sigproc_printf ("signal successfully delivered");
+    }
   else
     {
       pending_signals = 1;	/* FIXME: Probably need to be more tricky here */
       sig_set_pending (sig);
       sig_dispatch_pending (1);
-      Sleep (0);		/* Hopefully, other process will be waking up soon. */
+      low_priority_sleep (SLEEP_0_STAY_LOW);	/* Hopefully, other process will be waking up soon. */
       sigproc_printf ("couldn't send signal %d", sig);
     }
 
-  if (!hth)
-    sigproc_printf ("good.  Didn't suspend main thread, th %p", th);
-  else
-    {
-      res = ResumeThread (hth);
-      sigproc_printf ("ResumeThread returned %d", res);
-    }
-
   sigproc_printf ("returning %d", interrupted);
   return interrupted;
 }
@@ -989,6 +1014,7 @@ sig_handle (int sig, bool thisproc)
   /* FIXME: Should we still do this if SIGCONT has a handler? */
   if (sig == SIGCONT)
     {
+      DWORD stopped = myself->process_state & PID_STOPPED;
       myself->stopsig = 0;
       myself->process_state &= ~PID_STOPPED;
       /* Clear pending stop signals */
@@ -996,10 +1022,8 @@ sig_handle (int sig, bool thisproc)
       sig_clear (SIGTSTP);
       sig_clear (SIGTTIN);
       sig_clear (SIGTTOU);
-      /* Windows 95 hangs on resuming non-suspended thread */
-      SuspendThread (hMainThread);
-      while (ResumeThread (hMainThread) > 1)
-	;
+      if (stopped)
+	SetEvent (sigCONT);
       /* process pending signals */
       sig_dispatch_pending (1);
     }
@@ -1013,7 +1037,7 @@ sig_handle (int sig, bool thisproc)
   if (handler == (void *) SIG_DFL)
     {
       if (sig == SIGCHLD || sig == SIGIO || sig == SIGCONT || sig == SIGWINCH
-          || sig == SIGURG || (thisproc && hExeced && sig == SIGINT))
+	  || sig == SIGURG || (thisproc && hExeced && sig == SIGINT))
 	{
 	  sigproc_printf ("default signal %d ignored", sig);
 	  goto done;
@@ -1067,12 +1091,15 @@ sig_handle (int sig, bool thisproc)
   /* Never returns */
 }
 
+CRITICAL_SECTION NO_COPY exit_lock;
+
 /* Cover function to `do_exit' to handle exiting even in presence of more
    exceptions.  We used to call exit, but a SIGSEGV shouldn't cause atexit
    routines to run.  */
 static void
 signal_exit (int rc)
 {
+  EnterCriticalSection (&exit_lock);
   rc = EXIT_SIGNAL | (rc << 8);
   if (exit_already++)
     myself->exit (rc);
@@ -1080,8 +1107,8 @@ signal_exit (int rc)
   /* We'd like to stop the main thread from executing but when we do that it
      causes random, inexplicable hangs.  So, instead, we set up the priority
      of this thread really high so that it should do its thing and then exit. */
-  (void) SetThreadPriority (GetCurrentThread (), THREAD_PRIORITY_TIME_CRITICAL);
   (void) SetThreadPriority (hMainThread, THREAD_PRIORITY_IDLE);
+  (void) SetThreadPriority (GetCurrentThread (), THREAD_PRIORITY_TIME_CRITICAL);
 
   /* Unlock any main thread mutos since we're executing with prejudice. */
   muto *m;
@@ -1096,6 +1123,7 @@ signal_exit (int rc)
     TerminateProcess (hExeced, rc);
 
   sigproc_printf ("about to call do_exit (%x)", rc);
+  (void) SetEvent (signal_arrived);
   do_exit (rc);
 }
 
@@ -1104,12 +1132,13 @@ HANDLE NO_COPY title_mutex = NULL;
 void
 events_init (void)
 {
-  /* title_mutex protects modification of console title. It's neccessary
+  char *name;
+  /* title_mutex protects modification of console title. It's necessary
      while finding console window handle */
 
   if (!(title_mutex = CreateMutex (&sec_all_nih, FALSE,
-				   shared_name ("title_mutex", 0))))
-    api_fatal ("can't create title mutex, %E");
+				   name = shared_name ("title_mutex", 0))))
+    api_fatal ("can't create title mutex '%s', %E", name);
 
   ProtectHandle (title_mutex);
   new_muto (mask_sync);
@@ -1127,12 +1156,12 @@ events_init (void)
   debug_printf ("windows_system_directory '%s', windows_system_directory_length %d",
 		windows_system_directory, windows_system_directory_length);
   debug_printf ("cygwin_hmodule %p", cygwin_hmodule);
+  InitializeCriticalSection (&exit_lock);
 }
 
 void
 events_terminate (void)
 {
-  ForceCloseHandle (title_mutex);
   exit_already = 1;
 }
 
@@ -1162,7 +1191,6 @@ int
 sigframe::call_signal_handler ()
 {
   return unregister () ? call_signal_handler_now () : 0;
-
 }
 
 #define pid_offset (unsigned)(((_pinfo *)NULL)->pid)
@@ -1174,7 +1202,13 @@ reset_signal_arrived ()
   sigproc_printf ("reset signal_arrived");
 }
 
-void unused_sig_wrapper ()
+static void unused_sig_wrapper () __attribute__((const, unused));
+
+#undef errno
+#define errno ((DWORD volatile) _impure_ptr) + (((char *) &_impure_ptr->_errno) - ((char *) _impure_ptr))
+
+static void
+unused_sig_wrapper ()
 {
 /* Signal cleanup stuff.  Cleans up stack (too bad that we didn't
    prototype signal handlers as __stdcall), calls _set_process_mask
@@ -1236,7 +1270,7 @@ _sigdelayed0:								\n\
 	popl	%%eax							\n\
 	jmp	*%%eax							\n\
 __no_sig_end:								\n\
-" : "=m" (sigsave.sig) : "m" (&_impure_ptr->_errno),
+" : "=m" (sigsave.sig):  "X" ((char *) &_impure_ptr->_errno),
   "g" (sigsave.retaddr), "g" (sigsave.oldmask), "g" (sigsave.sig),
     "g" (sigsave.func), "g" (sigsave.saved_errno), "g" (sigsave.newmask)
 );
