@@ -1,6 +1,6 @@
 /* external.cc: Interface to Cygwin internals from external programs.
 
-   Copyright 1997, 1998, 1999, 2000, 2001 Red Hat, Inc.
+   Copyright 1997, 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
 
    Written by Christopher Faylor <cgf@cygnus.com>
 
@@ -14,7 +14,6 @@ details. */
 #include <errno.h>
 #include "security.h"
 #include "fhandler.h"
-#include "sync.h"
 #include "sigproc.h"
 #include "pinfo.h"
 #include <exceptions.h>
@@ -28,6 +27,7 @@ details. */
 #include "cygheap.h"
 #include "wincap.h"
 #include "heap.h"
+#include "cygthread.h"
 
 static external_pinfo *
 fillout_pinfo (pid_t pid, int winpid)
@@ -43,7 +43,7 @@ fillout_pinfo (pid_t pid, int winpid)
   static unsigned int i;
   if (!pids.npids || !nextpid)
     {
-      pids.init (winpid);
+      pids.set (winpid);
       i = 0;
     }
 
@@ -84,9 +84,12 @@ fillout_pinfo (pid_t pid, int winpid)
 	  ep.rusage_children = p->rusage_children;
 	  strcpy (ep.progname, p->progname);
 	  ep.strace_mask = 0;
-	  ep.strace_file = 0;
+	  ep.version = EXTERNAL_PINFO_VERSION;
 
 	  ep.process_state = p->process_state;
+
+	  ep.uid32 = p->uid;
+	  ep.gid32 = p->gid;
 	  break;
 	}
     }
@@ -143,13 +146,12 @@ cygwin_internal (cygwin_getinfo_types t, ...)
 	return 1;
 
       case CW_GETTHREADNAME:
-	return (DWORD) threadname (va_arg (arg, DWORD));
+	return (DWORD) cygthread::name (va_arg (arg, DWORD));
 
       case CW_SETTHREADNAME:
 	{
-	  char *name = va_arg (arg, char *);
-	  regthread (name, va_arg (arg, DWORD));
-	  return 1;
+	  set_errno (ENOSYS);
+	  return 0;
 	}
 
       case CW_GETPINFO:
@@ -159,9 +161,8 @@ cygwin_internal (cygwin_getinfo_types t, ...)
 	return (DWORD) cygwin_version_strings;
 
       case CW_READ_V1_MOUNT_TABLES:
-	/* Upgrade old v1 registry mounts to new location. */
-	mount_table->import_v1_mounts ();
-	return 0;
+	set_errno (ENOSYS);
+	return 1;
 
       case CW_USER_DATA:
 	return (DWORD) &__cygwin_user_data;
@@ -181,7 +182,7 @@ cygwin_internal (cygwin_getinfo_types t, ...)
 	return (DWORD) fillout_pinfo (va_arg (arg, pid_t), 1);
 
       case CW_INIT_EXCEPTIONS:
-	init_exceptions ((exception_list *) arg);
+	init_exceptions (va_arg (arg, exception_list *));
 	return 0;
 
       case CW_GET_CYGDRIVE_INFO:
@@ -195,16 +196,15 @@ cygwin_internal (cygwin_getinfo_types t, ...)
 
       case CW_SET_CYGWIN_REGISTRY_NAME:
 	{
-#	  define cr ((char *) arg)
+	  const char *cr = va_arg (arg, char *);
 	  if (check_null_empty_str_errno (cr))
 	    return (DWORD) NULL;
 	  cygheap->cygwin_regname = (char *) crealloc (cygheap->cygwin_regname,
 						       strlen (cr) + 1);
 	  strcpy (cygheap->cygwin_regname, cr);
+	}
       case CW_GET_CYGWIN_REGISTRY_NAME:
 	  return (DWORD) cygheap->cygwin_regname;
-#	  undef cr
-	}
 
       case CW_STRACE_TOGGLE:
 	{
@@ -232,6 +232,21 @@ cygwin_internal (cygwin_getinfo_types t, ...)
 	  pinfo p (va_arg (arg, pid_t));
 	  return p ? p->dwProcessId : 0;
 	}
+      case CW_EXTRACT_DOMAIN_AND_USER:
+	{
+	  struct passwd *pw = va_arg (arg, struct passwd *);
+	  char *domain = va_arg (arg, char *);
+	  char *user = va_arg (arg, char *);
+	  extract_nt_dom_user (pw, domain, user);
+	  return 0;
+	}
+      case CW_CMDLINE:
+	{
+	  size_t n;
+	  pid_t pid = va_arg (arg, pid_t);
+	  pinfo p (pid);
+	  return (DWORD) p->cmdline (n);
+	}
       default:
 	return (DWORD) -1;
     }
