@@ -9,11 +9,11 @@ Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
 details. */
 
 #include "winsup.h"
-#include <sys/fcntl.h>
 #include <errno.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <sys/cygwin.h>
+#include <sys/uio.h>
 #include <signal.h>
 #include "cygerrno.h"
 #include "perprocess.h"
@@ -24,9 +24,9 @@ details. */
 #include "dtable.h"
 #include "cygheap.h"
 #include "shared_info.h"
-#include "sigproc.h"
 #include "pinfo.h"
 #include <assert.h>
+#include <limits.h>
 
 static NO_COPY const int CHUNK_SIZE = 1024; /* Used for crlf conversions */
 
@@ -103,7 +103,7 @@ fhandler_base::set_readahead_valid (int val, int ch)
   if (!val)
     ralen = raixget = raixput = 0;
   if (ch != -1)
-    put_readahead(ch);
+    put_readahead (ch);
 }
 
 int
@@ -147,7 +147,8 @@ fhandler_base::get_readahead_into_buffer (char *buf, size_t buflen)
 /* Record the file name.
    Filenames are used mostly for debugging messages, and it's hoped that
    in cases where the name is really required, the filename wouldn't ever
-   be too long (e.g. devices or some such).  */
+   be too long (e.g. devices or some such).
+   The unix_path_name is also used by virtual fhandlers.  */
 void
 fhandler_base::set_name (const char *unix_path, const char *win32_path, int unit)
 {
@@ -159,8 +160,9 @@ fhandler_base::set_name (const char *unix_path, const char *win32_path, int unit
   else
     {
       const char *fmt = get_native_name ();
-      win32_path_name = (char *) cmalloc (HEAP_STR, strlen(fmt) + 16);
-      __small_sprintf (win32_path_name, fmt, unit);
+      char *w =  (char *) cmalloc (HEAP_STR, strlen (fmt) + 16);
+      __small_sprintf (w, fmt, unit);
+      win32_path_name = w;
     }
 
   if (win32_path_name == NULL)
@@ -176,12 +178,15 @@ fhandler_base::set_name (const char *unix_path, const char *win32_path, int unit
      path_conv. Ideally, we should pass in a format string and build the
      unix_path, too. */
   if (!is_device () || *win32_path_name != '\\')
-    unix_path_name = cstrdup (unix_path);
+    unix_path_name = unix_path;
   else
     {
-      unix_path_name = cstrdup (win32_path_name);
-      for (char *p = unix_path_name; (p = strchr (p, '\\')); p++)
-	*p = '/';
+      char *p = cstrdup (win32_path_name);
+      unix_path_name = p;
+      while ((p = strchr (p, '\\')) != NULL)
+	*p++ = '/';
+      if (unix_path)
+	cfree ((void *) unix_path);
     }
 
   if (unix_path_name == NULL)
@@ -192,13 +197,6 @@ fhandler_base::set_name (const char *unix_path, const char *win32_path, int unit
   namehash = hash_path_name (0, win32_path_name);
 }
 
-void
-fhandler_base::reset_unix_path_name (const char *unix_path)
-{
-  cfree (unix_path_name);
-  unix_path_name = cstrdup (unix_path);
-}
-
 /* Detect if we are sitting at EOF for conditions where Windows
    returns an error but UNIX doesn't.  */
 static int __stdcall
@@ -219,22 +217,66 @@ is_at_eof (HANDLE h, DWORD err)
   return 0;
 }
 
+void
+fhandler_base::set_flags (int flags, int supplied_bin)
+{
+  int bin;
+  int fmode;
+  debug_printf ("flags %p, supplied_bin %p", flags, supplied_bin);
+  if ((bin = flags & (O_BINARY | O_TEXT)))
+    debug_printf ("O_TEXT/O_BINARY set in flags %p", bin);
+  else if (get_r_binset () && get_w_binset ())
+    bin = get_r_binary () ? O_BINARY : O_TEXT;	// FIXME: Not quite right
+  else if ((fmode = get_default_fmode (flags)) & O_BINARY)
+    bin = O_BINARY;
+  else if (fmode & O_TEXT)
+    bin = O_TEXT;
+  else if (supplied_bin)
+    bin = supplied_bin;
+  else
+    bin = get_w_binary () || get_r_binary () || (binmode != O_TEXT)
+	  ? O_BINARY : O_TEXT;
+
+  openflags = flags | bin;
+
+  bin &= O_BINARY;
+  set_r_binary (bin);
+  set_w_binary (bin);
+  syscall_printf ("filemode set to %s", bin ? "binary" : "text");
+}
+
 /* Normal file i/o handlers.  */
 
 /* Cover function to ReadFile to achieve (as much as possible) Posix style
    semantics and use of errno.  */
-int
-fhandler_base::raw_read (void *ptr, size_t ulen)
+void
+fhandler_base::raw_read (void *ptr, size_t& ulen)
 {
-  DWORD bytes_read;
+#define bytes_read ((ssize_t) ulen)
 
-  if (!ReadFile (get_handle(), ptr, ulen, &bytes_read, 0))
+  HANDLE h = NULL;	/* grumble */
+  int prio = 0;		/* ditto */
+  DWORD len = ulen;
+
+  (ssize_t) ulen = -1;
+  if (read_state)
+    {
+      h = GetCurrentThread ();
+      prio = GetThreadPriority (h);
+      (void) SetThreadPriority (h, THREAD_PRIORITY_TIME_CRITICAL);
+      SetEvent (read_state);
+    }
+  BOOL res = ReadFile (get_handle (), ptr, len, (DWORD *) &ulen, 0);
+  if (read_state)
+    {
+      SetEvent (read_state);
+      (void) SetThreadPriority (h, prio);
+    }
+  if (!res)
     {
-      int errcode;
-
       /* Some errors are not really errors.  Detect such cases here.  */
 
-      errcode = GetLastError ();
+      DWORD  errcode = GetLastError ();
       switch (errcode)
 	{
 	case ERROR_BROKEN_PIPE:
@@ -246,23 +288,27 @@ fhandler_base::raw_read (void *ptr, size_t ulen)
 	  break;
 	case ERROR_NOACCESS:
 	  if (is_at_eof (get_handle (), errcode))
-	    return 0;
+	    {
+	      bytes_read = 0;
+	      break;
+	    }
 	case ERROR_INVALID_FUNCTION:
 	case ERROR_INVALID_PARAMETER:
+	case ERROR_INVALID_HANDLE:
 	  if (openflags & O_DIROPEN)
 	    {
 	      set_errno (EISDIR);
-	      return -1;
+	      bytes_read = -1;
+	      break;
 	    }
 	default:
 	  syscall_printf ("ReadFile %s failed, %E", unix_path_name);
 	  __seterrno_from_win_error (errcode);
-	  return -1;
+	  bytes_read = -1;
 	  break;
 	}
     }
-
-  return bytes_read;
+#undef bytes_read
 }
 
 /* Cover function to WriteFile to provide Posix interface and semantics
@@ -272,7 +318,7 @@ fhandler_base::raw_write (const void *ptr, size_t len)
 {
   DWORD bytes_written;
 
-  if (!WriteFile (get_handle(), ptr, len, &bytes_written, 0))
+  if (!WriteFile (get_handle (), ptr, len, &bytes_written, 0))
     {
       if (GetLastError () == ERROR_DISK_FULL && bytes_written > 0)
 	return bytes_written;
@@ -288,13 +334,17 @@ fhandler_base::raw_write (const void *ptr, size_t len)
 int
 fhandler_base::get_default_fmode (int flags)
 {
+  int fmode = __fmode;
   if (perfile_table)
     {
       size_t nlen = strlen (get_name ());
       unsigned accflags = ACCFLAGS (flags);
       for (__cygwin_perfile *pf = perfile_table; pf->name; pf++)
 	if (!*pf->name && ACCFLAGS (pf->flags) == accflags)
-	  return pf->flags & ~(O_RDONLY | O_WRONLY | O_RDWR);
+	  {
+	    fmode = pf->flags & ~(O_RDONLY | O_WRONLY | O_RDWR);
+	    break;
+	  }
 	else
 	  {
 	    size_t pflen = strlen (pf->name);
@@ -302,15 +352,18 @@ fhandler_base::get_default_fmode (int flags)
 	    if (pflen > nlen || (stem != get_name () && !isdirsep (stem[-1])))
 	      continue;
 	    else if (ACCFLAGS (pf->flags) == accflags && strcasematch (stem, pf->name))
-	      return pf->flags & ~(O_RDONLY | O_WRONLY | O_RDWR);
+	      {
+		fmode = pf->flags & ~(O_RDONLY | O_WRONLY | O_RDWR);
+		break;
+	      }
 	  }
     }
-  return __fmode;
+  return fmode;
 }
 
 /* Open system call handler function. */
 int
-fhandler_base::open (path_conv *, int flags, mode_t mode)
+fhandler_base::open (path_conv *pc, int flags, mode_t mode)
 {
   int res = 0;
   HANDLE x;
@@ -319,7 +372,7 @@ fhandler_base::open (path_conv *, int flags, mode_t mode)
   int creation_distribution;
   SECURITY_ATTRIBUTES sa = sec_none;
 
-  syscall_printf ("(%s, %p)", get_win32_name (), flags);
+  syscall_printf ("(%s, %p) query_open %d", get_win32_name (), flags, get_query_open ());
 
   if (get_win32_name () == NULL)
     {
@@ -360,7 +413,7 @@ fhandler_base::open (path_conv *, int flags, mode_t mode)
     creation_distribution = CREATE_NEW;
 
   if (flags & O_APPEND)
-    set_append_p();
+    set_append_p ();
 
   /* These flags are host dependent. */
   shared = wincap.shared ();
@@ -376,22 +429,24 @@ fhandler_base::open (path_conv *, int flags, mode_t mode)
     {
       char *c = strrchr (get_win32_name (), '\\');
       if ((c && c[1] == '.') || *get_win32_name () == '.')
-        file_attributes |= FILE_ATTRIBUTE_HIDDEN;
+	file_attributes |= FILE_ATTRIBUTE_HIDDEN;
     }
 #endif
 
   /* CreateFile() with dwDesiredAccess == 0 when called on remote
      share returns some handle, even if file doesn't exist. This code
      works around this bug. */
-  if (get_query_open () &&
-      isremote () &&
-      creation_distribution == OPEN_EXISTING &&
-      GetFileAttributes (get_win32_name ()) == INVALID_FILE_ATTRIBUTES)
+  if (get_query_open () && isremote () &&
+      creation_distribution == OPEN_EXISTING && pc && !pc->exists ())
     {
       set_errno (ENOENT);
       goto done;
     }
 
+  /* If mode has no write bits set, we set the R/O attribute. */
+  if (!(mode & (S_IWUSR | S_IWGRP | S_IWOTH)))
+    file_attributes |= FILE_ATTRIBUTE_READONLY;
+
   /* If the file should actually be created and ntsec is on,
      set files attributes. */
   if (flags & O_CREAT && get_device () == FH_DISK && allow_ntsec && has_acls ())
@@ -400,59 +455,31 @@ fhandler_base::open (path_conv *, int flags, mode_t mode)
   x = CreateFile (get_win32_name (), access, shared, &sa, creation_distribution,
 		  file_attributes, 0);
 
-  syscall_printf ("%p = CreateFileA (%s, %p, %p, %p, %p, %p, 0)",
+  if (x == INVALID_HANDLE_VALUE)
+    {
+      if (!wincap.can_open_directories () && pc && pc->isdir ())
+       {
+	 if (flags & (O_CREAT | O_EXCL) == (O_CREAT | O_EXCL))
+	   set_errno (EEXIST);
+	 else if (flags & (O_WRONLY | O_RDWR))
+	   set_errno (EISDIR);
+	 else
+	   set_nohandle (true);
+       }
+      else if (GetLastError () == ERROR_INVALID_HANDLE)
+       set_errno (ENOENT);
+      else
+       __seterrno ();
+      if (!get_nohandle ())
+       goto done;
+   }
+
+  syscall_printf ("%p = CreateFile (%s, %p, %p, %p, %p, %p, 0)",
 		  x, get_win32_name (), access, shared, &sa,
 		  creation_distribution, file_attributes);
 
-  if (x == INVALID_HANDLE_VALUE)
-    {
-      if (GetLastError () == ERROR_INVALID_HANDLE)
-	set_errno (ENOENT);
-      else
-	__seterrno ();
-      goto done;
-    }
-
-  /* Attributes may be set only if a file is _really_ created.
-     This code is now only used for ntea here since the files
-     security attributes are set in CreateFile () now. */
-  if (flags & O_CREAT && get_device () == FH_DISK
-      && GetLastError () != ERROR_ALREADY_EXISTS
-      && !allow_ntsec && allow_ntea)
-    set_file_attribute (has_acls (), get_win32_name (), mode);
-
   set_io_handle (x);
-  int bin;
-  int fmode;
-  if ((bin = flags & (O_BINARY | O_TEXT)))
-    /* nothing to do */;
-  else if ((fmode = get_default_fmode (flags)) & O_BINARY)
-    bin = O_BINARY;
-  else if (fmode & O_TEXT)
-    bin = O_TEXT;
-  else if (get_device () == FH_DISK)
-    bin = get_w_binary () || get_r_binary ();
-  else
-    bin = (binmode == O_BINARY) || get_w_binary () || get_r_binary ();
-
-  if (bin & O_TEXT)
-    bin = 0;
-
-  set_flags (flags | (bin ? O_BINARY : O_TEXT));
-
-  set_r_binary (bin);
-  set_w_binary (bin);
-  syscall_printf ("filemode set to %s", bin ? "binary" : "text");
-
-  if (get_device () != FH_TAPE
-      && get_device () != FH_FLOPPY
-      && get_device () != FH_SERIAL)
-    {
-      if (flags & O_APPEND)
-	SetFilePointer (get_handle(), 0, 0, FILE_END);
-      else
-	SetFilePointer (get_handle(), 0, 0, FILE_BEGIN);
-    }
+  set_flags (flags, pc ? pc->binmode () : 0);
 
   res = 1;
   set_open_status ();
@@ -470,14 +497,13 @@ done:
    an \n.  If last char is an \r, look ahead one more char, if \n then
    modify \r, if not, remember char.
 */
-int
-fhandler_base::read (void *in_ptr, size_t in_len)
+void
+fhandler_base::read (void *in_ptr, size_t& len)
 {
-  int len = (int) in_len;
   char *ptr = (char *) in_ptr;
-
+  ssize_t copied_chars = 0;
+  bool need_signal = !!read_state;
   int c;
-  int copied_chars = 0;
 
   while (len)
     if ((c = get_readahead ()) < 0)
@@ -489,73 +515,66 @@ fhandler_base::read (void *in_ptr, size_t in_len)
       }
 
   if (copied_chars && is_slow ())
-    return copied_chars;
-
-  if (len)
     {
-      int readlen = raw_read (ptr + copied_chars, len);
-      if (copied_chars == 0)
-	copied_chars = readlen;		/* Propagate error or EOF */
-      else if (readlen > 0)		/* FIXME: should flag EOF for next read */
-	copied_chars += readlen;
+      len = (size_t) copied_chars;
+      goto out;
     }
 
-  if (copied_chars <= 0)
-    return copied_chars;
-  if (get_r_binary ())
+  if (!len)
     {
-      debug_printf ("returning %d chars, binary mode", copied_chars);
-      return copied_chars;
+      len = (size_t) copied_chars;
+      goto out;
     }
 
-#if 0
-  char *ctrlzpos;
-  /* Scan buffer for a control-z and shorten the buffer to that length */
+  raw_read (ptr + copied_chars, len);
+  need_signal = false;
+  if (!copied_chars)
+    /* nothing */;
+  else if ((ssize_t) len > 0)
+    len += copied_chars;
+  else
+    len = copied_chars;
 
-  ctrlzpos = (char *) memchr ((char *) ptr, 0x1a, copied_chars);
-  if (ctrlzpos)
-    {
-      lseek ((ctrlzpos - ((char *) ptr + copied_chars)), SEEK_CUR);
-      copied_chars = ctrlzpos - (char *) ptr;
-    }
-
-  if (copied_chars == 0)
-    {
-      debug_printf ("returning 0 chars, text mode, CTRL-Z found");
-      return 0;
-    }
-#endif
+  if (get_r_binary () || len <= 0)
+    goto out;
 
   /* Scan buffer and turn \r\n into \n */
-  register char *src = (char *) ptr;
-  register char *dst = (char *) ptr;
-  register char *end = src + copied_chars - 1;
+  char *src, *dst, *end;
+  src = (char *) ptr;
+  dst = (char *) ptr;
+  end = src + len - 1;
 
   /* Read up to the last but one char - the last char needs special handling */
   while (src < end)
     {
-      *dst = *src++;
-      if (*dst != '\r' || *src != '\n')
-	dst++;
+      if (*src == '\r' && src[1] == '\n')
+	src++;
+      *dst++ = *src++;
     }
 
-  c = *src;
-  /* if last char is a '\r' then read one more to see if we should
-     translate this one too */
-  if (c == '\r')
+  /* If not beyond end and last char is a '\r' then read one more
+     to see if we should translate this one too */
+  if (src > end)
+    /* nothing */;
+  else if (*src != '\r')
+    *dst++ = *src;
+  else
     {
-      char c1 = 0;
-      len = raw_read (&c1, 1);
-      if (len <= 0)
+      char c1;
+      size_t c1len = 1;
+      raw_read (&c1, c1len);
+      if (c1len <= 0)
 	/* nothing */;
       else if (c1 == '\n')
-	c = '\n';
+	*dst++ = '\n';
       else
-	set_readahead_valid (1, c1);
+	{
+	  set_readahead_valid (1, c1);
+	  *dst++ = *src;
+	}
     }
 
-  *dst++ = c;
-  copied_chars = dst - (char *) ptr;
+  len = dst - (char *) ptr;
 
 #ifndef NOSTRACE
   if (strace.active)
@@ -575,8 +594,13 @@ fhandler_base::read (void *in_ptr, size_t in_len)
     }
 #endif
 
-  debug_printf ("returning %d chars, text mode", copied_chars);
-  return copied_chars;
+out:
+  if (need_signal)
+    SetEvent (read_state);
+
+  debug_printf ("returning %d, %s mode", len,
+		get_r_binary () ? "binary" : "text");
+  return;
 }
 
 int
@@ -585,7 +609,7 @@ fhandler_base::write (const void *ptr, size_t len)
   int res;
 
   if (get_append_p ())
-    SetFilePointer (get_handle(), 0, 0, FILE_END);
+    SetFilePointer (get_handle (), 0, 0, FILE_END);
   else if (wincap.has_lseek_bug () && get_check_win95_lseek_bug ())
     {
       /* Note: this bug doesn't happen on NT4, even though the documentation
@@ -601,7 +625,7 @@ fhandler_base::write (const void *ptr, size_t len)
 	     fill in the gap with zeros. - DJ */
 	  char zeros[512];
 	  int number_of_zeros_to_write = current_position - actual_length;
-	  memset(zeros, 0, 512);
+	  memset (zeros, 0, 512);
 	  SetFilePointer (get_handle (), 0, 0, FILE_END);
 	  while (number_of_zeros_to_write > 0)
 	    {
@@ -693,10 +717,129 @@ fhandler_base::write (const void *ptr, size_t len)
   return res;
 }
 
-__off32_t
-fhandler_base::lseek (__off32_t offset, int whence)
+ssize_t
+fhandler_base::readv (const struct iovec *const iov, const int iovcnt,
+		      ssize_t tot)
 {
-  __off32_t res;
+  assert (iov);
+  assert (iovcnt >= 1);
+
+  if (iovcnt == 1)
+    {
+      size_t len = iov->iov_len;
+      read (iov->iov_base, len);
+      return len;
+    }
+
+  if (tot == -1)		// i.e. if not pre-calculated by the caller.
+    {
+      tot = 0;
+      const struct iovec *iovptr = iov + iovcnt;
+      do
+	{
+	  iovptr -= 1;
+	  tot += iovptr->iov_len;
+	}
+      while (iovptr != iov);
+    }
+
+  assert (tot >= 0);
+
+  if (tot == 0)
+    return 0;
+
+  char *buf = (char *) alloca (tot);
+
+  if (!buf)
+    {
+      set_errno (ENOMEM);
+      return -1;
+    }
+
+  read (buf, (size_t) tot);
+
+  const struct iovec *iovptr = iov;
+  int nbytes = tot;
+
+  while (nbytes > 0)
+    {
+      const int frag = min (nbytes, (ssize_t) iovptr->iov_len);
+      memcpy (iovptr->iov_base, buf, frag);
+      buf += frag;
+      iovptr += 1;
+      nbytes -= frag;
+    }
+
+  return tot;
+}
+
+ssize_t
+fhandler_base::writev (const struct iovec *const iov, const int iovcnt,
+		       ssize_t tot)
+{
+  assert (iov);
+  assert (iovcnt >= 1);
+
+  if (iovcnt == 1)
+    return write (iov->iov_base, iov->iov_len);
+
+  if (tot == -1)		// i.e. if not pre-calculated by the caller.
+    {
+      tot = 0;
+      const struct iovec *iovptr = iov + iovcnt;
+      do
+	{
+	  iovptr -= 1;
+	  tot += iovptr->iov_len;
+	}
+      while (iovptr != iov);
+    }
+
+  assert (tot >= 0);
+
+  if (tot == 0)
+    return 0;
+
+  char *const buf = (char *) alloca (tot);
+
+  if (!buf)
+    {
+      set_errno (ENOMEM);
+      return -1;
+    }
+
+  char *bufptr = buf;
+  const struct iovec *iovptr = iov;
+  int nbytes = tot;
+
+  while (nbytes != 0)
+    {
+      const int frag = min (nbytes, (ssize_t) iovptr->iov_len);
+      memcpy (bufptr, iovptr->iov_base, frag);
+      bufptr += frag;
+      iovptr += 1;
+      nbytes -= frag;
+    }
+
+  return write (buf, tot);
+}
+
+__off64_t
+fhandler_base::lseek (__off64_t offset, int whence)
+{
+  __off64_t res;
+
+  /* 9x/Me doesn't support 64bit offsets.  We trap that here and return
+     EINVAL.  It doesn't make sense to simulate bigger offsets by a
+     SetFilePointer sequence since FAT and FAT32 don't support file
+     size >= 4GB anyway. */
+  if (!wincap.has_64bit_file_access ()
+      && (offset < LONG_MIN || offset > LONG_MAX))
+    {
+      debug_printf ("Win9x, offset not 32 bit.");
+      set_errno (EINVAL);
+      return (__off64_t)-1;
+    }
 
   /* Seeks on text files is tough, we rewind and read till we get to the
      right place.  */
@@ -708,70 +851,24 @@ fhandler_base::lseek (__off32_t offset, int whence)
       set_readahead_valid (0);
     }
 
-  debug_printf ("lseek (%s, %d, %d)", unix_path_name, offset, whence);
-
-#if 0	/* lseek has no business messing about with text-mode stuff */
-
-  if (!get_r_binary ())
-    {
-      int newplace;
-
-      if (whence == 0)
-	{
-	  newplace = offset;
-	}
-      else if (whence ==1)
-	{
-	  newplace = rpos +  offset;
-	}
-      else
-	{
-	  /* Seek from the end of a file.. */
-	  if (rsize == -1)
-	    {
-	      /* Find the size of the file by reading till the end */
-
-	      char b[CHUNK_SIZE];
-	      while (read (b, sizeof (b)) > 0)
-		;
-	      rsize = rpos;
-	    }
-	  newplace = rsize + offset;
-	}
-
-      if (rpos > newplace)
-	{
-	  SetFilePointer (handle, 0, 0, 0);
-	  rpos = 0;
-	}
-
-      /* You can never shrink something more than 50% by turning CRLF into LF,
-	 so we binary chop looking for the right place */
-
-      while (rpos < newplace)
-	{
-	  char b[CHUNK_SIZE];
-	  size_t span = (newplace - rpos) / 2;
-	  if (span == 0)
-	    span = 1;
-	  if (span > sizeof (b))
-	    span = sizeof (b);
-
-	  debug_printf ("lseek (%s, %d, %d) span %d, rpos %d newplace %d",
-		       name, offset, whence,span,rpos, newplace);
-	  read (b, span);
-	}
-
-      debug_printf ("Returning %d", newplace);
-      return newplace;
-    }
-#endif	/* end of deleted code dealing with text mode */
+  debug_printf ("lseek (%s, %D, %d)", unix_path_name, offset, whence);
 
   DWORD win32_whence = whence == SEEK_SET ? FILE_BEGIN
 		       : (whence == SEEK_CUR ? FILE_CURRENT : FILE_END);
 
-  res = SetFilePointer (get_handle(), offset, 0, win32_whence);
-  if (res == -1)
+  LONG off_low = offset & 0xffffffff;
+  LONG *poff_high, off_high;
+  if (!wincap.has_64bit_file_access ())
+    poff_high = NULL;
+  else
+    {
+      off_high =  offset >> 32;
+      poff_high = &off_high;
+    }
+
+  debug_printf ("setting file pointer to %u (high), %u (low)", off_high, off_low);
+  res = SetFilePointer (get_handle (), off_low, poff_high, win32_whence);
+  if (res == INVALID_SET_FILE_POINTER && GetLastError ())
     {
       __seterrno ();
     }
@@ -796,12 +893,12 @@ fhandler_base::close ()
 {
   int res = -1;
 
-  syscall_printf ("closing '%s' handle %p", get_name (), get_handle());
-  if (CloseHandle (get_handle()))
+  syscall_printf ("closing '%s' handle %p", get_name (), get_handle ());
+  if (get_nohandle () || CloseHandle (get_handle ()))
     res = 0;
   else
     {
-      paranoid_printf ("CloseHandle (%d <%s>) failed", get_handle(),
+      paranoid_printf ("CloseHandle (%d <%s>) failed", get_handle (),
 		       get_name ());
 
       __seterrno ();
@@ -824,12 +921,12 @@ fhandler_base::ioctl (unsigned int cmd, void *buf)
 int
 fhandler_base::lock (int, struct flock *)
 {
-  set_errno (ENOSYS);
+  set_errno (EINVAL);
   return -1;
 }
 
 extern "C" char * __stdcall
-rootdir(char *full_path)
+rootdir (char *full_path)
 {
   /* Possible choices:
    * d:... -> d:/
@@ -862,27 +959,32 @@ rootdir(char *full_path)
 }
 
 int __stdcall
-fhandler_base::fstat (struct stat *buf, path_conv *)
+fhandler_base::fstat (struct __stat64 *buf, path_conv *)
 {
+  debug_printf ("here");
   switch (get_device ())
     {
+    case FH_PIPE:
+      buf->st_mode = S_IFIFO | STD_RBITS | STD_WBITS | S_IWGRP | S_IWOTH;
+      break;
     case FH_PIPEW:
-      buf->st_mode = STD_WBITS | S_IWGRP | S_IWOTH;
+      buf->st_mode = S_IFIFO | STD_WBITS | S_IWGRP | S_IWOTH;
       break;
     case FH_PIPER:
-      buf->st_mode = STD_RBITS;
+      buf->st_mode = S_IFIFO | STD_RBITS;
+      break;
+    case FH_FLOPPY:
+      buf->st_mode = S_IFBLK | STD_RBITS | STD_WBITS | S_IWGRP | S_IWOTH;
       break;
     default:
-      buf->st_mode = STD_RBITS | STD_WBITS | S_IWGRP | S_IWOTH;
+      buf->st_mode = S_IFCHR | STD_RBITS | STD_WBITS | S_IWGRP | S_IWOTH;
       break;
     }
 
-  buf->st_mode |= get_device () == FH_FLOPPY ? S_IFBLK : S_IFCHR;
   buf->st_nlink = 1;
   buf->st_blksize = S_BLKSIZE;
-  buf->st_dev = buf->st_rdev = FHDEVN (get_device ()) << 8 | (get_unit () & 0xff);
-  buf->st_ino = get_namehash ();
-  buf->st_atime = buf->st_mtime = buf->st_ctime = time (NULL) - 1;
+  time_as_timestruc_t (&buf->st_ctim);
+  buf->st_atim = buf->st_mtim = buf->st_ctim;
   return 0;
 }
 
@@ -890,18 +992,18 @@ void
 fhandler_base::init (HANDLE f, DWORD a, mode_t bin)
 {
   set_io_handle (f);
-  set_r_binary (bin);
-  set_w_binary (bin);
   access = a;
   a &= GENERIC_READ | GENERIC_WRITE;
+  int flags = 0;
   if (a == GENERIC_READ)
-    set_flags (O_RDONLY);
-  if (a == GENERIC_WRITE)
-    set_flags (O_WRONLY);
-  if (a == (GENERIC_READ | GENERIC_WRITE))
-    set_flags (O_RDWR);
+    flags = O_RDONLY;
+  else if (a == GENERIC_WRITE)
+    flags = O_WRONLY;
+  else if (a == (GENERIC_READ | GENERIC_WRITE))
+    flags = O_RDWR;
+  set_flags (flags | bin);
   set_open_status ();
-  debug_printf ("created new fhandler_base for handle %p", f);
+  debug_printf ("created new fhandler_base for handle %p, bin %d", f, get_r_binary ());
 }
 
 void
@@ -916,18 +1018,19 @@ fhandler_base::dup (fhandler_base *child)
   debug_printf ("in fhandler_base dup");
 
   HANDLE nh;
-  if (get_nohandle ())
-    nh = NULL;
-  else if (!DuplicateHandle (hMainProc, get_handle(), hMainProc, &nh, 0, TRUE,
-			DUPLICATE_SAME_ACCESS))
+  if (!get_nohandle ())
     {
-      system_printf ("dup(%s) failed, handle %x, %E",
-		     get_name (), get_handle());
-      __seterrno ();
-      return -1;
-    }
+      if (!DuplicateHandle (hMainProc, get_handle (), hMainProc, &nh, 0, TRUE,
+			    DUPLICATE_SAME_ACCESS))
+	{
+	  system_printf ("dup(%s) failed, handle %x, %E",
+			 get_name (), get_handle ());
+	  __seterrno ();
+	  return -1;
+	}
 
-  child->set_io_handle (nh);
+      child->set_io_handle (nh);
+    }
   return 0;
 }
 
@@ -1059,25 +1162,18 @@ fhandler_base::fhandler_base (DWORD devtype, int unit):
   rabuflen (0),
   unix_path_name (NULL),
   win32_path_name (NULL),
-  open_status (0)
+  open_status (0),
+  read_state (NULL)
 {
-  int bin = __fmode & O_TEXT ? 0 : 1;
-  if (status != FH_DISK && status != FH_CONSOLE)
-    {
-      if (!get_r_binset ())
-	set_r_binary (bin);
-      if (!get_w_binset ())
-	set_w_binary (bin);
-    }
 }
 
 /* Normal I/O destructor */
 fhandler_base::~fhandler_base (void)
 {
   if (unix_path_name != NULL)
-    cfree (unix_path_name);
+    cfree ((void *) unix_path_name);
   if (win32_path_name != NULL)
-    cfree (win32_path_name);
+    cfree ((void *) win32_path_name);
   if (rabuf)
     free (rabuf);
   unix_path_name = win32_path_name = NULL;
@@ -1100,7 +1196,7 @@ fhandler_dev_null::dump (void)
 void
 fhandler_base::set_inheritance (HANDLE &h, int not_inheriting)
 {
-#ifdef DEBUGGING
+#ifdef DEBUGGING_AND_FDS_PROTECTED
   HANDLE oh = h;
 #endif
   /* Note that we could use SetHandleInformation here but it is not available
@@ -1111,37 +1207,33 @@ fhandler_base::set_inheritance (HANDLE &h, int not_inheriting)
   if (!DuplicateHandle (hMainProc, h, hMainProc, &h, 0, !not_inheriting,
 			     DUPLICATE_SAME_ACCESS | DUPLICATE_CLOSE_SOURCE))
     debug_printf ("DuplicateHandle failed, %E");
-#ifdef DEBUGGING
+#ifdef DEBUGGING_AND_FDS_PROTECTED
   if (h)
-    setclexec_pid (oh, h, not_inheriting);
+    setclexec (oh, h, not_inheriting);
 #endif
 }
 
 void
 fhandler_base::fork_fixup (HANDLE parent, HANDLE &h, const char *name)
 {
-#ifdef DEBUGGING
-  HANDLE oh = h;
-#endif
   if (/* !is_socket () && */ !get_close_on_exec ())
     debug_printf ("handle %p already opened", h);
   else if (!DuplicateHandle (parent, h, hMainProc, &h, 0, !get_close_on_exec (),
 			     DUPLICATE_SAME_ACCESS))
     system_printf ("%s - %E, handle %s<%p>", get_name (), name, h);
-#ifdef DEBUGGING
+#ifdef DEBUGGING_AND_FDS_PROTECTED
+  else if (get_close_on_exec ())
+    ProtectHandle (h);	/* would have to be fancier than this */
   else
-    {
-      debug_printf ("%s success - oldh %p, h %p", get_name (), oh, h);
-      // someday, maybe ProtectHandle2 (h, name);
-      setclexec_pid (oh, h, !get_close_on_exec ());
-    }
+    /* ProtectHandleINH (h) */;	/* Should already be protected */
 #endif
 }
 
 void
 fhandler_base::set_close_on_exec (int val)
 {
-  set_inheritance (io_handle, val);
+  if (!get_nohandle ())
+    set_inheritance (io_handle, val);
   set_close_on_exec_flag (val);
   debug_printf ("set close_on_exec for %s to %d", get_name (), val);
 }
@@ -1150,7 +1242,8 @@ void
 fhandler_base::fixup_after_fork (HANDLE parent)
 {
   debug_printf ("inheriting '%s' from parent", get_name ());
-  fork_fixup (parent, io_handle, "io_handle");
+  if (!get_nohandle ())
+    fork_fixup (parent, io_handle, "io_handle");
 }
 
 bool
@@ -1181,7 +1274,7 @@ fhandler_base::readdir (DIR *)
   return NULL;
 }
 
-__off32_t
+__off64_t
 fhandler_base::telldir (DIR *)
 {
   set_errno (ENOTDIR);
@@ -1189,7 +1282,7 @@ fhandler_base::telldir (DIR *)
 }
 
 void
-fhandler_base::seekdir (DIR *, __off32_t)
+fhandler_base::seekdir (DIR *, __off64_t)
 {
   set_errno (ENOTDIR);
   return;
