@@ -1,6 +1,6 @@
 /* fhandler.h
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -12,6 +12,7 @@ details. */
 #define _FHANDLER_H_
 
 #include <sys/ioctl.h>
+#include <fcntl.h>
 
 enum
 {
@@ -70,8 +71,11 @@ enum
   FH_CLIPBOARD = 0x00000017,	/* is a clipboard device */
   FH_OSS_DSP = 0x00000018,	/* is a dsp audio device */
   FH_CYGDRIVE= 0x00000019,	/* /cygdrive/x */
+  FH_PROC    = 0x0000001a,      /* /proc */
+  FH_REGISTRY =0x0000001b,      /* /proc/registry */
+  FH_PROCESS = 0x0000001c,      /* /proc/<n> */
 
-  FH_NDEV    = 0x0000001a,	/* Maximum number of devices */
+  FH_NDEV    = 0x0000001d,      /* Maximum number of devices */
   FH_DEVMASK = 0x00000fff,	/* devices live here */
   FH_BAD     = 0xffffffff
 };
@@ -97,15 +101,31 @@ enum
    both flags are set. */
 #define O_NONBLOCK_MASK (O_NONBLOCK | OLD_O_NDELAY)
 
+#define UNCONNECTED     0
+#define CONNECT_PENDING 1
+#define CONNECTED       2
+
 extern const char *windows_device_names[];
 extern struct __cygwin_perfile *perfile_table;
 #define __fmode (*(user_data->fmode_ptr))
+extern const char proc[];
+extern const int proc_len;
 
 class select_record;
 class path_conv;
 class fhandler_disk_file;
 typedef struct __DIR DIR;
 struct dirent;
+struct iovec;
+
+enum line_edit_status
+{
+  line_edit_ok = 0,
+  line_edit_input_done = 1,
+  line_edit_signalled = 2,
+  line_edit_error = 3,
+  line_edit_pipe_full = 4
+};
 
 enum bg_check_types
 {
@@ -118,8 +138,8 @@ enum bg_check_types
 enum executable_states
 {
   is_executable,
-  not_executable,
   dont_care_if_executable,
+  not_executable = dont_care_if_executable,
   dont_know_if_executable
 };
 
@@ -144,15 +164,14 @@ class fhandler_base
   size_t raixput;
   size_t rabuflen;
 
-  char *unix_path_name;
-  char *win32_path_name;
+  const char *unix_path_name;
+  const char *win32_path_name;
   DWORD open_status;
+  HANDLE read_state;
 
  public:
-  void set_name (const char * unix_path, const char * win32_path = NULL,
-		 int unit = 0);
+  void set_name (const char * unix_path, const char *win32_path = NULL, int unit = 0);
 
-  void reset_unix_path_name (const char *);
   virtual fhandler_base& operator =(fhandler_base &x);
   fhandler_base (DWORD dev, int unit = 0);
   virtual ~fhandler_base ();
@@ -171,13 +190,13 @@ class fhandler_base
   void set_async (int x) { FHCONDSETF (x, ASYNC); }
 
   int get_flags () { return openflags; }
-  void set_flags (int x) { openflags = x; }
+  void set_flags (int x, int supplied_bin = 0);
 
   bool is_nonblocking ();
   void set_nonblocking (int yes);
 
-  bool get_w_binary () { return FHISSETF (WBINARY); }
-  bool get_r_binary () { return FHISSETF (RBINARY); }
+  bool get_w_binary () { return FHISSETF (WBINSET) ? FHISSETF (WBINARY) : 1; }
+  bool get_r_binary () { return FHISSETF (RBINSET) ? FHISSETF (RBINARY) : 1; }
 
   bool get_w_binset () { return FHISSETF (WBINSET); }
   bool get_r_binset () { return FHISSETF (RBINSET); }
@@ -194,15 +213,15 @@ class fhandler_base
   DWORD get_open_status () {return open_status;}
   void reset_to_open_binmode ()
   {
-    status = status & ~(FH_WBINARY | FH_WBINSET | FH_RBINARY | FH_RBINSET);
-    status = status | ((FH_WBINARY | FH_WBINSET | FH_RBINARY | FH_RBINSET)
-		       & open_status);
+    set_flags ((get_flags () & ~(O_TEXT | O_BINARY))
+	       | ((open_status & (FH_WBINARY | FH_RBINARY)
+		   ? O_BINARY : O_TEXT)));
   }
 
   int get_default_fmode (int flags);
 
   bool get_r_no_interrupt () { return FHISSETF (NOEINTR); }
-  void set_r_no_interrupt (int b) { FHCONDSETF (b, NOEINTR); }
+  void set_r_no_interrupt (bool b) { FHCONDSETF (b, NOEINTR); }
 
   bool get_close_on_exec () { return FHISSETF (CLOEXEC); }
   int set_close_on_exec_flag (int b) { return FHCONDSETF (b, CLOEXEC); }
@@ -243,13 +262,14 @@ class fhandler_base
   }
   void set_execable_p () { FHSETF (EXECABL); }
   bool dont_care_if_execable () { return FHISSETF (DCEXEC); }
+  bool exec_state_isknown () { return FHISSETF (DCEXEC) || FHISSETF (EXECABL); }
 
   bool get_append_p () { return FHISSETF (APPEND); }
   void set_append_p (int val) { FHCONDSETF (val, APPEND); }
   void set_append_p () { FHSETF (APPEND); }
 
   bool get_query_open () { return FHISSETF (QUERYOPEN); }
-  void set_query_open (int val) { FHCONDSETF (val, QUERYOPEN); }
+  void set_query_open (bool val) { FHCONDSETF (val, QUERYOPEN); }
 
   bool get_readahead_valid () { return raixget < ralen; }
   int puts_readahead (const char *s, size_t len = (size_t) -1);
@@ -280,21 +300,23 @@ class fhandler_base
   /* fixup fd possibly non-inherited handles after fork */
   void fork_fixup (HANDLE parent, HANDLE &h, const char *name);
 
-  virtual int open (path_conv * real_path, int flags, mode_t mode = 0);
+  virtual int open (path_conv *real_path, int flags, mode_t mode = 0);
   virtual int close ();
-  virtual int __stdcall fstat (struct stat *buf, path_conv *) __attribute__ ((regparm (3)));
+  virtual int __stdcall fstat (struct __stat64 *buf, path_conv *) __attribute__ ((regparm (3)));
   virtual int ioctl (unsigned int cmd, void *);
   virtual int fcntl (int cmd, void *);
   virtual char const * ttyname () { return get_name(); }
-  virtual int __stdcall read (void *ptr, size_t len) __attribute__ ((regparm (3)));
+  virtual void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
   virtual int write (const void *ptr, size_t len);
-  virtual __off32_t lseek (__off32_t offset, int whence);
+  virtual ssize_t readv (const struct iovec *, int iovcnt, ssize_t tot = -1);
+  virtual ssize_t writev (const struct iovec *, int iovcnt, ssize_t tot = -1);
+  virtual __off64_t lseek (__off64_t offset, int whence);
   virtual int lock (int, struct flock *);
   virtual void dump ();
   virtual int dup (fhandler_base *child);
 
   virtual HANDLE mmap (caddr_t *addr, size_t len, DWORD access,
-		       int flags, __off32_t off);
+		       int flags, __off64_t off);
   virtual int munmap (HANDLE h, caddr_t addr, size_t len);
   virtual int msync (HANDLE h, caddr_t addr, size_t len, int flags);
   virtual BOOL fixup_mmap_after_fork (HANDLE h, DWORD access, DWORD offset,
@@ -319,7 +341,7 @@ class fhandler_base
   virtual class fhandler_console *is_console () { return 0; }
   virtual int is_windows () {return 0; }
 
-  virtual int raw_read (void *ptr, size_t ulen);
+  virtual void raw_read (void *ptr, size_t& ulen);
   virtual int raw_write (const void *ptr, size_t ulen);
 
   /* Virtual accessor functions to hide the fact
@@ -347,8 +369,8 @@ class fhandler_base
   virtual void set_eof () {}
   virtual DIR *opendir (path_conv& pc);
   virtual dirent *readdir (DIR *);
-  virtual __off32_t telldir (DIR *);
-  virtual void seekdir (DIR *, __off32_t);
+  virtual __off64_t telldir (DIR *);
+  virtual void seekdir (DIR *, __off64_t);
   virtual void rewinddir (DIR *);
   virtual int closedir (DIR *);
 };
@@ -357,16 +379,20 @@ class fhandler_socket: public fhandler_base
 {
  private:
   int addr_family;
+  int type;
   int connect_secret [4];
   HANDLE secret_event;
   struct _WSAPROTOCOL_INFOA *prot_info_ptr;
   char *sun_path;
+  int had_connect_or_listen;
+  int unit;
 
  public:
-  fhandler_socket ();
+  fhandler_socket (int unit);
   ~fhandler_socket ();
   int get_socket () { return (int) get_handle(); }
   fhandler_socket * is_socket () { return this; }
+  int get_unit () { return unit; }
 
   bool saw_shutdown_read () const {return FHISSETF (SHUTRD);}
   bool saw_shutdown_write () const {return FHISSETF (SHUTWR);}
@@ -374,15 +400,32 @@ class fhandler_socket: public fhandler_base
   void set_shutdown_read () {FHSETF (SHUTRD);}
   void set_shutdown_write () {FHSETF (SHUTWR);}
 
-  int recv (void *ptr, size_t len, unsigned int flags);
-  int __stdcall read (void *ptr, size_t len) __attribute__ ((regparm (3)));
+  bool is_unconnected () const {return had_connect_or_listen == UNCONNECTED;}
+  bool is_connect_pending () const {return had_connect_or_listen == CONNECT_PENDING;}
+  bool is_connected () const {return had_connect_or_listen == CONNECTED;}
+  void set_connect_state (int newstate) { had_connect_or_listen = newstate; }
 
-  int send (const void *ptr, size_t len, unsigned int flags);
-  int write (const void *ptr, size_t len);
+  int bind (const struct sockaddr *name, int namelen);
+  int connect (const struct sockaddr *name, int namelen);
+  int listen (int backlog);
+  int accept (struct sockaddr *peer, int *len);
+  int getsockname (struct sockaddr *name, int *namelen);
+  int getpeername (struct sockaddr *name, int *namelen);
+
+  ssize_t readv (const struct iovec *, int iovcnt, ssize_t tot = -1);
+  int recvfrom (void *ptr, size_t len, int flags,
+		struct sockaddr *from, int *fromlen);
+  int recvmsg (struct msghdr *msg, int flags, ssize_t tot = -1);
+
+  ssize_t writev (const struct iovec *, int iovcnt, ssize_t tot = -1);
+  int sendto (const void *ptr, size_t len, int flags,
+	      const struct sockaddr *to, int tolen);
+  int sendmsg (const struct msghdr *msg, int flags, ssize_t tot = -1);
 
   int ioctl (unsigned int cmd, void *);
   int fcntl (int cmd, void *);
-  __off32_t lseek (__off32_t, int) { return 0; }
+  __off64_t lseek (__off64_t, int) { return 0; }
+  int shutdown (int how);
   int close ();
   void hclose (HANDLE) {close ();}
   int dup (fhandler_base *child);
@@ -397,6 +440,8 @@ class fhandler_socket: public fhandler_base
   select_record *select_except (select_record *s);
   void set_addr_family (int af) {addr_family = af;}
   int get_addr_family () {return addr_family;}
+  void set_socket_type (int st) { type = st;}
+  int get_socket_type () {return type;}
   void set_sun_path (const char *path);
   char *get_sun_path () {return sun_path;}
   void set_connect_secret ();
@@ -405,7 +450,7 @@ class fhandler_socket: public fhandler_base
   int check_peer_secret_event (struct sockaddr_in *peer, int *secret = NULL);
   void signal_secret_event ();
   void close_secret_event ();
-  int __stdcall fstat (struct stat *buf, path_conv *) __attribute__ ((regparm (3)));
+  int __stdcall fstat (struct __stat64 *buf, path_conv *) __attribute__ ((regparm (3)));
 };
 
 class fhandler_pipe: public fhandler_base
@@ -417,20 +462,23 @@ class fhandler_pipe: public fhandler_base
   unsigned id;
  public:
   fhandler_pipe (DWORD devtype);
-  __off32_t lseek (__off32_t offset, int whence);
+  __off64_t lseek (__off64_t offset, int whence);
   select_record *select_read (select_record *s);
   select_record *select_write (select_record *s);
   select_record *select_except (select_record *s);
   void set_close_on_exec (int val);
-  int __stdcall read (void *ptr, size_t len) __attribute__ ((regparm (3)));
+  void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
   int close ();
   void create_guard (SECURITY_ATTRIBUTES *sa) {guard = CreateMutex (sa, FALSE, NULL);}
   int dup (fhandler_base *child);
+  int ioctl (unsigned int cmd, void *);
   void fixup_after_fork (HANDLE);
+  void fixup_after_exec (HANDLE);
   bool hit_eof ();
   void set_eof () {broken_pipe = true;}
   friend int make_pipe (int fildes[2], unsigned int psize, int mode);
   HANDLE get_guard () const {return guard;}
+  int ready_for_read (int fd, DWORD howlong);
 };
 
 class fhandler_dev_raw: public fhandler_base
@@ -466,7 +514,7 @@ class fhandler_dev_raw: public fhandler_base
   int open (path_conv *, int flags, mode_t mode = 0);
   int close (void);
 
-  int raw_read (void *ptr, size_t ulen);
+  void raw_read (void *ptr, size_t& ulen);
   int raw_write (const void *ptr, size_t ulen);
 
   int dup (fhandler_base *child);
@@ -489,7 +537,7 @@ class fhandler_dev_floppy: public fhandler_dev_raw
   virtual int open (path_conv *, int flags, mode_t mode = 0);
   virtual int close (void);
 
-  virtual __off32_t lseek (__off32_t offset, int whence);
+  virtual __off64_t lseek (__off64_t offset, int whence);
 
   virtual int ioctl (unsigned int cmd, void *buf);
 };
@@ -509,16 +557,16 @@ class fhandler_dev_tape: public fhandler_dev_raw
  public:
   fhandler_dev_tape (int unit);
 
-  int open (path_conv *, int flags, mode_t mode = 0);
-  int close (void);
+  virtual int open (path_conv *, int flags, mode_t mode = 0);
+  virtual int close (void);
 
-  __off32_t lseek (__off32_t offset, int whence);
+  virtual __off64_t lseek (__off64_t offset, int whence);
 
-  int __stdcall fstat (struct stat *buf, path_conv *) __attribute__ ((regparm (3)));
+  virtual int __stdcall fstat (struct __stat64 *buf, path_conv *) __attribute__ ((regparm (3)));
 
-  int dup (fhandler_base *child);
+  virtual int dup (fhandler_base *child);
 
-  int ioctl (unsigned int cmd, void *buf);
+  virtual int ioctl (unsigned int cmd, void *buf);
 
  private:
   int tape_write_marks (int marktype, DWORD len);
@@ -545,18 +593,29 @@ class fhandler_disk_file: public fhandler_base
   int close ();
   int lock (int, struct flock *);
   BOOL is_device () { return FALSE; }
-  int __stdcall fstat (struct stat *buf, path_conv *pc) __attribute__ ((regparm (3)));
-  int __stdcall fstat_helper (struct stat *buf) __attribute__ ((regparm (2)));
+  int __stdcall fstat (struct __stat64 *buf, path_conv *pc) __attribute__ ((regparm (3)));
+  int __stdcall fstat_helper (struct __stat64 *buf, path_conv *pc,
+			      FILETIME ftCreateionTime,
+			      FILETIME ftLastAccessTime,
+			      FILETIME ftLastWriteTime,
+			      DWORD nFileSizeHigh,
+			      DWORD nFileSizeLow,
+			      DWORD nFileIndexHigh = 0,
+			      DWORD nFileIndexLow = 0,
+			      DWORD nNumberOfLinks = 1)
+    __attribute__ ((regparm (3)));
+  int __stdcall fstat_by_handle (struct __stat64 *buf, path_conv *pc) __attribute__ ((regparm (3)));
+  int __stdcall fstat_by_name (struct __stat64 *buf, path_conv *pc) __attribute__ ((regparm (3)));
 
-  HANDLE mmap (caddr_t *addr, size_t len, DWORD access, int flags, __off32_t off);
+  HANDLE mmap (caddr_t *addr, size_t len, DWORD access, int flags, __off64_t off);
   int munmap (HANDLE h, caddr_t addr, size_t len);
   int msync (HANDLE h, caddr_t addr, size_t len, int flags);
   BOOL fixup_mmap_after_fork (HANDLE h, DWORD access, DWORD offset,
 			      DWORD size, void *address);
   DIR *opendir (path_conv& pc);
   struct dirent *readdir (DIR *);
-  __off32_t telldir (DIR *);
-  void seekdir (DIR *, __off32_t);
+  __off64_t telldir (DIR *);
+  void seekdir (DIR *, __off64_t);
   void rewinddir (DIR *);
   int closedir (DIR *);
 };
@@ -572,19 +631,21 @@ class fhandler_cygdrive: public fhandler_disk_file
   fhandler_cygdrive (int unit);
   DIR *opendir (path_conv& pc);
   struct dirent *readdir (DIR *);
-  __off32_t telldir (DIR *);
-  void seekdir (DIR *, __off32_t);
+  __off64_t telldir (DIR *);
+  void seekdir (DIR *, __off64_t);
   void rewinddir (DIR *);
   int closedir (DIR *);
-  int __stdcall fstat (struct stat *buf, path_conv *pc) __attribute__ ((regparm (3)));
+  int __stdcall fstat (struct __stat64 *buf, path_conv *pc) __attribute__ ((regparm (3)));
 };
 
 class fhandler_serial: public fhandler_base
 {
  private:
-  unsigned int vmin_;			/* from termios */
+  size_t vmin_;				/* from termios */
   unsigned int vtime_;			/* from termios */
   pid_t pgrp_;
+  int rts;				/* for Windows 9x purposes only */
+  int dtr;				/* for Windows 9x purposes only */
 
  public:
   int overlapped_armed;
@@ -599,14 +660,15 @@ class fhandler_serial: public fhandler_base
   void init (HANDLE h, DWORD a, mode_t flags);
   void overlapped_setup ();
   int dup (fhandler_base *child);
-  int raw_read (void *ptr, size_t ulen);
+  void raw_read (void *ptr, size_t& ulen);
   int raw_write (const void *ptr, size_t ulen);
   int tcsendbreak (int);
   int tcdrain ();
   int tcflow (int);
+  int ioctl (unsigned int cmd, void *);
   int tcsetattr (int a, const struct termios *t);
   int tcgetattr (struct termios *t);
-  __off32_t lseek (__off32_t, int) { return 0; }
+  __off64_t lseek (__off64_t, int) { return 0; }
   int tcflush (int);
   void dump ();
   int is_tty () { return 1; }
@@ -645,7 +707,7 @@ class fhandler_termios: public fhandler_base
     set_need_fork_fixup ();
   }
   HANDLE& get_output_handle () { return output_handle; }
-  int line_edit (const char *rptr, int nread, int always_accept = 0);
+  line_edit_status line_edit (const char *rptr, int nread, termios&);
   void set_output_handle (HANDLE h) { output_handle = h; }
   void tcinit (tty_min *this_tc, int force = FALSE);
   virtual int is_tty () { return 1; }
@@ -657,6 +719,7 @@ class fhandler_termios: public fhandler_base
   void fixup_after_fork (HANDLE);
   void fixup_after_exec (HANDLE parent) { fixup_after_fork (parent); }
   void echo_erase (int force = 0);
+  virtual __off64_t lseek (__off64_t, int);
 };
 
 enum ansi_intensity
@@ -667,21 +730,18 @@ enum ansi_intensity
   INTENSITY_BOLD
 };
 
-#define normal 1
-#define gotesc 2
-#define gotsquare 3
-#define gotarg1 4
-#define gotrsquare 5
-#define gotcommand 6
-#define gettitle 7
-#define eattitle 8
+#define normal 0
+#define gotesc 1
+#define gotsquare 2
+#define gotarg1 3
+#define gotrsquare 4
+#define gotcommand 5
+#define gettitle 6
+#define eattitle 7
 #define MAXARGS 10
 
-/* This is a input and output console handle */
-class fhandler_console: public fhandler_termios
+class dev_console
 {
- private:
-
   WORD default_color, underline_color, dim_color;
 
   /* Used to determine if an input keystroke should be modified with META. */
@@ -729,6 +789,14 @@ class fhandler_console: public fhandler_termios
   bool insert_mode;
   bool use_mouse;
   bool raw_win32_keyboard_mode;
+  friend class fhandler_console;
+};
+
+/* This is a input and output console handle */
+class fhandler_console: public fhandler_termios
+{
+ private:
+  static dev_console *dev_state;
 
 /* Output calls */
   void set_default_attr ();
@@ -751,6 +819,7 @@ class fhandler_console: public fhandler_termios
   void set_cursor_maybe ();
 
  public:
+  static int open_fhs;
 
   fhandler_console ();
 
@@ -760,7 +829,7 @@ class fhandler_console: public fhandler_termios
 
   int write (const void *ptr, size_t len);
   void doecho (const void *str, DWORD len) { (void) write (str, len); }
-  int __stdcall read (void *ptr, size_t len) __attribute__ ((regparm (3)));
+  void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
   int close ();
 
   int tcflush (int);
@@ -772,7 +841,7 @@ class fhandler_console: public fhandler_termios
 
   int ioctl (unsigned int cmd, void *);
   void init (HANDLE, DWORD, mode_t);
-  bool mouse_aware () {return use_mouse;}
+  bool mouse_aware () {return dev_state->use_mouse;}
 
   select_record *select_read (select_record *s);
   select_record *select_write (select_record *s);
@@ -781,6 +850,8 @@ class fhandler_console: public fhandler_termios
   void set_close_on_exec (int val);
   void fixup_after_fork (HANDLE parent);
   void set_input_state ();
+  void send_winch_maybe ();
+  static tty_min *get_tty_stuff (int);
 };
 
 class fhandler_tty_common: public fhandler_termios
@@ -829,17 +900,20 @@ class fhandler_tty_slave: public fhandler_tty_common
 
   int open (path_conv *, int flags, mode_t mode = 0);
   int write (const void *ptr, size_t len);
-  int __stdcall read (void *ptr, size_t len) __attribute__ ((regparm (3)));
+  void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
   void init (HANDLE, DWORD, mode_t);
 
   int tcsetattr (int a, const struct termios *t);
   int tcgetattr (struct termios *t);
   int tcflush (int);
   int ioctl (unsigned int cmd, void *);
+  int close ();
+  int dup (fhandler_base *child);
+  void fixup_after_fork (HANDLE parent);
 
-  __off32_t lseek (__off32_t, int) { return 0; }
+  __off64_t lseek (__off64_t, int) { return 0; }
   select_record *select_read (select_record *s);
-  int ready_for_read (int fd, DWORD howlong);
+  int cygserver_attach_tty (HANDLE*, HANDLE*);
 };
 
 class fhandler_pty_master: public fhandler_tty_common
@@ -856,7 +930,7 @@ class fhandler_pty_master: public fhandler_tty_common
   int accept_input ();
   int open (path_conv *, int flags, mode_t mode = 0);
   int write (const void *ptr, size_t len);
-  int __stdcall read (void *ptr, size_t len) __attribute__ ((regparm (3)));
+  void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
   int close ();
 
   int tcsetattr (int a, const struct termios *t);
@@ -864,7 +938,7 @@ class fhandler_pty_master: public fhandler_tty_common
   int tcflush (int);
   int ioctl (unsigned int cmd, void *);
 
-  __off32_t lseek (__off32_t, int) { return 0; }
+  __off64_t lseek (__off64_t, int) { return 0; }
   char *ptsname ();
 
   void set_close_on_exec (int val);
@@ -876,11 +950,11 @@ class fhandler_tty_master: public fhandler_pty_master
  public:
   /* Constructor */
   fhandler_console *console;	// device handler to perform real i/o.
-  HANDLE hThread;		// process_output thread handle.
 
   fhandler_tty_master (int unit);
   int init (int);
   int init_console ();
+  void set_winsize (bool);
   void fixup_after_fork (HANDLE parent);
   void fixup_after_exec (HANDLE);
 };
@@ -902,9 +976,8 @@ class fhandler_dev_zero: public fhandler_base
   fhandler_dev_zero ();
   int open (path_conv *, int flags, mode_t mode = 0);
   int write (const void *ptr, size_t len);
-  int __stdcall read (void *ptr, size_t len) __attribute__ ((regparm (3)));
-  __off32_t lseek (__off32_t offset, int whence);
-  int close (void);
+  void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
+  __off64_t lseek (__off64_t offset, int whence);
 
   void dump ();
 };
@@ -925,8 +998,8 @@ class fhandler_dev_random: public fhandler_base
   int get_unit () { return unit; }
   int open (path_conv *, int flags, mode_t mode = 0);
   int write (const void *ptr, size_t len);
-  int __stdcall read (void *ptr, size_t len) __attribute__ ((regparm (3)));
-  __off32_t lseek (__off32_t offset, int whence);
+  void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
+  __off64_t lseek (__off64_t offset, int whence);
   int close (void);
   int dup (fhandler_base *child);
 
@@ -938,7 +1011,7 @@ class fhandler_dev_mem: public fhandler_base
  protected:
   int unit;
   DWORD mem_size;
-  DWORD pos;
+  __off64_t pos;
 
  public:
   fhandler_dev_mem (int unit);
@@ -946,13 +1019,13 @@ class fhandler_dev_mem: public fhandler_base
 
   int open (path_conv *, int flags, mode_t mode = 0);
   int write (const void *ptr, size_t ulen);
-  int __stdcall read (void *ptr, size_t len) __attribute__ ((regparm (3)));
-  __off32_t lseek (__off32_t offset, int whence);
+  void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
+  __off64_t lseek (__off64_t offset, int whence);
   int close (void);
-  int __stdcall fstat (struct stat *buf, path_conv *) __attribute__ ((regparm (3)));
+  int __stdcall fstat (struct __stat64 *buf, path_conv *) __attribute__ ((regparm (3)));
   int dup (fhandler_base *child);
 
-  HANDLE mmap (caddr_t *addr, size_t len, DWORD access, int flags, __off32_t off);
+  HANDLE mmap (caddr_t *addr, size_t len, DWORD access, int flags, __off64_t off);
   int munmap (HANDLE h, caddr_t addr, size_t len);
   int msync (HANDLE h, caddr_t addr, size_t len, int flags);
   BOOL fixup_mmap_after_fork (HANDLE h, DWORD access, DWORD offset,
@@ -968,8 +1041,8 @@ class fhandler_dev_clipboard: public fhandler_base
   int is_windows (void) { return 1; }
   int open (path_conv *, int flags, mode_t mode = 0);
   int write (const void *ptr, size_t len);
-  int __stdcall read (void *ptr, size_t len) __attribute__ ((regparm (3)));
-  __off32_t lseek (__off32_t offset, int whence);
+  void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
+  __off64_t lseek (__off64_t offset, int whence);
   int close (void);
 
   int dup (fhandler_base *child);
@@ -977,7 +1050,7 @@ class fhandler_dev_clipboard: public fhandler_base
   void dump ();
 
  private:
-  __off32_t pos;
+  __off64_t pos;
   void *membuffer;
   size_t msize;
   bool eof;
@@ -993,9 +1066,9 @@ class fhandler_windows: public fhandler_base
   int is_windows (void) { return 1; }
   int open (path_conv *, int flags, mode_t mode = 0);
   int write (const void *ptr, size_t len);
-  int __stdcall read (void *ptr, size_t len) __attribute__ ((regparm (3)));
+  void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
   int ioctl (unsigned int cmd, void *);
-  __off32_t lseek (__off32_t, int) { return 0; }
+  __off64_t lseek (__off64_t, int) { return 0; }
   int close (void) { return 0; }
 
   void set_close_on_exec (int val);
@@ -1019,38 +1092,119 @@ class fhandler_dev_dsp : public fhandler_base
 
   int open (path_conv *, int flags, mode_t mode = 0);
   int write (const void *ptr, size_t len);
-  int __stdcall read (void *ptr, size_t len) __attribute__ ((regparm (3)));
+  void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
   int ioctl (unsigned int cmd, void *);
-  __off32_t lseek (__off32_t, int);
+  __off64_t lseek (__off64_t, int);
   int close (void);
   int dup (fhandler_base * child);
   void dump (void);
   void fixup_after_exec (HANDLE);
 };
 
+class fhandler_virtual : public fhandler_base
+{
+ protected:
+  char *filebuf;
+  size_t bufalloc;
+  __off64_t filesize;
+  __off64_t position;
+  int fileid; // unique within each class
+ public:
+
+  fhandler_virtual (DWORD devtype);
+  virtual ~fhandler_virtual();
+
+  virtual int exists();
+  DIR *opendir (path_conv& pc);
+  __off64_t telldir (DIR *);
+  void seekdir (DIR *, __off64_t);
+  void rewinddir (DIR *);
+  int closedir (DIR *);
+  int write (const void *ptr, size_t len);
+  void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
+  __off64_t lseek (__off64_t, int);
+  int dup (fhandler_base * child);
+  int open (path_conv *, int flags, mode_t mode = 0);
+  int close (void);
+  int __stdcall fstat (struct stat *buf, path_conv *pc) __attribute__ ((regparm (3)));
+  virtual bool fill_filebuf ();
+  void fixup_after_exec (HANDLE);
+};
+
+class fhandler_proc: public fhandler_virtual
+{
+ public:
+  fhandler_proc ();
+  fhandler_proc (DWORD devtype);
+  int exists();
+  struct dirent *readdir (DIR *);
+  static DWORD get_proc_fhandler(const char *path);
+
+  int open (path_conv *real_path, int flags, mode_t mode = 0);
+  int __stdcall fstat (struct __stat64 *buf, path_conv *) __attribute__ ((regparm (3)));
+  bool fill_filebuf ();
+};
+
+class fhandler_registry: public fhandler_proc
+{
+ private:
+  char *value_name;
+ public:
+  fhandler_registry ();
+  int exists();
+  struct dirent *readdir (DIR *);
+  __off64_t telldir (DIR *);
+  void seekdir (DIR *, __off64_t);
+  void rewinddir (DIR *);
+  int closedir (DIR *);
+
+  int open (path_conv *real_path, int flags, mode_t mode = 0);
+  int __stdcall fstat (struct __stat64 *buf, path_conv *) __attribute__ ((regparm (3)));
+  bool fill_filebuf ();
+  int close (void);
+};
+
+class pinfo;
+class fhandler_process: public fhandler_proc
+{
+  pid_t pid;
+ public:
+  fhandler_process ();
+  int exists();
+  struct dirent *readdir (DIR *);
+  int open (path_conv *real_path, int flags, mode_t mode = 0);
+  int __stdcall fstat (struct __stat64 *buf, path_conv *) __attribute__ ((regparm (3)));
+  bool fill_filebuf ();
+};
+
 typedef union
 {
-  char base[sizeof(fhandler_base)];
-  char console[sizeof(fhandler_console)];
-  char dev_clipboard[sizeof(fhandler_dev_clipboard)];
-  char dev_dsp[sizeof(fhandler_dev_dsp)];
-  char dev_floppy[sizeof(fhandler_dev_floppy)];
-  char dev_mem[sizeof(fhandler_dev_mem)];
-  char dev_null[sizeof(fhandler_dev_null)];
-  char dev_random[sizeof(fhandler_dev_random)];
-  char dev_raw[sizeof(fhandler_dev_raw)];
-  char dev_tape[sizeof(fhandler_dev_tape)];
-  char dev_zero[sizeof(fhandler_dev_zero)];
-  char disk_file[sizeof(fhandler_disk_file)];
-  char pipe[sizeof(fhandler_pipe)];
-  char pty_master[sizeof(fhandler_pty_master)];
-  char serial[sizeof(fhandler_serial)];
-  char socket[sizeof(fhandler_socket)];
-  char termios[sizeof(fhandler_termios)];
-  char tty_common[sizeof(fhandler_tty_common)];
-  char tty_master[sizeof(fhandler_tty_master)];
-  char tty_slave[sizeof(fhandler_tty_slave)];
-  char windows[sizeof(fhandler_windows)];
+  char __base[sizeof (fhandler_base)];
+  char __console[sizeof (fhandler_console)];
+  char __cygdrive[sizeof (fhandler_cygdrive)];
+  char __dev_clipboard[sizeof (fhandler_dev_clipboard)];
+  char __dev_dsp[sizeof (fhandler_dev_dsp)];
+  char __dev_floppy[sizeof (fhandler_dev_floppy)];
+  char __dev_mem[sizeof (fhandler_dev_mem)];
+  char __dev_null[sizeof (fhandler_dev_null)];
+  char __dev_random[sizeof (fhandler_dev_random)];
+  char __dev_raw[sizeof (fhandler_dev_raw)];
+  char __dev_tape[sizeof (fhandler_dev_tape)];
+  char __dev_zero[sizeof (fhandler_dev_zero)];
+  char __disk_file[sizeof (fhandler_disk_file)];
+  char __pipe[sizeof (fhandler_pipe)];
+  char __proc[sizeof (fhandler_proc)];
+  char __process[sizeof (fhandler_process)];
+  char __pty_master[sizeof (fhandler_pty_master)];
+  char __registry[sizeof (fhandler_registry)];
+  char __serial[sizeof (fhandler_serial)];
+  char __socket[sizeof (fhandler_socket)];
+  char __termios[sizeof (fhandler_termios)];
+  char __tty_common[sizeof (fhandler_tty_common)];
+  char __tty_master[sizeof (fhandler_tty_master)];
+  char __tty_slave[sizeof (fhandler_tty_slave)];
+  char __virtual[sizeof (fhandler_virtual)];
+  char __windows[sizeof (fhandler_windows)];
 } fhandler_union;
 
 struct select_record
@@ -1062,6 +1216,7 @@ struct select_record
   bool windows_handle;
   bool read_ready, write_ready, except_ready;
   bool read_selected, write_selected, except_selected;
+  bool except_on_write;
   int (*startup) (select_record *me, class select_stuff *stuff);
   int (*peek) (select_record *, bool);
   int (*verify) (select_record *me, fd_set *readfds, fd_set *writefds,
@@ -1070,9 +1225,10 @@ struct select_record
   struct select_record *next;
 
   select_record (fhandler_base *in_fh = NULL) : fd (0), h (NULL),
-		 fh (in_fh), saw_error (0), windows_handle (0),
-		 read_ready (0), write_ready (0), except_ready (0),
-		 read_selected (0), write_selected (0), except_selected (0),
+		 fh (in_fh), saw_error (false), windows_handle (false),
+		 read_ready (false), write_ready (false), except_ready (false),
+		 read_selected (false), write_selected (false),
+		 except_selected (false), except_on_write (false),
 		 startup (NULL), peek (NULL), verify (NULL), cleanup (NULL),
 		 next (NULL) {}
 };
