@@ -1,6 +1,6 @@
 /* fhandler_tty.cc
 
-   Copyright 1997, 1998, 2000, 2001 Red Hat, Inc.
+   Copyright 1997, 1998, 2000, 2001, 2002, 2003 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -10,7 +10,6 @@ details. */
 
 #include "winsup.h"
 #include <stdio.h>
-#include <fcntl.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <errno.h>
@@ -21,11 +20,12 @@ details. */
 #include "fhandler.h"
 #include "path.h"
 #include "dtable.h"
-#include "sync.h"
 #include "sigproc.h"
 #include "pinfo.h"
 #include "cygheap.h"
 #include "shared_info.h"
+#include "cygwin/cygserver.h"
+#include "cygthread.h"
 
 /* Tty master stuff */
 
@@ -36,14 +36,23 @@ static DWORD WINAPI process_output (void *);		// Output queue thread
 static DWORD WINAPI process_ioctl (void *);		// Ioctl requests thread
 
 fhandler_tty_master::fhandler_tty_master (int unit)
-  : fhandler_pty_master (FH_TTYM, unit), console (NULL), hThread (NULL)
+  : fhandler_pty_master (FH_TTYM, unit), console (NULL)
 {
 }
 
+void
+fhandler_tty_master::set_winsize (bool sendSIGWINCH)
+{
+  winsize w;
+  console->ioctl (TIOCGWINSZ, &w);
+  get_ttyp ()->winsize = w;
+  if (sendSIGWINCH)
+    tc->kill_pgrp (SIGWINCH);
+}
+
 int
 fhandler_tty_master::init (int ntty)
 {
-  HANDLE h;
   termios_printf ("Creating master for tty%d", ntty);
 
   if (init_console ())
@@ -60,40 +69,22 @@ fhandler_tty_master::init (int ntty)
 
   cygwin_shared->tty[ttynum]->common_init (this);
 
+  set_winsize (false);
+
   inuse = get_ttyp ()->create_inuse (TTY_MASTER_ALIVE);
 
-  h = makethread (process_input, NULL, 0, "ttyin");
-  if (h == NULL)
-    {
-      termios_printf ("can't create input thread");
-      return -1;
-    }
-  else
-    {
-      SetThreadPriority (h, THREAD_PRIORITY_HIGHEST);
-      CloseHandle (h);
-    }
+  cygthread *h;
+  h = new cygthread (process_input, cygself, "ttyin");
+  h->SetThreadPriority (THREAD_PRIORITY_HIGHEST);
+  h->zap_h ();
 
-  h = makethread (process_ioctl, NULL, 0, "ttyioctl");
-  if (h == NULL)
-    {
-      termios_printf ("can't create ioctl thread");
-      return -1;
-    }
-  else
-    {
-      SetThreadPriority (h, THREAD_PRIORITY_HIGHEST);
-      CloseHandle (h);
-    }
+  h = new cygthread (process_ioctl, cygself, "ttyioctl");
+  h->SetThreadPriority (THREAD_PRIORITY_HIGHEST);
+  h->zap_h ();
 
-  hThread = makethread (process_output, NULL, 0, "ttyout");
-  if (hThread != NULL)
-    SetThreadPriority (hThread, THREAD_PRIORITY_HIGHEST);
-  else
-    {
-      termios_printf ("can't create output thread");
-      return -1;
-    }
+  h = new cygthread (process_output, cygself, "ttyout");
+  h->SetThreadPriority (THREAD_PRIORITY_HIGHEST);
+  h->zap_h ();
 
   return 0;
 }
@@ -125,7 +116,7 @@ fhandler_tty_common::__acquire_output_mutex (const char *fn, int ln,
 #else
       ostack[osi].fn = fn;
       ostack[osi].ln = ln;
-      ostack[osi].tname = threadname (0, 0);
+      ostack[osi].tname = cygthread::name ();
       termios_printf ("acquired for %s:%d, osi %d", fn, ln, osi);
       osi++;
 #endif
@@ -166,46 +157,49 @@ fhandler_pty_master::doecho (const void *str, DWORD len)
 int
 fhandler_pty_master::accept_input ()
 {
-  DWORD bytes_left, written;
-  DWORD n;
-  DWORD rc;
-  char* p;
+  DWORD bytes_left;
+  int ret = 1;
 
-  rc = WaitForSingleObject (input_mutex, INFINITE);
+  (void) WaitForSingleObject (input_mutex, INFINITE);
 
-  bytes_left = n = eat_readahead (-1);
-  get_ttyp ()->read_retval = 0;
-  p = rabuf;
+  bytes_left = eat_readahead (-1);
 
-  if (n != 0)
+  if (!bytes_left)
     {
-      while (bytes_left > 0)
+      termios_printf ("sending EOF to slave");
+      get_ttyp ()->read_retval = 0;
+    }
+  else
+    {
+      char *p = rabuf;
+      DWORD rc;
+      DWORD written = 0;
+
+      termios_printf ("about to write %d chars to slave", bytes_left);
+      rc = WriteFile (get_output_handle (), p, bytes_left, &written, NULL);
+      if (!rc)
 	{
-	  termios_printf ("about to write %d chars to slave", bytes_left);
-	  rc = WriteFile (get_output_handle (), p, bytes_left, &written, NULL);
-	  if (!rc)
-	    {
-	      debug_printf ("error writing to pipe %E");
-	      break;
-	    }
-	  get_ttyp ()->read_retval += written;
+	  debug_printf ("error writing to pipe %E");
+	  get_ttyp ()->read_retval = -1;
+	  ret = -1;
+	}
+      else
+	{
+	  get_ttyp ()->read_retval = 1;
 	  p += written;
 	  bytes_left -= written;
 	  if (bytes_left > 0)
 	    {
 	      debug_printf ("to_slave pipe is full");
-	      SetEvent (input_available_event);
-	      ReleaseMutex (input_mutex);
-	      Sleep (10);
-	      rc = WaitForSingleObject (input_mutex, INFINITE);
+	      puts_readahead (p, bytes_left);
+	      ret = 0;
 	    }
 	}
     }
-  else
-    termios_printf ("sending EOF to slave");
+
   SetEvent (input_available_event);
   ReleaseMutex (input_mutex);
-  return get_ttyp ()->read_retval;
+  return ret;
 }
 
 static DWORD WINAPI
@@ -215,9 +209,11 @@ process_input (void *)
 
   while (1)
     {
-      int nraw = tty_master->console->read ((void *) rawbuf,
-				      (size_t) INP_BUFFER_SIZE);
-      (void) tty_master->line_edit (rawbuf, nraw);
+      size_t nraw = INP_BUFFER_SIZE;
+      tty_master->console->read ((void *) rawbuf, nraw);
+      if (tty_master->line_edit (rawbuf, nraw, tty_master->get_ttyp ()->ti)
+	  == line_edit_signalled)
+	tty_master->console->eat_readahead (-1);
     }
 }
 
@@ -394,19 +390,15 @@ out:
 static DWORD WINAPI
 process_output (void *)
 {
-  char buf[OUT_BUFFER_SIZE*2];
+  char buf[OUT_BUFFER_SIZE * 2];
 
   for (;;)
     {
       int n = tty_master->process_slave_output (buf, OUT_BUFFER_SIZE, 0);
-      if (n < 0)
+      if (n <= 0)
 	{
-	  termios_printf ("ReadFile %E");
-	  ExitThread (0);
-	}
-      if (n == 0)
-	{
-	  /* End of file.  */
+	  if (n < 0)
+	    termios_printf ("ReadFile %E");
 	  ExitThread (0);
 	}
       n = tty_master->console->write ((void *) buf, (size_t) n);
@@ -437,11 +429,13 @@ process_ioctl (void *)
 fhandler_tty_slave::fhandler_tty_slave (int num)
   : fhandler_tty_common (FH_TTYS, num)
 {
+  set_r_no_interrupt (1);
 }
 
 fhandler_tty_slave::fhandler_tty_slave ()
   : fhandler_tty_common (FH_TTYS, 0)
 {
+  set_r_no_interrupt (1);
 }
 
 /* FIXME: This function needs to close handles when it has
@@ -454,7 +448,7 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
   attach_tty (ttynum);
   tc->set_ctty (ttynum, flags);
 
-  set_flags (flags);
+  set_flags ((flags & ~O_TEXT) | O_BINARY);
   /* Create synchronisation events */
   char buf[40];
 
@@ -510,45 +504,94 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
       return 0;
     }
 
-  HANDLE tty_owner = OpenProcess (PROCESS_DUP_HANDLE, FALSE,
-				     get_ttyp ()->master_pid);
-  if (tty_owner == NULL)
+  HANDLE from_master_local, to_master_local;
+
+  if (!wincap.has_security () ||
+      cygserver_running == CYGSERVER_UNAVAIL ||
+      !cygserver_attach_tty (&from_master_local, &to_master_local))
     {
-      termios_printf ("can't open tty (%d) handle process %d",
-		      ttynum, get_ttyp ()->master_pid);
-      __seterrno ();
-      return 0;
+      termios_printf ("cannot dup handles via server. using old method.");
+
+      HANDLE tty_owner = OpenProcess (PROCESS_DUP_HANDLE, FALSE,
+				      get_ttyp ()->master_pid);
+      termios_printf ("tty own handle %p",tty_owner);
+      if (tty_owner == NULL)
+	{
+	  termios_printf ("can't open tty (%d) handle process %d",
+			  ttynum, get_ttyp ()->master_pid);
+	  __seterrno ();
+	  return 0;
+	}
+
+      if (!DuplicateHandle (tty_owner, get_ttyp ()->from_master,
+			    hMainProc, &from_master_local, 0, TRUE,
+			    DUPLICATE_SAME_ACCESS))
+	{
+	  termios_printf ("can't duplicate input, %E");
+	  __seterrno ();
+	  return 0;
+	}
+
+      if (!DuplicateHandle (tty_owner, get_ttyp ()->to_master,
+			  hMainProc, &to_master_local, 0, TRUE,
+			  DUPLICATE_SAME_ACCESS))
+	{
+	  termios_printf ("can't duplicate output, %E");
+	  __seterrno ();
+	  return 0;
+	}
+      CloseHandle (tty_owner);
     }
 
-  HANDLE nh;
-  if (!DuplicateHandle (tty_owner, get_ttyp ()->from_master, hMainProc, &nh, 0, TRUE,
-			DUPLICATE_SAME_ACCESS))
-    {
-      termios_printf ("can't duplicate input, %E");
-      __seterrno ();
-      return 0;
-    }
-  set_io_handle (nh);
-  ProtectHandle1 (nh, from_pty);
-  termios_printf ("duplicated from_master %p->%p from tty_owner %p",
-		  get_ttyp ()->from_master, nh, tty_owner);
-  if (!DuplicateHandle (tty_owner, get_ttyp ()->to_master, hMainProc, &nh, 0, TRUE,
-			DUPLICATE_SAME_ACCESS))
-    {
-      termios_printf ("can't duplicate output, %E");
-      __seterrno ();
-      return 0;
-    }
-  set_output_handle (nh);
-  ProtectHandle1 (nh, to_pty);
-  CloseHandle (tty_owner);
+  termios_printf ("duplicated from_master %p->%p from tty_owner",
+      get_ttyp ()->from_master, from_master_local);
+  termios_printf ("duplicated to_master %p->%p from tty_owner",
+      get_ttyp ()->to_master, to_master_local);
+
+  set_io_handle (from_master_local);
+  set_output_handle (to_master_local);
 
   set_open_status ();
+  if (!output_done_event)
+    {
+      fhandler_console::open_fhs++;
+      termios_printf ("incremented open_fhs %d", fhandler_console::open_fhs);
+    }
   termios_printf ("tty%d opened", ttynum);
 
   return 1;
 }
 
+int
+fhandler_tty_slave::close ()
+{
+  if (!output_done_event && !cygheap->fdtab.in_vfork_cleanup ())
+    {
+      fhandler_console::open_fhs--;
+      termios_printf ("decremented open_fhs %d", fhandler_console::open_fhs);
+    }
+  return fhandler_tty_common::close ();
+}
+
+int
+fhandler_tty_slave::cygserver_attach_tty (LPHANDLE from_master_ptr,
+					  LPHANDLE to_master_ptr)
+{
+  if (!from_master_ptr || !to_master_ptr)
+    return 0;
+
+  client_request_attach_tty req ((DWORD) get_ttyp ()->master_pid,
+				 (HANDLE) get_ttyp ()->from_master,
+				 (HANDLE) get_ttyp ()->to_master);
+
+  if (req.make_request () == -1 || req.error_code ())
+    return 0;
+
+  *from_master_ptr = req.from_master ();
+  *to_master_ptr = req.to_master ();
+  return 1;
+}
+
 void
 fhandler_tty_slave::init (HANDLE, DWORD a, mode_t)
 {
@@ -602,7 +645,7 @@ fhandler_tty_slave::write (const void *ptr, size_t len)
 	    default:
 	      __seterrno_from_win_error (err);
 	    }
-	  _raise (SIGHUP);		/* FIXME: Should this be SIGTTOU? */
+	  raise (SIGHUP);		/* FIXME: Should this be SIGTTOU? */
 	  towrite = (DWORD) -1;
 	  break;
 	}
@@ -619,12 +662,11 @@ fhandler_tty_slave::write (const void *ptr, size_t len)
   return towrite;
 }
 
-int __stdcall
-fhandler_tty_slave::read (void *ptr, size_t len)
+void __stdcall
+fhandler_tty_slave::read (void *ptr, size_t& len)
 {
-  DWORD n;
   int totalread = 0;
-  int vmin = INT_MAX;
+  int vmin = 0;
   int vtime = 0;	/* Initialized to prevent -Wuninitialized warning */
   size_t readlen;
   DWORD bytes_in_pipe;
@@ -634,28 +676,43 @@ fhandler_tty_slave::read (void *ptr, size_t len)
   DWORD rc;
   HANDLE w4[2];
 
-  termios_printf ("read(%x, %d) handle %d", ptr, len, get_handle ());
+  termios_printf ("read(%x, %d) handle %p", ptr, len, get_handle ());
 
-  if (!(get_ttyp ()->ti.c_lflag & ICANON))
-    {
-      vmin = min (INP_BUFFER_SIZE, get_ttyp ()->ti.c_cc[VMIN]);
-      vtime = get_ttyp ()->ti.c_cc[VTIME];
-      if (vmin < 0) vmin = 0;
-      if (vtime < 0) vtime = 0;
-      if (vmin == 0)
-	time_to_wait = INFINITE;
-      else
-	time_to_wait = (vtime == 0 ? INFINITE : 100 * vtime);
-    }
-  else
+  if ((get_ttyp ()->ti.c_lflag & ICANON))
     time_to_wait = INFINITE;
+  else
+    {
+      vmin = get_ttyp ()->ti.c_cc[VMIN];
+      if (vmin > INP_BUFFER_SIZE)
+	vmin = INP_BUFFER_SIZE;
+      vtime = get_ttyp ()->ti.c_cc[VTIME];
+      if (vmin < 0)
+	vmin = 0;
+      if (vtime < 0)
+	vtime = 0;
+      if (!vmin && !vtime)
+	time_to_wait = 0;
+      else
+	time_to_wait = !vtime ? INFINITE : 100 * vtime;
+    }
 
   w4[0] = signal_arrived;
   w4[1] = input_available_event;
 
+  DWORD waiter = INFINITE;
   while (len)
     {
-      rc = WaitForMultipleObjects (2, w4, FALSE, time_to_wait);
+      rc = WaitForMultipleObjects (2, w4, FALSE, waiter);
+
+      if (rc == WAIT_TIMEOUT)
+	break;
+
+      if (rc == WAIT_FAILED)
+	{
+	  termios_printf ("wait for input event failed, %E");
+	  break;
+	}
+
       if (rc == WAIT_OBJECT_0)
 	{
 	  /* if we've received signal after successfully reading some data,
@@ -663,15 +720,10 @@ fhandler_tty_slave::read (void *ptr, size_t len)
 	  if (totalread > 0)
 	    break;
 	  set_sig_errno (EINTR);
-	  return -1;
+	  (ssize_t) len = -1;
+	  return;
 	}
-      else if (rc == WAIT_FAILED)
-	{
-	  termios_printf ("wait for input event failed, %E");
-	  break;
-	}
-      else if (rc == WAIT_TIMEOUT)
-	break;
+
       rc = WaitForSingleObject (input_mutex, 1000);
       if (rc == WAIT_FAILED)
 	{
@@ -683,20 +735,33 @@ fhandler_tty_slave::read (void *ptr, size_t len)
 	  termios_printf ("failed to acquire input mutex after input event arrived");
 	  break;
 	}
-      if (!PeekNamedPipe (get_handle (), peek_buf, sizeof(peek_buf), &bytes_in_pipe, NULL, NULL))
+      if (!PeekNamedPipe (get_handle (), peek_buf, sizeof (peek_buf), &bytes_in_pipe, NULL, NULL))
 	{
 	  termios_printf ("PeekNamedPipe failed, %E");
-	  _raise (SIGHUP);
+	  raise (SIGHUP);
 	  bytes_in_pipe = 0;
 	}
+
+      if (!vmin && !time_to_wait)
+	{
+	  ReleaseMutex (input_mutex);
+	  (ssize_t) len = bytes_in_pipe;
+	  return;
+	}
+
       readlen = min (bytes_in_pipe, min (len, sizeof (buf)));
+
+      if (vmin && readlen > (unsigned) vmin)
+	readlen = vmin;
+
+      DWORD n = 0;
       if (readlen)
 	{
 	  termios_printf ("reading %d bytes (vtime %d)", readlen, vtime);
 	  if (ReadFile (get_handle (), buf, readlen, &n, NULL) == FALSE)
 	    {
 	      termios_printf ("read failed, %E");
-	      _raise (SIGHUP);
+	      raise (SIGHUP);
 	    }
 	  /* MSDN states that 5th prameter can be used to determine total
 	     number of bytes in pipe, but for some reason this number doesn't
@@ -705,7 +770,7 @@ fhandler_tty_slave::read (void *ptr, size_t len)
 	  if (!PeekNamedPipe (get_handle (), peek_buf, 1, &bytes_in_pipe, NULL, NULL))
 	    {
 	      termios_printf ("PeekNamedPipe failed, %E");
-	      _raise (SIGHUP);
+	      raise (SIGHUP);
 	      bytes_in_pipe = 0;
 	    }
 	  if (n)
@@ -735,7 +800,7 @@ fhandler_tty_slave::read (void *ptr, size_t len)
 	}
       if (get_ttyp ()->ti.c_lflag & ICANON || is_nonblocking ())
 	break;
-      if (totalread >= vmin && (vmin > 0 || totalread > 0))
+      if (vmin && totalread >= vmin)
 	break;
 
       /* vmin == 0 && vtime == 0:
@@ -755,9 +820,24 @@ fhandler_tty_slave::read (void *ptr, size_t len)
 
       if (vmin == 0)
 	break;
+
+      if (n)
+	waiter = time_to_wait;
     }
   termios_printf ("%d=read(%x, %d)", totalread, ptr, len);
-  return totalread;
+  (ssize_t) len = totalread;
+  return;
+}
+
+int
+fhandler_tty_slave::dup (fhandler_base *child)
+{
+  if (!output_done_event)
+    {
+      fhandler_console::open_fhs++;
+      termios_printf ("incremented open_fhs %d", fhandler_console::open_fhs);
+    }
+  return fhandler_tty_common::dup (child);
 }
 
 int
@@ -884,13 +964,13 @@ fhandler_tty_slave::ioctl (unsigned int cmd, void *arg)
 {
   termios_printf ("ioctl (%x)", cmd);
 
-  if (myself->pgid && get_ttyp ()->getpgid () != myself->pgid &&
-	 myself->ctty == ttynum && (get_ttyp ()->ti.c_lflag & TOSTOP))
+  if (myself->pgid && get_ttyp ()->getpgid () != myself->pgid
+      && myself->ctty == ttynum && (get_ttyp ()->ti.c_lflag & TOSTOP))
     {
       /* background process */
       termios_printf ("bg ioctl pgid %d, tpgid %d, ctty %d",
 		      myself->pgid, get_ttyp ()->getpgid (), myself->ctty);
-      _raise (SIGTTOU);
+      raise (SIGTTOU);
     }
 
   switch (cmd)
@@ -916,18 +996,29 @@ fhandler_tty_slave::ioctl (unsigned int cmd, void *arg)
       get_ttyp ()->arg.winsize = get_ttyp ()->winsize;
       if (ioctl_request_event)
 	SetEvent (ioctl_request_event);
-      * (struct winsize *) arg = get_ttyp ()->arg.winsize;
+      *(struct winsize *) arg = get_ttyp ()->arg.winsize;
       if (ioctl_done_event)
 	WaitForSingleObject (ioctl_done_event, INFINITE);
       get_ttyp ()->winsize = get_ttyp ()->arg.winsize;
       break;
     case TIOCSWINSZ:
-      get_ttyp ()->ioctl_retval = -1;
-      get_ttyp ()->arg.winsize = * (struct winsize *) arg;
-      if (ioctl_request_event)
-	SetEvent (ioctl_request_event);
-      if (ioctl_done_event)
-	WaitForSingleObject (ioctl_done_event, INFINITE);
+      if (get_ttyp ()->winsize.ws_row != ((struct winsize *) arg)->ws_row
+	  || get_ttyp ()->winsize.ws_col != ((struct winsize *) arg)->ws_col)
+	{
+	  get_ttyp ()->arg.winsize = *(struct winsize *) arg;
+	  if (ioctl_request_event)
+	    {
+	      get_ttyp ()->ioctl_retval = -1;
+	      SetEvent (ioctl_request_event);
+	    }
+	  else
+	    {
+	      get_ttyp ()->winsize = *(struct winsize *) arg;
+	      kill (-get_ttyp ()->getpgid (), SIGWINCH);
+	    }
+	  if (ioctl_done_event)
+	    WaitForSingleObject (ioctl_done_event, INFINITE);
+	}
       break;
     }
 
@@ -955,7 +1046,7 @@ fhandler_pty_master::open (path_conv *, int flags, mode_t)
 
   cygwin_shared->tty[ttynum]->common_init (this);
   inuse = get_ttyp ()->create_inuse (TTY_MASTER_ALIVE);
-  set_flags (flags);
+  set_flags ((flags & ~O_TEXT) | O_BINARY);
   set_open_status ();
 
   termios_printf ("opened pty master tty%d<%p>", ttynum, this);
@@ -1004,7 +1095,7 @@ fhandler_pty_master::close ()
   while (accept_input () > 0)
     continue;
 #endif
-  this->fhandler_tty_common::close ();
+  fhandler_tty_common::close ();
 
   if (!get_ttyp ()->master_alive ())
     {
@@ -1028,14 +1119,28 @@ fhandler_pty_master::close ()
 int
 fhandler_pty_master::write (const void *ptr, size_t len)
 {
-  (void) line_edit ((char *) ptr, len);
-  return len;
+  int i;
+  char *p = (char *) ptr;
+  termios ti = tc->ti;
+
+  for (i = 0; i < (int) len; i++)
+    {
+      line_edit_status status = line_edit (p++, 1, ti);
+      if (status > line_edit_signalled)
+	{
+	  if (status != line_edit_pipe_full)
+	    i = -1;
+	  break;
+	}
+    }
+  return i;
 }
 
-int __stdcall
-fhandler_pty_master::read (void *ptr, size_t len)
+void __stdcall
+fhandler_pty_master::read (void *ptr, size_t& len)
 {
-  return process_slave_output ((char *) ptr, len, pktmode);
+  (ssize_t) len = process_slave_output ((char *) ptr, len, pktmode);
+  return;
 }
 
 int
@@ -1067,11 +1172,15 @@ fhandler_pty_master::ioctl (unsigned int cmd, void *arg)
 	pktmode = * (int *) arg;
 	break;
       case TIOCGWINSZ:
-	* (struct winsize *) arg = get_ttyp ()->winsize;
+	*(struct winsize *) arg = get_ttyp ()->winsize;
 	break;
       case TIOCSWINSZ:
-	get_ttyp ()->winsize = * (struct winsize *) arg;
-	_kill (-get_ttyp ()->getpgid (), SIGWINCH);
+	if (get_ttyp ()->winsize.ws_row != ((struct winsize *) arg)->ws_row
+	    || get_ttyp ()->winsize.ws_col != ((struct winsize *) arg)->ws_col)
+	  {
+	    get_ttyp ()->winsize = * (struct winsize *) arg;
+	    kill (-get_ttyp ()->getpgid (), SIGWINCH);
+	  }
 	break;
       case FIONBIO:
 	set_nonblocking (*(int *) arg);
@@ -1096,7 +1205,7 @@ void
 fhandler_tty_common::set_close_on_exec (int val)
 {
 #ifndef DEBUGGING
-  this->fhandler_base::set_close_on_exec (val);
+  fhandler_base::set_close_on_exec (val);
 #else
   /* FIXME: This is a duplication from fhandler_base::set_close_on_exec.
      It is here because we need to specify the "from_pty" stuff here or
@@ -1118,10 +1227,21 @@ fhandler_tty_common::set_close_on_exec (int val)
   set_inheritance (output_handle, val);
 }
 
+void
+fhandler_tty_slave::fixup_after_fork (HANDLE parent)
+{
+  if (!output_done_event)
+    {
+      fhandler_console::open_fhs++;
+      termios_printf ("incremented open_fhs %d", fhandler_console::open_fhs);
+    }
+  fhandler_tty_common::fixup_after_fork (parent);
+}
+
 void
 fhandler_tty_common::fixup_after_fork (HANDLE parent)
 {
-  this->fhandler_termios::fixup_after_fork (parent);
+  fhandler_termios::fixup_after_fork (parent);
   if (output_done_event)
     fork_fixup (parent, output_done_event, "output_done_event");
   if (ioctl_request_event)
@@ -1140,7 +1260,7 @@ fhandler_tty_common::fixup_after_fork (HANDLE parent)
 void
 fhandler_pty_master::set_close_on_exec (int val)
 {
-  this->fhandler_tty_common::set_close_on_exec (val);
+  fhandler_tty_common::set_close_on_exec (val);
 
   /* FIXME: There is a console handle leak here. */
   if (get_ttyp ()->master_pid == GetCurrentProcessId ())
@@ -1155,7 +1275,7 @@ fhandler_pty_master::set_close_on_exec (int val)
 void
 fhandler_tty_master::fixup_after_fork (HANDLE child)
 {
-  this->fhandler_pty_master::fixup_after_fork (child);
+  fhandler_pty_master::fixup_after_fork (child);
   console->fixup_after_fork (child);
 }
 
@@ -1164,7 +1284,6 @@ fhandler_tty_master::fixup_after_exec (HANDLE)
 {
   console->close ();
   init_console ();
-  return;
 }
 
 int
