@@ -23,7 +23,8 @@ fhandler_dev_zero::fhandler_dev_zero ()
 int
 fhandler_dev_zero::open (path_conv *, int flags, mode_t)
 {
-  set_flags (flags);
+  set_flags ((flags & ~O_TEXT) | O_BINARY);
+  set_nohandle (true);
   set_open_status ();
   return 1;
 }
@@ -34,21 +35,15 @@ fhandler_dev_zero::write (const void *, size_t len)
   return len;
 }
 
-int __stdcall
-fhandler_dev_zero::read (void *ptr, size_t len)
+void __stdcall
+fhandler_dev_zero::read (void *ptr, size_t& len)
 {
-  memset(ptr, 0, len);
-  return len;
+  memset (ptr, 0, len);
+  return;
 }
 
-__off32_t
-fhandler_dev_zero::lseek (__off32_t, int)
-{
-  return 0;
-}
-
-int
-fhandler_dev_zero::close (void)
+__off64_t
+fhandler_dev_zero::lseek (__off64_t, int)
 {
   return 0;
 }
@@ -56,5 +51,5 @@ fhandler_dev_zero::close (void)
 void
 fhandler_dev_zero::dump ()
 {
-  paranoid_printf("here, fhandler_dev_zero");
+  paranoid_printf ("here, fhandler_dev_zero");
 }
