@@ -1,6 +1,6 @@
 /* mmap.cc
 
-   Copyright 1996, 1997, 1998, 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 2000, 2001, 2002, 2003 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -20,15 +20,13 @@ details. */
 #include "dtable.h"
 #include "cygerrno.h"
 #include "cygheap.h"
-#include "sync.h"
-#include "sigproc.h"
 #include "pinfo.h"
 #include "sys/cygwin.h"
 
-#define PAGE_CNT(bytes) howmany(bytes,getpagesize())
+#define PAGE_CNT(bytes) howmany((bytes),getpagesize())
 
-#define PGBITS		(sizeof(DWORD)*8)
-#define MAPSIZE(pages)	howmany(pages,PGBITS)
+#define PGBITS		(sizeof (DWORD)*8)
+#define MAPSIZE(pages)	howmany ((pages), PGBITS)
 
 #define MAP_SET(n)	(map_map_[(n)/PGBITS] |= (1L << ((n) % PGBITS)))
 #define MAP_CLR(n)	(map_map_[(n)/PGBITS] &= ~(1L << ((n) % PGBITS)))
@@ -47,13 +45,13 @@ class mmap_record
     HANDLE mapping_handle_;
     int devtype_;
     DWORD access_mode_;
-    DWORD offset_;
+    __off64_t offset_;
     DWORD size_to_map_;
     caddr_t base_address_;
     DWORD *map_map_;
 
   public:
-    mmap_record (int fd, HANDLE h, DWORD ac, DWORD o, DWORD s, caddr_t b) :
+    mmap_record (int fd, HANDLE h, DWORD ac, __off64_t o, DWORD s, caddr_t b) :
        fdesc_ (fd),
        mapping_handle_ (h),
        devtype_ (0),
@@ -78,26 +76,14 @@ class mmap_record
     DWORD get_size () const { return size_to_map_; }
     caddr_t get_address () const { return base_address_; }
     DWORD *get_map () const { return map_map_; }
-
-    void alloc_map ()
-      {
-	/* Allocate one bit per page */
-	map_map_ = (DWORD *) calloc (MAPSIZE(PAGE_CNT (size_to_map_)),
-				     sizeof (DWORD));
-	if (wincap.virtual_protect_works_on_shared_pages ())
-	  {
-	    DWORD old_prot;
-	    if (!VirtualProtect (base_address_, size_to_map_,
-				 PAGE_NOACCESS, &old_prot))
-	      syscall_printf ("-1 = alloc_map (): %E");
-	  }
-      }
+    void alloc_map (__off64_t off, DWORD len);
     void free_map () { if (map_map_) free (map_map_); }
 
     DWORD find_empty (DWORD pages);
-    DWORD map_map (DWORD off, DWORD len);
+    __off64_t map_map (__off64_t off, DWORD len);
     BOOL unmap_map (caddr_t addr, DWORD len);
     void fixup_map (void);
+    int access (caddr_t address);
 
     fhandler_base *alloc_fh ();
     void free_fh (fhandler_base *fh);
@@ -124,9 +110,41 @@ mmap_record::find_empty (DWORD pages)
   return (DWORD)-1;
 }
 
-DWORD
-mmap_record::map_map (DWORD off, DWORD len)
+void
+mmap_record::alloc_map (__off64_t off, DWORD len)
 {
+  /* Allocate one bit per page */
+  map_map_ = (DWORD *) calloc (MAPSIZE (PAGE_CNT (size_to_map_)),
+			       sizeof (DWORD));
+  if (wincap.virtual_protect_works_on_shared_pages ())
+    {
+      DWORD old_prot;
+
+      off -= offset_;
+      len = PAGE_CNT (len) * getpagesize ();
+      if (off > 0 &&
+	  !VirtualProtect (base_address_, off, PAGE_NOACCESS, &old_prot))
+	syscall_printf ("VirtualProtect(%x,%d) failed: %E", base_address_, off);
+      if (off + len < size_to_map_
+	  && !VirtualProtect (base_address_ + off + len,
+			      size_to_map_ - len - off,
+			      PAGE_NOACCESS, &old_prot))
+	syscall_printf ("VirtualProtect(%x,%d) failed: %E",
+			base_address_ + off + len, size_to_map_ - len - off);
+      off /= getpagesize ();
+      len /= getpagesize ();
+      while (len-- > 0)
+	MAP_SET (off + len);
+    }
+}
+
+__off64_t
+mmap_record::map_map (__off64_t off, DWORD len)
+{
+  /* Used ONLY if this mapping matches into the chunk of another already
+     performed mapping in a special case of MAP_ANON|MAP_PRIVATE.
+
+     Otherwise it's job is now done by alloc_map(). */
   DWORD prot, old_prot;
   switch (access_mode_)
     {
@@ -141,33 +159,22 @@ mmap_record::map_map (DWORD off, DWORD len)
       break;
     }
 
+  debug_printf ("map_map (fd=%d, off=%D, len=%d)", fdesc_, off, len);
   len = PAGE_CNT (len);
-  if (fdesc_ == -1 && !off)
-    {
-      off = find_empty (len);
-      if (off != (DWORD)-1)
-	{
-	  if (wincap.virtual_protect_works_on_shared_pages ()
-	      && !VirtualProtect (base_address_ + off * getpagesize (),
-				  len * getpagesize (), prot, &old_prot))
-	    syscall_printf ("-1 = map_map (): %E");
 
-	  while (len-- > 0)
-	    MAP_SET (off + len);
-	  return off * getpagesize ();
-	}
-      return 0L;
-    }
-  off -= offset_;
-  DWORD start = off / getpagesize ();
+  if ((off = find_empty (len)) == (DWORD)-1)
+    return 0L;
   if (wincap.virtual_protect_works_on_shared_pages ()
-      && !VirtualProtect (base_address_ + start * getpagesize (),
+      && !VirtualProtect (base_address_ + off * getpagesize (),
 			  len * getpagesize (), prot, &old_prot))
-    syscall_printf ("-1 = map_map (): %E");
+    {
+      __seterrno ();
+      return (__off64_t)-1;
+    }
 
-  for (; len-- > 0; ++start)
-    MAP_SET (start);
-  return off;
+  while (len-- > 0)
+    MAP_SET (off + len);
+  return off * getpagesize ();
 }
 
 BOOL
@@ -186,7 +193,7 @@ mmap_record::unmap_map (caddr_t addr, DWORD len)
     MAP_CLR (off);
   /* Return TRUE if all pages are free'd which may result in unmapping
      the whole chunk. */
-  for (len = MAPSIZE(PAGE_CNT (size_to_map_)); len > 0; )
+  for (len = MAPSIZE (PAGE_CNT (size_to_map_)); len > 0; )
     if (map_map_[--len])
       return FALSE;
   return TRUE;
@@ -213,10 +220,17 @@ mmap_record::fixup_map ()
     }
 
   for (DWORD off = PAGE_CNT (size_to_map_); off > 0; --off)
-    VirtualProtect (base_address_ + off * getpagesize (),
-		    getpagesize (),
-		    MAP_ISSET (off - 1) ? prot : PAGE_NOACCESS,
-		    &old_prot);
+    VirtualProtect (base_address_ + off * getpagesize (), getpagesize (),
+		    MAP_ISSET (off - 1) ? prot : PAGE_NOACCESS, &old_prot);
+}
+
+int
+mmap_record::access (caddr_t address)
+{
+  if (address < base_address_ || address >= base_address_ + size_to_map_)
+    return 0;
+  DWORD off = (address - base_address_) / getpagesize ();
+  return MAP_ISSET (off);
 }
 
 static fhandler_disk_file fh_paging_file;
@@ -235,7 +249,7 @@ mmap_record::alloc_fh ()
      the call to fork(). This requires creating a fhandler
      of the correct type to be sure to call the method of the
      correct class. */
-  return cygheap->fdtab.build_fhandler (-1, get_device (), NULL);
+  return cygheap->fdtab.build_fhandler (-1, get_device ());
 }
 
 void
@@ -253,16 +267,17 @@ public:
   DWORD hash;
   list ();
   ~list ();
-  mmap_record *add_record (mmap_record r);
+  mmap_record *add_record (mmap_record r, __off64_t off, DWORD len);
   void erase (int i);
-  mmap_record *match (DWORD off, DWORD len);
-  __off32_t match (caddr_t addr, DWORD len, __off32_t start);
+  void erase ();
+  mmap_record *match (__off64_t off, DWORD len);
+  long match (caddr_t addr, DWORD len, long start);
 };
 
 list::list ()
 : nrecs (0), maxrecs (10), fd (0), hash (0)
 {
-  recs = (mmap_record *) malloc (10 * sizeof(mmap_record));
+  recs = (mmap_record *) malloc (10 * sizeof (mmap_record));
 }
 
 list::~list ()
@@ -273,7 +288,7 @@ list::~list ()
 }
 
 mmap_record *
-list::add_record (mmap_record r)
+list::add_record (mmap_record r, __off64_t off, DWORD len)
 {
   if (nrecs == maxrecs)
     {
@@ -281,13 +296,13 @@ list::add_record (mmap_record r)
       recs = (mmap_record *) realloc (recs, maxrecs * sizeof (mmap_record));
     }
   recs[nrecs] = r;
-  recs[nrecs].alloc_map ();
+  recs[nrecs].alloc_map (off, len);
   return recs + nrecs++;
 }
 
 /* Used in mmap() */
 mmap_record *
-list::match (DWORD off, DWORD len)
+list::match (__off64_t off, DWORD len)
 {
   if (fd == -1 && !off)
     {
@@ -300,21 +315,23 @@ list::match (DWORD off, DWORD len)
     {
       for (int i = 0; i < nrecs; ++i)
 	if (off >= recs[i].get_offset ()
-	    && off + len <= recs[i].get_offset () + recs[i].get_size ())
+	    && off + len <= recs[i].get_offset ()
+			 + (PAGE_CNT (recs[i].get_size ()) * getpagesize ()))
 	  return recs + i;
     }
   return NULL;
 }
 
 /* Used in munmap() */
-__off32_t
+long
 list::match (caddr_t addr, DWORD len, __off32_t start)
 {
   for (int i = start + 1; i < nrecs; ++i)
     if (addr >= recs[i].get_address ()
-	&& addr + len <= recs[i].get_address () + recs[i].get_size ())
+	&& addr + len <= recs[i].get_address ()
+			 + (PAGE_CNT (recs[i].get_size ()) * getpagesize ()))
       return i;
-  return ILLEGAL_SEEK;
+  return -1;
 }
 
 void
@@ -326,6 +343,12 @@ list::erase (int i)
   nrecs--;
 }
 
+void
+list::erase ()
+{
+  erase (nrecs-1);
+}
+
 class map {
 public:
   list **lists;
@@ -339,7 +362,7 @@ public:
 
 map::map ()
 {
-  lists = (list **) malloc (10 * sizeof(list *));
+  lists = (list **) malloc (10 * sizeof (list *));
   nlists = 0;
   maxlists = 10;
 }
@@ -392,17 +415,16 @@ map::erase (int i)
  * Code to keep a record of all mmap'ed areas in a process.
  * Needed to duplicate tham in a child of fork().
  * mmap_record classes are kept in an STL list in an STL map, keyed
- * by file descriptor. This is *NOT* duplicated accross a fork(), it
+ * by file descriptor. This is *NOT* duplicated across a fork(), it
  * needs to be specially handled by the fork code.
  */
 
 static map *mmapped_areas;
 
-extern "C"
-caddr_t
-mmap (caddr_t addr, size_t len, int prot, int flags, int fd, __off32_t off)
+extern "C" caddr_t
+mmap64 (caddr_t addr, size_t len, int prot, int flags, int fd, __off64_t off)
 {
-  syscall_printf ("addr %x, len %d, prot %x, flags %x, fd %d, off %d",
+  syscall_printf ("addr %x, len %d, prot %x, flags %x, fd %d, off %D",
 		  addr, len, prot, flags, fd, off);
 
   static DWORD granularity;
@@ -425,7 +447,7 @@ mmap (caddr_t addr, size_t len, int prot, int flags, int fd, __off32_t off)
       return MAP_FAILED;
     }
 
-  SetResourceLock(LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
+  SetResourceLock (LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
 
   if (mmapped_areas == NULL)
     {
@@ -435,7 +457,7 @@ mmap (caddr_t addr, size_t len, int prot, int flags, int fd, __off32_t off)
 	{
 	  set_errno (ENOMEM);
 	  syscall_printf ("-1 = mmap(): ENOMEM");
-	  ReleaseResourceLock(LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
+	  ReleaseResourceLock (LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
 	  return MAP_FAILED;
 	}
     }
@@ -444,8 +466,8 @@ mmap (caddr_t addr, size_t len, int prot, int flags, int fd, __off32_t off)
     fd = -1;
 
   /* Map always in multipliers of `granularity'-sized chunks. */
-  DWORD gran_off = off & ~(granularity - 1);
-  DWORD gran_len = howmany (len, granularity) * granularity;
+  __off64_t gran_off = off & ~(granularity - 1);
+  DWORD gran_len = howmany (off + len, granularity) * granularity - gran_off;
 
   fhandler_base *fh;
   caddr_t base = addr;
@@ -458,13 +480,15 @@ mmap (caddr_t addr, size_t len, int prot, int flags, int fd, __off32_t off)
       if (cfd < 0)
 	{
 	  syscall_printf ("-1 = mmap(): EBADF");
-	  ReleaseResourceLock(LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
+	  ReleaseResourceLock (LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
 	  return MAP_FAILED;
 	}
       fh = cfd;
       if (fh->get_device () == FH_DISK)
 	{
-	  DWORD fsiz = GetFileSize (fh->get_handle (), NULL);
+	  DWORD high;
+	  DWORD low = GetFileSize (fh->get_handle (), &high);
+	  __off64_t fsiz = ((__off64_t)high << 32) + low;
 	  fsiz -= gran_off;
 	  if (gran_len > fsiz)
 	    gran_len = fsiz;
@@ -479,20 +503,25 @@ mmap (caddr_t addr, size_t len, int prot, int flags, int fd, __off32_t off)
       fh = &fh_paging_file;
     }
 
-  list *l = mmapped_areas->get_list_by_fd (fd);
+  list *map_list = mmapped_areas->get_list_by_fd (fd);
 
   /* First check if this mapping matches into the chunk of another
      already performed mapping. Only valid for MAP_ANON in a special
      case of MAP_PRIVATE. */
-  if (l && fd == -1 && off == 0 && !(flags & MAP_FIXED))
+  if (map_list && fd == -1 && off == 0 && !(flags & MAP_FIXED))
     {
       mmap_record *rec;
-      if ((rec = l->match (off, len)) != NULL)
+      if ((rec = map_list->match (off, len)) != NULL)
 	{
-	  off = rec->map_map (off, len);
+	  if ((off = rec->map_map (off, len)) == (__off64_t)-1)
+	    {
+	      syscall_printf ("-1 = mmap()");
+	      ReleaseResourceLock (LOCK_MMAP_LIST, READ_LOCK|WRITE_LOCK, "mmap");
+	      return MAP_FAILED;
+	    }
 	  caddr_t ret = rec->get_address () + off;
 	  syscall_printf ("%x = mmap() succeeded", ret);
-	  ReleaseResourceLock(LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
+	  ReleaseResourceLock (LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
 	  return ret;
 	}
     }
@@ -516,7 +545,7 @@ mmap (caddr_t addr, size_t len, int prot, int flags, int fd, __off32_t off)
 
   if (h == INVALID_HANDLE_VALUE)
     {
-      ReleaseResourceLock(LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
+      ReleaseResourceLock (LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
       return MAP_FAILED;
     }
 
@@ -530,35 +559,39 @@ mmap (caddr_t addr, size_t len, int prot, int flags, int fd, __off32_t off)
   /* Get list of mmapped areas for this fd, create a new one if
      one does not exist yet.
   */
-  if (l == 0)
+  if (!map_list)
     {
       /* Create a new one */
-      l = new list;
-      if (l == 0)
+      map_list = new list;
+      if (!map_list)
 	{
 	  fh->munmap (h, base, gran_len);
 	  set_errno (ENOMEM);
 	  syscall_printf ("-1 = mmap(): ENOMEM");
-	  ReleaseResourceLock(LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
+	  ReleaseResourceLock (LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
 	  return MAP_FAILED;
 	}
-      l = mmapped_areas->add_list (l, fd);
+      map_list = mmapped_areas->add_list (map_list, fd);
   }
 
   /* Insert into the list */
-  mmap_record *rec = l->add_record (mmap_rec);
-  off = rec->map_map (off, len);
-  caddr_t ret = rec->get_address () + off;
+  mmap_record *rec = map_list->add_record (mmap_rec, off, len);
+  caddr_t ret = rec->get_address () + (off - gran_off);
   syscall_printf ("%x = mmap() succeeded", ret);
-  ReleaseResourceLock(LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
+  ReleaseResourceLock (LOCK_MMAP_LIST, READ_LOCK | WRITE_LOCK, "mmap");
   return ret;
 }
 
+extern "C" caddr_t
+mmap (caddr_t addr, size_t len, int prot, int flags, int fd, __off32_t off)
+{
+  return mmap64 (addr, len, prot, flags, fd, (__off64_t)off);
+}
+
 /* munmap () removes an mmapped area.  It insists that base area
    requested is the same as that mmapped, error if not. */
 
-extern "C"
-int
+extern "C" int
 munmap (caddr_t addr, size_t len)
 {
   syscall_printf ("munmap (addr %x, len %d)", addr, len);
@@ -571,13 +604,13 @@ munmap (caddr_t addr, size_t len)
       return -1;
     }
 
-  SetResourceLock(LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "munmap");
+  SetResourceLock (LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "munmap");
   /* Check if a mmap'ed area was ever created */
   if (mmapped_areas == NULL)
     {
       syscall_printf ("-1 = munmap(): mmapped_areas == NULL");
       set_errno (EINVAL);
-      ReleaseResourceLock(LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "munmap");
+      ReleaseResourceLock (LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "munmap");
       return -1;
     }
 
@@ -586,13 +619,13 @@ munmap (caddr_t addr, size_t len)
 
   for (int it = 0; it < mmapped_areas->nlists; ++it)
     {
-      list *l = mmapped_areas->lists[it];
-      if (l)
+      list *map_list = mmapped_areas->lists[it];
+      if (map_list)
 	{
-	  __off32_t li = ILLEGAL_SEEK;
-	  if ((li = l->match(addr, len, li)) >= 0)
+	  long li = -1;
+	  if ((li = map_list->match(addr, len, li)) >= 0)
 	    {
-	      mmap_record *rec = l->recs + li;
+	      mmap_record *rec = map_list->recs + li;
 	      if (rec->unmap_map (addr, len))
 		{
 		  fhandler_base *fh = rec->alloc_fh ();
@@ -600,10 +633,10 @@ munmap (caddr_t addr, size_t len)
 		  rec->free_fh (fh);
 
 		  /* Delete the entry. */
-		  l->erase (li);
+		  map_list->erase (li);
 		}
 	      syscall_printf ("0 = munmap(): %x", addr);
-	      ReleaseResourceLock(LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "munmap");
+	      ReleaseResourceLock (LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "munmap");
 	      return 0;
 	    }
 	}
@@ -612,14 +645,13 @@ munmap (caddr_t addr, size_t len)
   set_errno (EINVAL);
   syscall_printf ("-1 = munmap(): EINVAL");
 
-  ReleaseResourceLock(LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "munmap");
+  ReleaseResourceLock (LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "munmap");
   return -1;
 }
 
 /* Sync file with memory. Ignore flags for now. */
 
-extern "C"
-int
+extern "C" int
 msync (caddr_t addr, size_t len, int flags)
 {
   syscall_printf ("addr = %x, len = %d, flags = %x",
@@ -634,13 +666,13 @@ msync (caddr_t addr, size_t len, int flags)
       return -1;
     }
 
-  SetResourceLock(LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "msync");
+  SetResourceLock (LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "msync");
   /* Check if a mmap'ed area was ever created */
   if (mmapped_areas == NULL)
     {
       syscall_printf ("-1 = msync(): mmapped_areas == NULL");
       set_errno (EINVAL);
-      ReleaseResourceLock(LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "msync");
+      ReleaseResourceLock (LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "msync");
       return -1;
     }
 
@@ -649,14 +681,18 @@ msync (caddr_t addr, size_t len, int flags)
 
   for (int it = 0; it < mmapped_areas->nlists; ++it)
     {
-      list *l = mmapped_areas->lists[it];
-      if (l != 0)
+      list *map_list = mmapped_areas->lists[it];
+      if (map_list != 0)
 	{
-	  for (int li = 0; li < l->nrecs; ++li)
+	  for (int li = 0; li < map_list->nrecs; ++li)
 	    {
-	      mmap_record *rec = l->recs + li;
-	      if (rec->get_address () == addr)
+	      mmap_record *rec = map_list->recs + li;
+	      if (rec->access (addr))
 		{
+		  /* Check whole area given by len. */
+		  for (DWORD i = getpagesize (); i < len; ++i)
+		    if (!rec->access (addr + i))
+		      goto invalid_address_range;
 		  fhandler_base *fh = rec->alloc_fh ();
 		  int ret = fh->msync (rec->get_handle (), addr, len, flags);
 		  rec->free_fh (fh);
@@ -666,18 +702,20 @@ msync (caddr_t addr, size_t len, int flags)
 		  else
 		    syscall_printf ("0 = msync()");
 
-		  ReleaseResourceLock(LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "msync");
+		  ReleaseResourceLock (LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK,
+				       "msync");
 		  return 0;
 		}
-	     }
-	 }
-     }
+	    }
+	}
+    }
 
+invalid_address_range:
   /* SUSv2: Return code if indicated memory was not mapped is ENOMEM. */
   set_errno (ENOMEM);
   syscall_printf ("-1 = msync(): ENOMEM");
 
-  ReleaseResourceLock(LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "msync");
+  ReleaseResourceLock (LOCK_MMAP_LIST, WRITE_LOCK | READ_LOCK, "msync");
   return -1;
 }
 
@@ -695,7 +733,7 @@ msync (caddr_t addr, size_t len, int flags)
 */
 HANDLE
 fhandler_base::mmap (caddr_t *addr, size_t len, DWORD access,
-		     int flags, __off32_t off)
+		     int flags, __off64_t off)
 {
   set_errno (ENODEV);
   return INVALID_HANDLE_VALUE;
@@ -726,16 +764,22 @@ fhandler_base::fixup_mmap_after_fork (HANDLE h, DWORD access, DWORD offset,
 /* Implementation for disk files. */
 HANDLE
 fhandler_disk_file::mmap (caddr_t *addr, size_t len, DWORD access,
-			  int flags, __off32_t off)
+			  int flags, __off64_t off)
 {
   DWORD protect;
 
-  if (access & FILE_MAP_COPY)
-    protect = PAGE_WRITECOPY;
-  else if (access & FILE_MAP_WRITE)
-    protect = PAGE_READWRITE;
-  else
-    protect = PAGE_READONLY;
+  switch (access)
+    {
+      case FILE_MAP_WRITE:
+	protect = PAGE_READWRITE;
+	break;
+      case FILE_MAP_READ:
+	protect = PAGE_READONLY;
+	break;
+      default:
+	protect = PAGE_WRITECOPY;
+	break;
+    }
 
   HANDLE h;
 
@@ -756,7 +800,7 @@ fhandler_disk_file::mmap (caddr_t *addr, size_t len, DWORD access,
 
       debug_printf ("named sharing");
       if (!(h = OpenFileMapping (access, TRUE, namebuf)))
-	h = CreateFileMapping (get_handle(), &sec_none, protect, 0, 0, namebuf);
+	h = CreateFileMapping (get_handle (), &sec_none, protect, 0, 0, namebuf);
     }
   else
     h = CreateFileMapping (get_handle (), &sec_none, protect, 0,
@@ -769,9 +813,10 @@ fhandler_disk_file::mmap (caddr_t *addr, size_t len, DWORD access,
       return INVALID_HANDLE_VALUE;
     }
 
-  void *base = MapViewOfFileEx (h, access, 0, off, len,
+  DWORD high = off >> 32, low = off & 0xffffffff;
+  void *base = MapViewOfFileEx (h, access, high, low, len,
 			       (flags & MAP_FIXED) ? *addr : NULL);
-  debug_printf ("%x = MapViewOfFileEx (h:%x, access:%x, 0, off:%d, len:%d, addr:%x)", base, h, access, off, len, (flags & MAP_FIXED) ? *addr : NULL);
+  debug_printf ("%x = MapViewOfFileEx (h:%x, access:%x, 0, off:%D, len:%d, addr:%x)", base, h, access, off, len, (flags & MAP_FIXED) ? *addr : NULL);
   if (!base || ((flags & MAP_FIXED) && base != *addr))
     {
       if (!base)
@@ -817,13 +862,19 @@ fhandler_disk_file::fixup_mmap_after_fork (HANDLE h, DWORD access, DWORD offset,
 {
   /* Re-create the MapViewOfFileEx call */
   void *base = MapViewOfFileEx (h, access, 0, offset, size, address);
+  if (base != address)
+    {
+      MEMORY_BASIC_INFORMATION m;
+      (void) VirtualQuery (address, &m, sizeof (m));
+      system_printf ("requested %p != %p mem alloc base %p, state %p, size %d, %E",
+		     address, base, m.AllocationBase, m.State, m.RegionSize);
+    }
   return base == address;
 }
 
 /* Set memory protection */
 
-extern "C"
-int
+extern "C" int
 mprotect (caddr_t addr, size_t len, int prot)
 {
   DWORD old_prot;
@@ -831,29 +882,39 @@ mprotect (caddr_t addr, size_t len, int prot)
 
   syscall_printf ("mprotect (addr %x, len %d, prot %x)", addr, len, prot);
 
-  if (prot == PROT_NONE)
-    new_prot = PAGE_NOACCESS;
-  else
+  if (!wincap.virtual_protect_works_on_shared_pages ()
+      && addr >= (caddr_t)0x80000000 && addr <= (caddr_t)0xBFFFFFFF)
     {
-      switch (prot)
-	{
-	  case PROT_READ | PROT_WRITE | PROT_EXEC:
-	    new_prot = PAGE_EXECUTE_READWRITE;
-	    break;
-	  case PROT_READ | PROT_WRITE:
-	    new_prot = PAGE_READWRITE;
-	    break;
-	  case PROT_READ | PROT_EXEC:
-	    new_prot = PAGE_EXECUTE_READ;
-	    break;
-	  case PROT_READ:
-	    new_prot = PAGE_READONLY;
-	    break;
-	  default:
-	    syscall_printf ("-1 = mprotect (): invalid prot value");
-	    set_errno (EINVAL);
-	    return -1;
-	 }
+      syscall_printf ("0 = mprotect (9x: No VirtualProtect on shared memory)");
+      return 0;
+    }
+
+  switch (prot)
+    {
+      case PROT_READ | PROT_WRITE | PROT_EXEC:
+      case PROT_WRITE | PROT_EXEC:
+	new_prot = PAGE_EXECUTE_READWRITE;
+	break;
+      case PROT_READ | PROT_WRITE:
+      case PROT_WRITE:
+	new_prot = PAGE_READWRITE;
+	break;
+      case PROT_READ | PROT_EXEC:
+	new_prot = PAGE_EXECUTE_READ;
+	break;
+      case PROT_READ:
+	new_prot = PAGE_READONLY;
+	break;
+      case PROT_EXEC:
+	new_prot = PAGE_EXECUTE;
+	break;
+      case PROT_NONE:
+	new_prot = PAGE_NOACCESS;
+	break;
+      default:
+	syscall_printf ("-1 = mprotect (): invalid prot value");
+	set_errno (EINVAL);
+	return -1;
      }
 
   if (VirtualProtect (addr, len, new_prot, &old_prot) == 0)
@@ -877,7 +938,7 @@ mprotect (caddr_t addr, size_t len, int prot)
  */
 
 int __stdcall
-fixup_mmaps_after_fork ()
+fixup_mmaps_after_fork (HANDLE parent)
 {
 
   debug_printf ("recreate_mmaps_after_fork, mmapped_areas %p", mmapped_areas);
@@ -889,13 +950,13 @@ fixup_mmaps_after_fork ()
   /* Iterate through the map */
   for (int it = 0; it < mmapped_areas->nlists; ++it)
     {
-      list *l = mmapped_areas->lists[it];
-      if (l != 0)
+      list *map_list = mmapped_areas->lists[it];
+      if (map_list)
 	{
 	  int li;
-	  for (li = 0; li < l->nrecs; ++li)
+	  for (li = 0; li < map_list->nrecs; ++li)
 	    {
-	      mmap_record *rec = l->recs + li;
+	      mmap_record *rec = map_list->recs + li;
 
 	      debug_printf ("fd %d, h %x, access %x, offset %d, size %d, address %p",
 		  rec->get_fd (), rec->get_handle (), rec->get_access (),
@@ -910,10 +971,68 @@ fixup_mmaps_after_fork ()
 	      rec->free_fh (fh);
 
 	      if (!ret)
+		return -1;
+	      if (rec->get_access () == FILE_MAP_COPY)
 		{
-		  system_printf ("base address fails to match requested address %p",
-				 rec->get_address ());
-		  return -1;
+		  for (char *address = rec->get_address ();
+		       address < rec->get_address () + rec->get_size ();
+		       address += getpagesize ())
+		    if (rec->access (address)
+			&& !ReadProcessMemory (parent, address, address,
+					       getpagesize (), NULL))
+		      {
+			DWORD old_prot;
+			DWORD last_error = GetLastError ();
+
+			if (last_error != ERROR_PARTIAL_COPY
+			    && last_error != ERROR_NOACCESS
+			    || !wincap.virtual_protect_works_on_shared_pages ())
+			  {
+			    system_printf ("ReadProcessMemory failed for "
+					   "MAP_PRIVATE address %p, %E",
+					   rec->get_address ());
+			    return -1;
+			  }
+			if (!VirtualProtectEx (parent,
+					       address, getpagesize (),
+					       PAGE_READONLY, &old_prot))
+			  {
+			    system_printf ("VirtualProtectEx failed for "
+					   "MAP_PRIVATE address %p, %E",
+					   rec->get_address ());
+			    return -1;
+			  }
+			else
+			  {
+			    BOOL ret;
+			    DWORD dummy_prot;
+
+			    ret = ReadProcessMemory (parent, address, address,
+						     getpagesize (), NULL);
+			    if (!VirtualProtectEx(parent,
+						  address, getpagesize (),
+						  old_prot, &dummy_prot))
+			      system_printf ("WARNING: VirtualProtectEx to "
+					     "return to previous state "
+					     "in parent failed for "
+					     "MAP_PRIVATE address %p, %E",
+					     rec->get_address ());
+			    if (!VirtualProtect (address, getpagesize (),
+						 old_prot, &dummy_prot))
+			      system_printf ("WARNING: VirtualProtect to copy "
+					     "protection to child failed for"
+					     "MAP_PRIVATE address %p, %E",
+					     rec->get_address ());
+			    if (!ret)
+			      {
+				system_printf ("ReadProcessMemory (2nd try) "
+					       "failed for "
+					       "MAP_PRIVATE address %p, %E",
+					       rec->get_address ());
+				return -1;
+			      }
+			  }
+		      }
 		}
 	      rec->fixup_map ();
 	    }
