@@ -8,13 +8,15 @@ This software is a copyrighted work licensed under the terms of the
 Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
 details. */
 
+#ifndef _PINFO_H
+#define _PINFO_H
 /* Signal constants (have to define them here, unfortunately) */
 
 enum
 {
   __SIGFLUSH	    = -2,
   __SIGSTRACE	    = -1,
-  __SIGUNUSED	    =  0,
+  __SIGCOMMUNE	    =  0,
   __SIGOFFSET	    =  2
 };
 
@@ -23,6 +25,17 @@ enum
 #include <sys/resource.h>
 #include "thread.h"
 
+struct commune_result
+{
+  char *s;
+  int n;
+};
+
+enum picom
+{
+  PICOM_CMDLINE = 1
+};
+
 class _pinfo
 {
 public:
@@ -39,7 +52,7 @@ public:
      we only use this handle from the parent. */
   HANDLE hProcess;
 
-#define PINFO_REDIR_SIZE ((DWORD) &(((_pinfo *)NULL)->hProcess) + sizeof (DWORD))
+#define PINFO_REDIR_SIZE ((char *) &myself.procinfo->hProcess - (char *) myself.procinfo)
 
   /* Handle associated with initial Windows pid which started it all. */
   HANDLE pid_handle;
@@ -65,8 +78,8 @@ public:
      if not found.  This data resides in the shared data area (allowing
      tasks to store whatever they want here) so it's for informational
      purposes only. */
-  __uid16_t uid;		/* User ID */
-  __gid16_t gid;		/* Group ID */
+  __uid32_t uid;	/* User ID */
+  __gid32_t gid;	/* Group ID */
   pid_t pgid;		/* Process group ID */
   pid_t sid;		/* Session ID */
   int ctty;		/* Control tty */
@@ -80,6 +93,11 @@ public:
   /* Non-zero if process was stopped by a signal. */
   char stopsig;
 
+  /* commune */
+  pid_t hello_pid;
+  HANDLE tothem;
+  HANDLE fromthem;
+
   void exit (UINT n, bool norecord = 0) __attribute__ ((noreturn, regparm(2)));
 
   inline void set_has_pgid_children ()
@@ -95,7 +113,7 @@ public:
     return thread2signal ? thread2signal->sigs[sig] : sigs[sig];
   }
 
-  inline void copysigs (_pinfo *p) {sigs = p->sigs;}
+  inline void copysigs (_pinfo *p) {memcpy (sigs, p->sigs, sizeof (sigs));}
 
   inline sigset_t& getsigmask ()
   {
@@ -117,19 +135,26 @@ public:
   }
 
   inline void setthread2signal (void *thr) {thread2signal = (pthread *) thr;}
+  void commune_recv ();
+  commune_result commune_send (DWORD);
+  bool alive ();
+  char *cmdline (size_t &);
+
+  friend void __stdcall set_myself (pid_t, HANDLE);
 
 private:
   struct sigaction sigs[NSIG];
   sigset_t sig_mask;		/* one set for everything to ignore. */
   LONG _sigtodo[NSIG + __SIGOFFSET];
-  pthread *thread2signal;  // NULL means means thread any other means a pthread
+  pthread *thread2signal;  // NULL means thread any other means a pthread
+  CRITICAL_SECTION lock;
 };
 
 class pinfo
 {
   HANDLE h;
   _pinfo *procinfo;
-  int destroy;
+  bool destroy;
 public:
   void init (pid_t n, DWORD create = 0, HANDLE h = NULL) __attribute__ ((regparm(3)));
   pinfo () {}
@@ -152,7 +177,16 @@ public:
   _pinfo *operator * () const {return procinfo;}
   operator _pinfo * () const {return procinfo;}
   // operator bool () const {return (int) h;}
-  int remember () {destroy = 0; return proc_subproc (PROC_ADDCHILD, (DWORD) this);}
+#ifndef _SIGPROC_H
+  int remember () {system_printf ("remember is not here"); return 0;}
+#else
+  int remember ()
+  {
+    int res = proc_subproc (PROC_ADDCHILD, (DWORD) this);
+    destroy = res ? false : true;
+    return res;
+  }
+#endif
   HANDLE shared_handle () {return h;}
 };
 
@@ -169,17 +203,19 @@ class winpids
   DWORD enumNT (bool winpid);
   DWORD enum9x (bool winpid);
   void add (DWORD& nelem, bool, DWORD pid);
+  static CRITICAL_SECTION cs;
 public:
   DWORD npids;
   inline void reset () { npids = 0; release (); }
-  void init (bool winpid);
+  void set (bool winpid);
   winpids (int): enum_processes (&winpids::enum_init) { reset (); }
   winpids (): pidlist (NULL), npidlist (0), pinfolist (NULL),
-	      enum_processes (&winpids::enum_init), npids (0) { init (0); }
+	      enum_processes (&winpids::enum_init), npids (0) { set (0); }
   inline DWORD& winpid (int i) const {return pidlist[i];}
   inline _pinfo *operator [] (int i) const {return (_pinfo *) pinfolist[i];}
   ~winpids ();
   void release ();
+  static void init ();
 };
 
 extern __inline pid_t
@@ -193,14 +229,14 @@ void __stdcall set_myself (pid_t pid, HANDLE h = NULL);
 extern pinfo myself;
 
 #define _P_VFORK 0
-extern "C" int _spawnve (HANDLE hToken, int mode, const char *path,
-			 const char *const *argv, const char *const *envp);
-
 extern void __stdcall pinfo_fixup_after_fork ();
 extern HANDLE hexec_proc;
 
 /* For mmaps across fork(). */
-int __stdcall fixup_mmaps_after_fork ();
+int __stdcall fixup_mmaps_after_fork (HANDLE parent);
+/* for shm areas across fork (). */
+int __stdcall fixup_shms_after_fork ();
 
 void __stdcall fill_rusage (struct rusage *, HANDLE);
 void __stdcall add_rusage (struct rusage *, struct rusage *);
+#endif /*_PINFO_H*/
