@@ -1,6 +1,6 @@
-/* secacl.cc: Sun compatible ACL functions.
+/* sec_acl.cc: Sun compatible ACL functions.
 
-   Copyright 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 2000, 2001, 2002, 2003 Red Hat, Inc.
 
    Written by Corinna Vinschen <corinna@vinschen.de>
 
@@ -24,33 +24,31 @@ details. */
 #include <wingdi.h>
 #include <winuser.h>
 #include "cygerrno.h"
-#include "perprocess.h"
 #include "security.h"
 #include "fhandler.h"
 #include "path.h"
 #include "dtable.h"
-#include "sync.h"
-#include "sigproc.h"
 #include "pinfo.h"
 #include "cygheap.h"
+#include "pwdgrp.h"
 
-extern "C" int aclsort (int nentries, int, __aclent16_t *aclbufp);
-extern "C" int acl (const char *path, int cmd, int nentries, __aclent16_t *aclbufp);
+extern "C" int aclsort32 (int nentries, int, __aclent32_t *aclbufp);
+extern "C" int acl32 (const char *path, int cmd, int nentries, __aclent32_t *aclbufp);
 
 static int
-searchace (__aclent16_t *aclp, int nentries, int type, int id = -1)
+searchace (__aclent32_t *aclp, int nentries, int type, __uid32_t id = ILLEGAL_UID)
 {
   int i;
 
   for (i = 0; i < nentries; ++i)
-    if ((aclp[i].a_type == type && (id < 0 || aclp[i].a_id == id))
+    if ((aclp[i].a_type == type && (id == ILLEGAL_UID || aclp[i].a_id == id))
 	|| !aclp[i].a_type)
       return i;
   return -1;
 }
 
 static int
-setacl (const char *file, int nentries, __aclent16_t *aclbufp)
+setacl (const char *file, int nentries, __aclent32_t *aclbufp)
 {
   DWORD sd_size = 4096;
   char sd_buf[4096];
@@ -65,7 +63,7 @@ setacl (const char *file, int nentries, __aclent16_t *aclbufp)
   BOOL dummy;
 
   /* Get owner SID. */
-  PSID owner_sid = NULL;
+  PSID owner_sid;
   if (!GetSecurityDescriptorOwner (psd, &owner_sid, &dummy))
     {
       __seterrno ();
@@ -74,7 +72,7 @@ setacl (const char *file, int nentries, __aclent16_t *aclbufp)
   cygsid owner (owner_sid);
 
   /* Get group SID. */
-  PSID group_sid = NULL;
+  PSID group_sid;
   if (!GetSecurityDescriptorGroup (psd, &group_sid, &dummy))
     {
       __seterrno ();
@@ -89,13 +87,12 @@ setacl (const char *file, int nentries, __aclent16_t *aclbufp)
       __seterrno ();
       return -1;
     }
-  if (!SetSecurityDescriptorOwner(&sd, owner, FALSE))
+  if (!SetSecurityDescriptorOwner (&sd, owner, FALSE))
     {
       __seterrno ();
       return -1;
     }
-  if (group
-      && !SetSecurityDescriptorGroup(&sd, group, FALSE))
+  if (!SetSecurityDescriptorGroup (&sd, group, FALSE))
     {
       __seterrno ();
       return -1;
@@ -109,7 +106,7 @@ setacl (const char *file, int nentries, __aclent16_t *aclbufp)
 
   cygsid sid;
   struct passwd *pw;
-  struct __group16 *gr;
+  struct __group32 *gr;
   int pos;
 
   if (!InitializeAcl (acl, 3072, ACL_REVISION))
@@ -119,65 +116,77 @@ setacl (const char *file, int nentries, __aclent16_t *aclbufp)
     }
   for (int i = 0; i < nentries; ++i)
     {
-      DWORD allow = STANDARD_RIGHTS_READ
-		    | FILE_READ_ATTRIBUTES | FILE_READ_EA;
+      DWORD allow;
+      /* Owner has more standard rights set. */
+      if ((aclbufp[i].a_type & ~ACL_DEFAULT) == USER_OBJ)
+	allow = STANDARD_RIGHTS_ALL | FILE_WRITE_ATTRIBUTES | FILE_WRITE_EA;
+      else
+	allow = STANDARD_RIGHTS_READ | FILE_READ_ATTRIBUTES | FILE_READ_EA;
       if (aclbufp[i].a_perm & S_IROTH)
 	allow |= FILE_GENERIC_READ;
       if (aclbufp[i].a_perm & S_IWOTH)
-	allow |= STANDARD_RIGHTS_ALL | FILE_GENERIC_WRITE;
+	allow |= STANDARD_RIGHTS_WRITE | FILE_GENERIC_WRITE;
       if (aclbufp[i].a_perm & S_IXOTH)
 	allow |= FILE_GENERIC_EXECUTE;
       if ((aclbufp[i].a_perm & (S_IWOTH | S_IXOTH)) == (S_IWOTH | S_IXOTH))
 	allow |= FILE_DELETE_CHILD;
       /* Set inherit property. */
       DWORD inheritance = (aclbufp[i].a_type & ACL_DEFAULT)
-			  ? INHERIT_ONLY : DONT_INHERIT;
+			  ? (SUB_CONTAINERS_AND_OBJECTS_INHERIT | INHERIT_ONLY)
+			  : NO_INHERITANCE;
       /*
        * If a specific acl contains a corresponding default entry with
        * identical permissions, only one Windows ACE with proper
        * inheritance bits is created.
        */
       if (!(aclbufp[i].a_type & ACL_DEFAULT)
-	  && (pos = searchace (aclbufp, nentries,
+	  && aclbufp[i].a_type & (USER|GROUP|OTHER_OBJ)
+	  && (pos = searchace (aclbufp + i + 1, nentries - i - 1,
 			       aclbufp[i].a_type | ACL_DEFAULT,
 			       (aclbufp[i].a_type & (USER|GROUP))
-			       ? aclbufp[i].a_id : -1)) >= 0
-	  && pos < nentries
+			       ? aclbufp[i].a_id : ILLEGAL_UID)) >= 0
 	  && aclbufp[i].a_perm == aclbufp[pos].a_perm)
 	{
-	  inheritance = INHERIT_ALL;
-	  /* This eliminates the corresponding default entry. */
-	  aclbufp[pos].a_type = 0;
+	  inheritance = SUB_CONTAINERS_AND_OBJECTS_INHERIT;
+	  /* This invalidates the corresponding default entry. */
+	  aclbufp[pos].a_type = USER|GROUP|ACL_DEFAULT;
 	}
       switch (aclbufp[i].a_type)
 	{
 	case USER_OBJ:
-	case DEF_USER_OBJ:
-	  allow |= STANDARD_RIGHTS_ALL & ~DELETE;
 	  if (!add_access_allowed_ace (acl, ace_off++, allow,
 					owner, acl_len, inheritance))
 	    return -1;
 	  break;
+	case DEF_USER_OBJ:
+	  if (!add_access_allowed_ace (acl, ace_off++, allow,
+				       well_known_creator_owner_sid, acl_len, inheritance))
+	    return -1;
+	  break;
 	case USER:
 	case DEF_USER:
-	  if (!(pw = getpwuid (aclbufp[i].a_id))
+	  if (!(pw = internal_getpwuid (aclbufp[i].a_id))
 	      || !sid.getfrompw (pw)
 	      || !add_access_allowed_ace (acl, ace_off++, allow,
-					   sid, acl_len, inheritance))
+					  sid, acl_len, inheritance))
 	    return -1;
 	  break;
 	case GROUP_OBJ:
+	  if (!add_access_allowed_ace (acl, ace_off++, allow,
+				       group, acl_len, inheritance))
+	    return -1;
+	  break;
 	case DEF_GROUP_OBJ:
 	  if (!add_access_allowed_ace (acl, ace_off++, allow,
-					group, acl_len, inheritance))
+				       well_known_creator_group_sid, acl_len, inheritance))
 	    return -1;
 	  break;
 	case GROUP:
 	case DEF_GROUP:
-	  if (!(gr = getgrgid (aclbufp[i].a_id))
+	  if (!(gr = internal_getgrgid (aclbufp[i].a_id))
 	      || !sid.getfromgr (gr)
 	      || !add_access_allowed_ace (acl, ace_off++, allow,
-					   sid, acl_len, inheritance))
+					  sid, acl_len, inheritance))
 	    return -1;
 	  break;
 	case OTHER_OBJ:
@@ -215,33 +224,39 @@ setacl (const char *file, int nentries, __aclent16_t *aclbufp)
   return write_sd (file, psd, sd_size);
 }
 
+/* Temporary access denied bits */
+#define DENY_R 040000
+#define DENY_W 020000
+#define DENY_X 010000
+
 static void
-getace (__aclent16_t &acl, int type, int id, DWORD win_ace_mask, DWORD win_ace_type)
+getace (__aclent32_t &acl, int type, int id, DWORD win_ace_mask,
+	DWORD win_ace_type)
 {
   acl.a_type = type;
   acl.a_id = id;
 
-  if (win_ace_mask & FILE_READ_DATA)
+  if ((win_ace_mask & FILE_READ_DATA) && !(acl.a_perm & (S_IROTH | DENY_R)))
     if (win_ace_type == ACCESS_ALLOWED_ACE_TYPE)
-      acl.a_perm |= (acl.a_perm & S_IRGRP) ? 0 : S_IRUSR;
+      acl.a_perm |= S_IROTH;
     else if (win_ace_type == ACCESS_DENIED_ACE_TYPE)
-      acl.a_perm &= ~S_IRGRP;
+      acl.a_perm |= DENY_R;
 
-  if (win_ace_mask & FILE_WRITE_DATA)
+  if ((win_ace_mask & FILE_WRITE_DATA) && !(acl.a_perm & (S_IWOTH | DENY_W)))
     if (win_ace_type == ACCESS_ALLOWED_ACE_TYPE)
-      acl.a_perm |= (acl.a_perm & S_IWGRP) ? 0 : S_IWUSR;
+      acl.a_perm |= S_IWOTH;
     else if (win_ace_type == ACCESS_DENIED_ACE_TYPE)
-      acl.a_perm &= ~S_IWGRP;
+      acl.a_perm |= DENY_W;
 
-  if (win_ace_mask & FILE_EXECUTE)
+  if ((win_ace_mask & FILE_EXECUTE) && !(acl.a_perm & (S_IXOTH | DENY_X)))
     if (win_ace_type == ACCESS_ALLOWED_ACE_TYPE)
-      acl.a_perm |= (acl.a_perm & S_IXGRP) ? 0 : S_IXUSR;
+      acl.a_perm |= S_IXOTH;
     else if (win_ace_type == ACCESS_DENIED_ACE_TYPE)
-      acl.a_perm &= ~S_IXGRP;
+      acl.a_perm |= DENY_X;
 }
 
 static int
-getacl (const char *file, DWORD attr, int nentries, __aclent16_t *aclbufp)
+getacl (const char *file, DWORD attr, int nentries, __aclent32_t *aclbufp)
 {
   DWORD sd_size = 4096;
   char sd_buf[4096];
@@ -254,35 +269,39 @@ getacl (const char *file, DWORD attr, int nentries, __aclent16_t *aclbufp)
       return ret;
     }
 
-  PSID owner_sid;
-  PSID group_sid;
+  cygpsid owner_sid;
+  cygpsid group_sid;
   BOOL dummy;
-  __uid16_t uid;
-  __gid16_t gid;
+  __uid32_t uid;
+  __gid32_t gid;
 
-  if (!GetSecurityDescriptorOwner (psd, &owner_sid, &dummy))
+  if (!GetSecurityDescriptorOwner (psd, (PSID *) &owner_sid, &dummy))
     {
       debug_printf ("GetSecurityDescriptorOwner %E");
       __seterrno ();
       return -1;
     }
-  uid = cygsid (owner_sid).get_uid ();
+  uid = owner_sid.get_uid ();
 
-  if (!GetSecurityDescriptorGroup (psd, &group_sid, &dummy))
+  if (!GetSecurityDescriptorGroup (psd, (PSID *) &group_sid, &dummy))
     {
       debug_printf ("GetSecurityDescriptorGroup %E");
       __seterrno ();
       return -1;
     }
-  gid = cygsid (group_sid).get_gid ();
+  gid = group_sid.get_gid ();
 
-  __aclent16_t lacl[MAX_ACL_ENTRIES];
-  memset (&lacl, 0, MAX_ACL_ENTRIES * sizeof (__aclent16_t));
+  __aclent32_t lacl[MAX_ACL_ENTRIES];
+  memset (&lacl, 0, MAX_ACL_ENTRIES * sizeof (__aclent32_t));
   lacl[0].a_type = USER_OBJ;
   lacl[0].a_id = uid;
   lacl[1].a_type = GROUP_OBJ;
   lacl[1].a_id = gid;
   lacl[2].a_type = OTHER_OBJ;
+  lacl[2].a_id = ILLEGAL_GID;
+  lacl[3].a_type = CLASS_OBJ;
+  lacl[3].a_id = ILLEGAL_GID;
+  lacl[3].a_perm = S_IROTH | S_IWOTH | S_IXOTH;
 
   PACL acl;
   BOOL acl_exists;
@@ -294,176 +313,102 @@ getacl (const char *file, DWORD attr, int nentries, __aclent16_t *aclbufp)
       return -1;
     }
 
-  int pos, i;
+  int pos, i, types_def = 0;
 
   if (!acl_exists || !acl)
+    for (pos = 0; pos < 3; ++pos) /* Don't change CLASS_OBJ entry */
+      lacl[pos].a_perm = S_IROTH | S_IWOTH | S_IXOTH;
+  else
     {
-      for (pos = 0; pos < MIN_ACL_ENTRIES; ++pos)
-	lacl[pos].a_perm = S_IRWXU | S_IRWXG | S_IRWXO;
-      pos = nentries < MIN_ACL_ENTRIES ? nentries : MIN_ACL_ENTRIES;
-      memcpy (aclbufp, lacl, pos * sizeof (__aclent16_t));
-      return pos;
-    }
-
-  for (i = 0; i < acl->AceCount && (!nentries || i < nentries); ++i)
-    {
-      ACCESS_ALLOWED_ACE *ace;
-
-      if (!GetAce (acl, i, (PVOID *) &ace))
-	continue;
-
-      cygsid ace_sid ((PSID) &ace->SidStart);
-      int id;
-      int type = 0;
-
-      if (ace_sid == owner_sid)
+      for (i = 0; i < acl->AceCount; ++i)
 	{
-	  type = USER_OBJ;
-	  id = uid;
-	}
-      else if (ace_sid == group_sid)
-	{
-	  type = GROUP_OBJ;
-	  id = gid;
-	}
-      else if (ace_sid == well_known_world_sid)
-	{
-	  type = OTHER_OBJ;
-	  id = 0;
-	}
-      else
-	{
-	  id = ace_sid.get_id (FALSE, &type);
-	  if (type != GROUP)
+	  ACCESS_ALLOWED_ACE *ace;
+
+	  if (!GetAce (acl, i, (PVOID *) &ace))
+	    continue;
+
+	  cygpsid ace_sid ((PSID) &ace->SidStart);
+	  int id;
+	  int type = 0;
+
+	  if (ace_sid == well_known_world_sid)
 	    {
-	      int type2 = 0;
-	      int id2 = ace_sid.get_id (TRUE, &type2);
-	      if (type2 == GROUP)
-		{
-		  id = id2;
-		  type = GROUP;
-		}
+	      type = OTHER_OBJ;
+	      id = ILLEGAL_GID;
+	    }
+	  else if (ace_sid == group_sid)
+	    {
+	      type = GROUP_OBJ;
+	      id = gid;
+	    }
+	  else if (ace_sid == owner_sid)
+	    {
+	      type = USER_OBJ;
+	      id = uid;
+	    }
+	  else if (ace_sid == well_known_creator_group_sid)
+	    {
+	      type = GROUP_OBJ | ACL_DEFAULT;
+	      id = ILLEGAL_GID;
+	    }
+	  else if (ace_sid == well_known_creator_owner_sid)
+	    {
+	      type = USER_OBJ | ACL_DEFAULT;
+	      id = ILLEGAL_GID;
+	    }
+	  else
+	    id = ace_sid.get_id (TRUE, &type);
+
+	  if (!type)
+	    continue;
+	  if (!(ace->Header.AceFlags & INHERIT_ONLY || type & ACL_DEFAULT))
+	    {
+	      if ((pos = searchace (lacl, MAX_ACL_ENTRIES, type, id)) >= 0)
+		getace (lacl[pos], type, id, ace->Mask, ace->Header.AceType);
+	    }
+	  if ((ace->Header.AceFlags & SUB_CONTAINERS_AND_OBJECTS_INHERIT)
+	      && (attr & FILE_ATTRIBUTE_DIRECTORY))
+	    {
+	      if (type == USER_OBJ)
+		type = USER;
+	      else if (type == GROUP_OBJ)
+		type = GROUP;
+	      type |= ACL_DEFAULT;
+	      types_def |= type;
+	      if ((pos = searchace (lacl, MAX_ACL_ENTRIES, type, id)) >= 0)
+		getace (lacl[pos], type, id, ace->Mask, ace->Header.AceType);
 	    }
 	}
-      if (!type)
-	continue;
-      if (!(ace->Header.AceFlags & INHERIT_ONLY_ACE))
+      /* Include DEF_CLASS_OBJ if any default ace exists */
+      if ((types_def & (USER|GROUP))
+	  && ((pos = searchace (lacl, MAX_ACL_ENTRIES, DEF_CLASS_OBJ)) >= 0))
 	{
-	  if ((pos = searchace (lacl, MAX_ACL_ENTRIES, type, id)) >= 0)
-	    getace (lacl[pos], type, id, ace->Mask, ace->Header.AceType);
-	}
-      if ((ace->Header.AceFlags & INHERIT_ALL)
-	  && (attr & FILE_ATTRIBUTE_DIRECTORY))
-	{
-	  type |= ACL_DEFAULT;
-	  if ((pos = searchace (lacl, MAX_ACL_ENTRIES, type, id)) >= 0)
-	    getace (lacl[pos], type, id, ace->Mask, ace->Header.AceType);
+	  lacl[pos].a_type = DEF_CLASS_OBJ;
+	  lacl[pos].a_id = ILLEGAL_GID;
+	  lacl[pos].a_perm = S_IRWXU | S_IRWXG | S_IRWXO;
 	}
     }
   if ((pos = searchace (lacl, MAX_ACL_ENTRIES, 0)) < 0)
     pos = MAX_ACL_ENTRIES;
-  for (i = 0; i < pos; ++i)
-    {
-      lacl[i].a_perm = (lacl[i].a_perm & S_IRWXU)
-		       & ~((lacl[i].a_perm & S_IRWXG) << 3);
-      lacl[i].a_perm |= (lacl[i].a_perm & S_IRWXU) >> 3
-			| (lacl[i].a_perm & S_IRWXU) >> 6;
-    }
-  if ((searchace (lacl, MAX_ACL_ENTRIES, USER) >= 0
-       || searchace (lacl, MAX_ACL_ENTRIES, GROUP) >= 0)
-      && (pos = searchace (lacl, MAX_ACL_ENTRIES, CLASS_OBJ)) >= 0)
-    {
-      lacl[pos].a_type = CLASS_OBJ;
-      lacl[pos].a_perm =
-	  lacl[searchace (lacl, MAX_ACL_ENTRIES, GROUP_OBJ)].a_perm;
-    }
-  int dgpos;
-  if ((searchace (lacl, MAX_ACL_ENTRIES, DEF_USER) >= 0
-       || searchace (lacl, MAX_ACL_ENTRIES, DEF_GROUP) >= 0)
-      && (dgpos = searchace (lacl, MAX_ACL_ENTRIES, DEF_GROUP_OBJ)) >= 0
-      && (pos = searchace (lacl, MAX_ACL_ENTRIES, DEF_CLASS_OBJ)) >= 0
-      && (attr & FILE_ATTRIBUTE_DIRECTORY))
-    {
-      lacl[pos].a_type = DEF_CLASS_OBJ;
-      lacl[pos].a_perm = lacl[dgpos].a_perm;
-    }
-  if ((pos = searchace (lacl, MAX_ACL_ENTRIES, 0)) < 0)
-    pos = MAX_ACL_ENTRIES;
-  if (pos > nentries)
-    pos = nentries;
-  if (aclbufp)
-    memcpy (aclbufp, lacl, pos * sizeof (__aclent16_t));
-  aclsort (pos, 0, aclbufp);
+  if (aclbufp) {
+    if (owner_sid == group_sid)
+      lacl[0].a_perm = lacl[1].a_perm;
+    if (pos > nentries)
+      {
+	set_errno (ENOSPC);
+	return -1;
+      }
+    memcpy (aclbufp, lacl, pos * sizeof (__aclent32_t));
+    for (i = 0; i < pos; ++i)
+      aclbufp[i].a_perm &= ~(DENY_R | DENY_W | DENY_X);
+    aclsort32 (pos, 0, aclbufp);
+  }
   syscall_printf ("%d = getacl (%s)", pos, file);
   return pos;
 }
 
-int
-acl_access (const char *path, int flags)
-{
-  __aclent16_t acls[MAX_ACL_ENTRIES];
-  int cnt;
-
-  if ((cnt = acl (path, GETACL, MAX_ACL_ENTRIES, acls)) < 1)
-    return -1;
-
-  /* Only check existance. */
-  if (!(flags & (R_OK|W_OK|X_OK)))
-    return 0;
-
-  for (int i = 0; i < cnt; ++i)
-    {
-      switch (acls[i].a_type)
-	{
-	case USER_OBJ:
-	case USER:
-	  if (acls[i].a_id != myself->uid)
-	    {
-	      /*
-	       * Check if user is a NT group:
-	       * Take SID from passwd, search SID in group, check is_grp_member.
-	       */
-	      cygsid owner;
-	      cygsid group;
-	      struct passwd *pw;
-	      struct __group16 *gr = NULL;
-
-	      if ((pw = getpwuid (acls[i].a_id)) != NULL
-		  && owner.getfrompw (pw))
-		{
-		  for (int gidx = 0; (gr = internal_getgrent (gidx)); ++gidx)
-		    if (group.getfromgr (gr)
-			&& owner == group
-			&& is_grp_member (myself->uid, gr->gr_gid))
-		      break;
-		}
-	      if (!gr)
-		continue;
-	    }
-	  break;
-	case GROUP_OBJ:
-	case GROUP:
-	  if (acls[i].a_id != myself->gid &&
-	      !is_grp_member (myself->uid, acls[i].a_id))
-	    continue;
-	  break;
-	case OTHER_OBJ:
-	  break;
-	default:
-	  continue;
-	}
-      if ((!(flags & R_OK) || (acls[i].a_perm & S_IREAD))
-	  && (!(flags & W_OK) || (acls[i].a_perm & S_IWRITE))
-	  && (!(flags & X_OK) || (acls[i].a_perm & S_IEXEC)))
-	return 0;
-    }
-  set_errno (EACCES);
-  return -1;
-}
-
-static
-int
-acl_worker (const char *path, int cmd, int nentries, __aclent16_t *aclbufp,
+static int
+acl_worker (const char *path, int cmd, int nentries, __aclent32_t *aclbufp,
 	    int nofollow)
 {
   extern suffix_info stat_suffixes[];
@@ -474,9 +419,9 @@ acl_worker (const char *path, int cmd, int nentries, __aclent16_t *aclbufp,
       syscall_printf ("-1 = acl (%s)", path);
       return -1;
     }
-  if (!real_path.has_acls ())
+  if (!real_path.has_acls () || !allow_ntsec)
     {
-      struct stat st;
+      struct __stat64 st;
       int ret = -1;
 
       switch (cmd)
@@ -485,53 +430,30 @@ acl_worker (const char *path, int cmd, int nentries, __aclent16_t *aclbufp,
 	  set_errno (ENOSYS);
 	  break;
 	case GETACL:
-	  if (nentries < 1)
-	    set_errno (EINVAL);
-	  else if ((nofollow && !lstat (path, &st))
-		   || (!nofollow && !stat (path, &st)))
+	  if (!aclbufp)
+	    set_errno(EFAULT);
+	  else if (nentries < MIN_ACL_ENTRIES)
+	    set_errno (ENOSPC);
+	  else if ((nofollow && !lstat64 (path, &st))
+		   || (!nofollow && !stat64 (path, &st)))
 	    {
-	      __aclent16_t lacl[4];
-	      if (nentries > 0)
-		{
-		  lacl[0].a_type = USER_OBJ;
-		  lacl[0].a_id = st.st_uid;
-		  lacl[0].a_perm = (st.st_mode & S_IRWXU)
-				   | (st.st_mode & S_IRWXU) >> 3
-				   | (st.st_mode & S_IRWXU) >> 6;
-		}
-	      if (nentries > 1)
-		{
-		  lacl[1].a_type = GROUP_OBJ;
-		  lacl[1].a_id = st.st_gid;
-		  lacl[1].a_perm = (st.st_mode & S_IRWXG)
-				   | (st.st_mode & S_IRWXG) << 3
-				   | (st.st_mode & S_IRWXG) >> 3;
-		}
-	      if (nentries > 2)
-		{
-		  lacl[2].a_type = OTHER_OBJ;
-		  lacl[2].a_id = 0;
-		  lacl[2].a_perm = (st.st_mode & S_IRWXO)
-				   | (st.st_mode & S_IRWXO) << 6
-				   | (st.st_mode & S_IRWXO) << 3;
-		}
-	      if (nentries > 3)
-		{
-		  lacl[3].a_type = CLASS_OBJ;
-		  lacl[3].a_id = 0;
-		  lacl[3].a_perm = (st.st_mode & S_IRWXG)
-				   | (st.st_mode & S_IRWXG) << 3
-				   | (st.st_mode & S_IRWXG) >> 3;
-		}
-	      if (nentries > 4)
-		nentries = 4;
-	      if (aclbufp)
-		memcpy (aclbufp, lacl, nentries * sizeof (__aclent16_t));
-	      ret = nentries;
+	      aclbufp[0].a_type = USER_OBJ;
+	      aclbufp[0].a_id = st.st_uid;
+	      aclbufp[0].a_perm = (st.st_mode & S_IRWXU) >> 6;
+	      aclbufp[1].a_type = GROUP_OBJ;
+	      aclbufp[1].a_id = st.st_gid;
+	      aclbufp[1].a_perm = (st.st_mode & S_IRWXG) >> 3;
+	      aclbufp[2].a_type = OTHER_OBJ;
+	      aclbufp[2].a_id = ILLEGAL_GID;
+	      aclbufp[2].a_perm = st.st_mode & S_IRWXO;
+	      aclbufp[3].a_type = CLASS_OBJ;
+	      aclbufp[3].a_id = ILLEGAL_GID;
+	      aclbufp[3].a_perm = S_IRWXU | S_IRWXG | S_IRWXO;
+	      ret = MIN_ACL_ENTRIES;
 	    }
 	  break;
 	case GETACLCNT:
-	  ret = 4;
+	  ret = MIN_ACL_ENTRIES;
 	  break;
 	}
       syscall_printf ("%d = acl (%s)", ret, path);
@@ -540,45 +462,44 @@ acl_worker (const char *path, int cmd, int nentries, __aclent16_t *aclbufp,
   switch (cmd)
     {
       case SETACL:
-	if (!aclsort(nentries, 0, aclbufp))
+	if (!aclsort32 (nentries, 0, aclbufp))
 	  return setacl (real_path.get_win32 (),
 			 nentries, aclbufp);
 	break;
       case GETACL:
-	if (nentries < 1)
-	  break;
-	return getacl (real_path.get_win32 (),
-		       real_path.file_attributes (),
-		       nentries, aclbufp);
+	if (!aclbufp)
+	  set_errno(EFAULT);
+	else
+	  return getacl (real_path.get_win32 (),
+			 real_path.file_attributes (),
+			 nentries, aclbufp);
+	break;
       case GETACLCNT:
 	return getacl (real_path.get_win32 (),
 		       real_path.file_attributes (),
 		       0, NULL);
       default:
+	set_errno (EINVAL);
 	break;
     }
-  set_errno (EINVAL);
   syscall_printf ("-1 = acl (%s)", path);
   return -1;
 }
 
-extern "C"
-int
-acl (const char *path, int cmd, int nentries, __aclent16_t *aclbufp)
+extern "C" int
+acl32 (const char *path, int cmd, int nentries, __aclent32_t *aclbufp)
 {
   return acl_worker (path, cmd, nentries, aclbufp, 0);
 }
 
-extern "C"
-int
-lacl (const char *path, int cmd, int nentries, __aclent16_t *aclbufp)
+extern "C" int
+lacl32 (const char *path, int cmd, int nentries, __aclent32_t *aclbufp)
 {
   return acl_worker (path, cmd, nentries, aclbufp, 1);
 }
 
-extern "C"
-int
-facl (int fd, int cmd, int nentries, __aclent16_t *aclbufp)
+extern "C" int
+facl32 (int fd, int cmd, int nentries, __aclent32_t *aclbufp)
 {
   cygheap_fdget cfd (fd);
   if (cfd < 0)
@@ -597,9 +518,8 @@ facl (int fd, int cmd, int nentries, __aclent16_t *aclbufp)
   return acl_worker (path, cmd, nentries, aclbufp, 0);
 }
 
-extern "C"
-int
-aclcheck (__aclent16_t *aclbufp, int nentries, int *which)
+extern "C" int
+aclcheck32 (__aclent32_t *aclbufp, int nentries, int *which)
 {
   BOOL has_user_obj = FALSE;
   BOOL has_group_obj = FALSE;
@@ -730,35 +650,33 @@ aclcheck (__aclent16_t *aclbufp, int nentries, int *which)
   return 0;
 }
 
-extern "C"
-int acecmp (const void *a1, const void *a2)
+static int
+acecmp (const void *a1, const void *a2)
 {
-#define ace(i) ((const __aclent16_t *) a##i)
-  int ret = ace(1)->a_type - ace(2)->a_type;
+#define ace(i) ((const __aclent32_t *) a##i)
+  int ret = ace (1)->a_type - ace (2)->a_type;
   if (!ret)
-    ret = ace(1)->a_id - ace(2)->a_id;
+    ret = ace (1)->a_id - ace (2)->a_id;
   return ret;
 #undef ace
 }
 
-extern "C"
-int
-aclsort (int nentries, int, __aclent16_t *aclbufp)
+extern "C" int
+aclsort32 (int nentries, int, __aclent32_t *aclbufp)
 {
-  if (aclcheck (aclbufp, nentries, NULL))
+  if (aclcheck32 (aclbufp, nentries, NULL))
     return -1;
   if (!aclbufp || nentries < 1)
     {
       set_errno (EINVAL);
       return -1;
     }
-  qsort((void *) aclbufp, nentries, sizeof (__aclent16_t), acecmp);
+  qsort ((void *) aclbufp, nentries, sizeof (__aclent32_t), acecmp);
   return 0;
 }
 
-extern "C"
-int
-acltomode (__aclent16_t *aclbufp, int nentries, mode_t *modep)
+extern "C" int
+acltomode32 (__aclent32_t *aclbufp, int nentries, mode_t *modep)
 {
   int pos;
 
@@ -768,32 +686,36 @@ acltomode (__aclent16_t *aclbufp, int nentries, mode_t *modep)
       return -1;
     }
   *modep = 0;
-  if ((pos = searchace (aclbufp, nentries, USER_OBJ)) < 0)
+  if ((pos = searchace (aclbufp, nentries, USER_OBJ)) < 0
+      || !aclbufp[pos].a_type)
     {
       set_errno (EINVAL);
       return -1;
     }
-  *modep |= aclbufp[pos].a_perm & S_IRWXU;
-  if ((pos = searchace (aclbufp, nentries, GROUP_OBJ)) < 0)
+  *modep |= (aclbufp[pos].a_perm & S_IRWXO) << 6;
+  if ((pos = searchace (aclbufp, nentries, GROUP_OBJ)) < 0
+      || !aclbufp[pos].a_type)
     {
       set_errno (EINVAL);
       return -1;
     }
-  if (searchace (aclbufp, nentries, CLASS_OBJ) < 0)
-    pos = searchace (aclbufp, nentries, CLASS_OBJ);
-  *modep |= (aclbufp[pos].a_perm & S_IRWXU) >> 3;
-  if ((pos = searchace (aclbufp, nentries, OTHER_OBJ)) < 0)
+  *modep |= (aclbufp[pos].a_perm & S_IRWXO) << 3;
+  int cpos;
+  if ((cpos = searchace (aclbufp, nentries, CLASS_OBJ)) >= 0
+      && aclbufp[cpos].a_type == CLASS_OBJ)
+    *modep |= ((aclbufp[pos].a_perm & S_IRWXO) & aclbufp[cpos].a_perm) << 3;
+  if ((pos = searchace (aclbufp, nentries, OTHER_OBJ)) < 0
+      || !aclbufp[pos].a_type)
     {
       set_errno (EINVAL);
       return -1;
     }
-  *modep |= (aclbufp[pos].a_perm & S_IRWXU) >> 6;
+  *modep |= aclbufp[pos].a_perm & S_IRWXO;
   return 0;
 }
 
-extern "C"
-int
-aclfrommode(__aclent16_t *aclbufp, int nentries, mode_t *modep)
+extern "C" int
+aclfrommode32 (__aclent32_t *aclbufp, int nentries, mode_t *modep)
 {
   int pos;
 
@@ -802,47 +724,43 @@ aclfrommode(__aclent16_t *aclbufp, int nentries, mode_t *modep)
       set_errno (EINVAL);
       return -1;
     }
-  if ((pos = searchace (aclbufp, nentries, USER_OBJ)) < 0)
+  if ((pos = searchace (aclbufp, nentries, USER_OBJ)) < 0
+      || !aclbufp[pos].a_type)
     {
       set_errno (EINVAL);
       return -1;
     }
-  aclbufp[pos].a_perm = (*modep & S_IRWXU)
-			| (*modep & S_IRWXU) >> 3
-			| (*modep & S_IRWXU) >> 6;
-  if ((pos = searchace (aclbufp, nentries, GROUP_OBJ)) < 0)
+  aclbufp[pos].a_perm = (*modep & S_IRWXU) >> 6;
+  if ((pos = searchace (aclbufp, nentries, GROUP_OBJ)) < 0
+      || !aclbufp[pos].a_type)
     {
       set_errno (EINVAL);
       return -1;
     }
-  if (searchace (aclbufp, nentries, CLASS_OBJ) < 0)
-    pos = searchace (aclbufp, nentries, CLASS_OBJ);
-  aclbufp[pos].a_perm = (*modep & S_IRWXG)
-			| (*modep & S_IRWXG) << 3
-			| (*modep & S_IRWXG) >> 3;
-  if ((pos = searchace (aclbufp, nentries, OTHER_OBJ)) < 0)
+  aclbufp[pos].a_perm = (*modep & S_IRWXG) >> 3;
+  if ((pos = searchace (aclbufp, nentries, CLASS_OBJ)) >= 0
+      && aclbufp[pos].a_type == CLASS_OBJ)
+    aclbufp[pos].a_perm = (*modep & S_IRWXG) >> 3;
+  if ((pos = searchace (aclbufp, nentries, OTHER_OBJ)) < 0
+      || !aclbufp[pos].a_type)
     {
       set_errno (EINVAL);
       return -1;
     }
-  aclbufp[pos].a_perm = (*modep & S_IRWXO)
-			| (*modep & S_IRWXO) << 6
-			| (*modep & S_IRWXO) << 3;
+  aclbufp[pos].a_perm = (*modep & S_IRWXO);
   return 0;
 }
 
-extern "C"
-int
-acltopbits (__aclent16_t *aclbufp, int nentries, mode_t *pbitsp)
+extern "C" int
+acltopbits32 (__aclent32_t *aclbufp, int nentries, mode_t *pbitsp)
 {
-  return acltomode (aclbufp, nentries, pbitsp);
+  return acltomode32 (aclbufp, nentries, pbitsp);
 }
 
-extern "C"
-int
-aclfrompbits (__aclent16_t *aclbufp, int nentries, mode_t *pbitsp)
+extern "C" int
+aclfrompbits32 (__aclent32_t *aclbufp, int nentries, mode_t *pbitsp)
 {
-  return aclfrommode (aclbufp, nentries, pbitsp);
+  return aclfrommode32 (aclbufp, nentries, pbitsp);
 }
 
 static char *
@@ -850,19 +768,18 @@ permtostr (mode_t perm)
 {
   static char pbuf[4];
 
-  pbuf[0] = (perm & S_IREAD) ? 'r' : '-';
-  pbuf[1] = (perm & S_IWRITE) ? 'w' : '-';
-  pbuf[2] = (perm & S_IEXEC) ? 'x' : '-';
+  pbuf[0] = (perm & S_IROTH) ? 'r' : '-';
+  pbuf[1] = (perm & S_IWOTH) ? 'w' : '-';
+  pbuf[2] = (perm & S_IXOTH) ? 'x' : '-';
   pbuf[3] = '\0';
   return pbuf;
 }
 
-extern "C"
-char *
-acltotext (__aclent16_t *aclbufp, int aclcnt)
+extern "C" char *
+acltotext32 (__aclent32_t *aclbufp, int aclcnt)
 {
   if (!aclbufp || aclcnt < 1 || aclcnt > MAX_ACL_ENTRIES
-      || aclcheck (aclbufp, aclcnt, NULL))
+      || aclcheck32 (aclbufp, aclcnt, NULL))
     {
       set_errno (EINVAL);
       return NULL;
@@ -920,23 +837,22 @@ permfromstr (char *perm)
   if (strlen (perm) != 3)
     return 01000;
   if (perm[0] == 'r')
-    mode |= S_IRUSR | S_IRGRP | S_IROTH;
+    mode |= S_IROTH;
   else if (perm[0] != '-')
     return 01000;
   if (perm[1] == 'w')
-    mode |= S_IWUSR | S_IWGRP | S_IWOTH;
+    mode |= S_IWOTH;
   else if (perm[1] != '-')
     return 01000;
   if (perm[2] == 'x')
-    mode |= S_IXUSR | S_IXGRP | S_IXOTH;
+    mode |= S_IXOTH;
   else if (perm[2] != '-')
     return 01000;
   return mode;
 }
 
-extern "C"
-__aclent16_t *
-aclfromtext (char *acltextp, int *)
+extern "C" __aclent32_t *
+aclfromtext32 (char *acltextp, int *)
 {
   if (!acltextp)
     {
@@ -944,7 +860,7 @@ aclfromtext (char *acltextp, int *)
       return NULL;
     }
   char buf[strlen (acltextp) + 1];
-  __aclent16_t lacl[MAX_ACL_ENTRIES];
+  __aclent32_t lacl[MAX_ACL_ENTRIES];
   memset (lacl, 0, sizeof lacl);
   int pos = 0;
   strcpy (buf, acltextp);
@@ -968,18 +884,18 @@ aclfromtext (char *acltextp, int *)
 	      c += 5;
 	      if (isalpha (*c))
 		{
-		  struct passwd *pw = getpwnam (c);
+		  struct passwd *pw = internal_getpwnam (c);
 		  if (!pw)
 		    {
 		      set_errno (EINVAL);
 		      return NULL;
 		    }
 		  lacl[pos].a_id = pw->pw_uid;
-		  c = strchr (c, ':');
+		  c = strechr (c, ':');
 		}
 	      else if (isdigit (*c))
 		lacl[pos].a_id = strtol (c, &c, 10);
-	      if (!c || *c != ':')
+	      if (*c != ':')
 		{
 		  set_errno (EINVAL);
 		  return NULL;
@@ -996,18 +912,18 @@ aclfromtext (char *acltextp, int *)
 	      c += 5;
 	      if (isalpha (*c))
 		{
-		  struct __group16 *gr = getgrnam (c);
+		  struct __group32 *gr = internal_getgrnam (c);
 		  if (!gr)
 		    {
 		      set_errno (EINVAL);
 		      return NULL;
 		    }
 		  lacl[pos].a_id = gr->gr_gid;
-		  c = strchr (c, ':');
+		  c = strechr (c, ':');
 		}
 	      else if (isdigit (*c))
 		lacl[pos].a_id = strtol (c, &c, 10);
-	      if (!c || *c != ':')
+	      if (*c != ':')
 		{
 		  set_errno (EINVAL);
 		  return NULL;
@@ -1041,9 +957,86 @@ aclfromtext (char *acltextp, int *)
 	}
       ++pos;
     }
-  __aclent16_t *aclp = (__aclent16_t *) malloc (pos * sizeof (__aclent16_t));
+  __aclent32_t *aclp = (__aclent32_t *) malloc (pos * sizeof (__aclent32_t));
   if (aclp)
-    memcpy (aclp, lacl, pos * sizeof (__aclent16_t));
+    memcpy (aclp, lacl, pos * sizeof (__aclent32_t));
   return aclp;
 }
 
+/* __aclent16_t and __aclent32_t have same size and same member offsets */
+static __aclent32_t *
+acl16to32 (__aclent16_t *aclbufp, int nentries)
+{
+  __aclent32_t *aclbufp32 = (__aclent32_t *) aclbufp;
+  if (aclbufp32)
+    for (int i = 0; i < nentries; i++)
+      aclbufp32[i].a_id &= USHRT_MAX;
+  return aclbufp32;
+}
+
+extern "C" int
+acl (const char *path, int cmd, int nentries, __aclent16_t *aclbufp)
+{
+  return acl32 (path, cmd, nentries, acl16to32 (aclbufp, nentries));
+}
+
+extern "C" int
+facl (int fd, int cmd, int nentries, __aclent16_t *aclbufp)
+{
+  return facl32 (fd, cmd, nentries, acl16to32 (aclbufp, nentries));
+}
+
+extern "C" int
+lacl (const char *path, int cmd, int nentries, __aclent16_t *aclbufp)
+{
+  return lacl32 (path, cmd, nentries, acl16to32 (aclbufp, nentries));
+}
+
+extern "C" int
+aclcheck (__aclent16_t *aclbufp, int nentries, int *which)
+{
+  return aclcheck32 (acl16to32 (aclbufp, nentries), nentries, which);
+}
+
+extern "C" int
+aclsort (int nentries, int i, __aclent16_t *aclbufp)
+{
+  return aclsort32 (nentries, i, acl16to32 (aclbufp, nentries));
+}
+
+
+extern "C" int
+acltomode (__aclent16_t *aclbufp, int nentries, mode_t *modep)
+{
+  return acltomode32 (acl16to32 (aclbufp, nentries), nentries, modep);
+}
+
+extern "C" int
+aclfrommode (__aclent16_t *aclbufp, int nentries, mode_t *modep)
+{
+  return aclfrommode32 ((__aclent32_t *)aclbufp, nentries, modep);
+}
+
+extern "C" int
+acltopbits (__aclent16_t *aclbufp, int nentries, mode_t *pbitsp)
+{
+  return acltopbits32 (acl16to32 (aclbufp, nentries), nentries, pbitsp);
+}
+
+extern "C" int
+aclfrompbits (__aclent16_t *aclbufp, int nentries, mode_t *pbitsp)
+{
+  return aclfrompbits32 ((__aclent32_t *)aclbufp, nentries, pbitsp);
+}
+
+extern "C" char *
+acltotext (__aclent16_t *aclbufp, int aclcnt)
+{
+  return acltotext32 (acl16to32 (aclbufp, aclcnt), aclcnt);
+}
+
+extern "C" __aclent16_t *
+aclfromtext (char *acltextp, int * aclcnt)
+{
+  return (__aclent16_t *) aclfromtext32 (acltextp, aclcnt);
+}
