@@ -17,6 +17,7 @@ details. */
 #include <errno.h>
 #include <stdlib.h>
 #include <sys/cygwin.h>
+#include <assert.h>
 #include "cygerrno.h"
 #include "sync.h"
 #include "sigproc.h"
@@ -29,8 +30,8 @@ details. */
 #include "child_info_magic.h"
 #define NEED_VFORK
 #include "perthread.h"
-#include <assert.h>
 #include "shared_info.h"
+#include "cygthread.h"
 
 /*
  * Convenience defines
@@ -40,8 +41,6 @@ details. */
 #define WSPX		   20000 // Wait for wait_sig to terminate
 #define WWSP		   20000 // Wait for wait_subproc to terminate
 
-#define WAIT_SIG_PRIORITY		THREAD_PRIORITY_NORMAL
-
 #define TOTSIGS	(NSIG + __SIGOFFSET)
 
 #define wake_wait_subproc() SetEvent (events[0])
@@ -79,34 +78,35 @@ HANDLE NO_COPY signal_arrived;		// Event signaled when a signal has
 Static DWORD proc_loop_wait = 1000;	// Wait for subprocesses to exit
 Static DWORD sig_loop_wait = INFINITE;	// Wait for signals to arrive
 
-Static HANDLE sigcatch_nonmain = NULL;	// The semaphore signaled when
+Static HANDLE sigcatch_nonmain;		// The semaphore signaled when
 					//  signals are available for
 					//  processing from non-main thread
-Static HANDLE sigcatch_main = NULL;	// Signalled when main thread sends a
+Static HANDLE sigcatch_main;		// Signalled when main thread sends a
 					//  signal
-Static HANDLE sigcatch_nosync = NULL;	// Signal wait_sig to scan sigtodo
+Static HANDLE sigcatch_nosync;		// Signal wait_sig to scan sigtodo
 					//  but not to bother with any
 					//  synchronization
-Static HANDLE sigcomplete_main = NULL;	// Event signaled when a signal has
+Static HANDLE sigcomplete_main;		// Event signaled when a signal has
 					//  finished processing for the main
 					//  thread
-Static HANDLE sigcomplete_nonmain = NULL;// Semaphore raised for non-main
+Static HANDLE sigcomplete_nonmain;	// Semaphore raised for non-main
 					//  threads when a signal has finished
 					//  processing
-Static HANDLE hwait_sig = NULL;		// Handle of wait_sig thread
-Static HANDLE hwait_subproc = NULL;	// Handle of sig_subproc thread
+HANDLE NO_COPY sigCONT;			// Used to "STOP" a process
+Static cygthread *hwait_sig;		// Handle of wait_sig thread
+Static cygthread *hwait_subproc;	// Handle of sig_subproc thread
 
-Static HANDLE wait_sig_inited = NULL;	// Control synchronization of
+Static HANDLE wait_sig_inited;		// Control synchronization of
 					//  message queue startup
 
 /* Used by WaitForMultipleObjects.  These are handles to child processes.
  */
-Static HANDLE events[PSIZE + 1] = {0};  // All my children's handles++
+Static HANDLE events[PSIZE + 1];	  // All my children's handles++
 #define hchildren (events + 1)		// Where the children handles begin
 Static char cpchildren[PSIZE * sizeof (pinfo)];		// All my children info
-Static int nchildren = 0;		// Number of active children
-Static char czombies[NZOMBIES * sizeof (pinfo)];		// All my deceased children info
-Static int nzombies = 0;		// Number of deceased children
+Static int nchildren;			// Number of active children
+Static char czombies[(NZOMBIES + 1) * sizeof (pinfo)];		// All my deceased children info
+Static int nzombies;			// Number of deceased children
 
 #define pchildren ((pinfo *) cpchildren)
 #define zombies ((pinfo *) czombies)
@@ -167,19 +167,15 @@ out:
   return res;
 }
 
-__inline static void
-wait_for_me ()
+void __stdcall
+wait_for_sigthread ()
 {
-  /* See if this is the first signal call after initialization.
-   * If so, wait for notification that all initialization has completed.
-   * Then set the handle to NULL to avoid checking this again.
-   */
-  if (wait_sig_inited)
-    {
-      (void) WaitForSingleObject (wait_sig_inited, INFINITE);
-      (void) ForceCloseHandle (wait_sig_inited);
-      wait_sig_inited = NULL;
-    }
+  sigproc_printf ("wait_sig_inited %p", wait_sig_inited);
+  HANDLE hsig_inited = wait_sig_inited;
+  assert (hsig_inited);
+  (void) WaitForSingleObject (hsig_inited, INFINITE);
+  wait_sig_inited = NULL;
+  (void) ForceCloseHandle1 (hsig_inited, wait_sig_inited);
 }
 
 /* Get the sync_proc_subproc muto to control access to
@@ -210,7 +206,7 @@ proc_can_be_signalled (_pinfo *p)
 {
   if (p == myself_nowait || p == myself_nowait_nonmain || p == myself)
     {
-      wait_for_me ();
+      assert (!wait_sig_inited);
       return 1;
     }
 
@@ -460,22 +456,11 @@ proc_terminate (void)
   /* Signal processing is assumed to be blocked in this routine. */
   if (hwait_subproc)
     {
-      int rc;
       proc_loop_wait = 0;	// Tell wait_subproc thread to exit
+      sync_proc_subproc->acquire (WPSP);
       wake_wait_subproc ();	// Wake wait_subproc loop
-
-      /* Wait for wait_subproc thread to exit (but not *too* long) */
-      if ((rc = WaitForSingleObject (hwait_subproc, WWSP)) != WAIT_OBJECT_0)
-	if (rc == WAIT_TIMEOUT)
-	  system_printf ("WFSO(hwait_subproc) timed out");
-	else
-	  system_printf ("WFSO(hwait_subproc), rc %d, %E", rc);
-
-      HANDLE h = hwait_subproc;
       hwait_subproc = NULL;
-      ForceCloseHandle1 (h, hwait_subproc);
 
-      sync_proc_subproc->acquire(WPSP);
       (void) proc_subproc (PROC_CLEARWAIT, 1);
 
       /* Clean out zombie processes from the pid list. */
@@ -487,8 +472,9 @@ proc_terminate (void)
 	      ForceCloseHandle1 (zombies[i]->hProcess, childhProc);
 	      ForceCloseHandle1 (zombies[i]->pid_handle, pid_handle);
 	    }
+	  zombies[i]->ppid = 1;
 	  zombies[i]->process_state = PID_EXITED;	/* CGF FIXME - still needed? */
-	  zombies[i].release();		// FIXME: this breaks older gccs for some reason
+	  zombies[i].release ();	// FIXME: this breaks older gccs for some reason
 	}
 
       /* Disassociate my subprocesses */
@@ -532,6 +518,12 @@ sig_dispatch_pending (int justwake)
 {
   if (!hwait_sig)
     return 0;
+  DWORD tid = GetCurrentThreadId ();
+
+  sigframe thisframe (mainthread);
+
+  if (tid == sigtid && !justwake)
+    justwake = 1;
 
   int was_pending = pending_signals;
 #ifdef DEBUGGING
@@ -545,7 +537,6 @@ sig_dispatch_pending (int justwake)
 #endif
   else
     {
-      wait_for_me ();
       if (!justwake)
 	(void) sig_send (myself, __SIGFLUSH);
       else if (ReleaseSemaphore (sigcatch_nosync, 1, NULL))
@@ -560,6 +551,9 @@ sig_dispatch_pending (int justwake)
 	system_printf ("%E releasing sigcatch_nosync(%p)", sigcatch_nosync);
 
     }
+  if (was_pending && !justwake)
+    thisframe.call_signal_handler ();
+
   return was_pending;
 }
 
@@ -576,24 +570,19 @@ sigproc_init ()
   wait_sig_inited = CreateEvent (&sec_none_nih, TRUE, FALSE, NULL);
   ProtectHandle (wait_sig_inited);
 
-  /* local event signaled when main thread has been dispatched
-     to a signal handler function. */
-  signal_arrived = CreateEvent(&sec_none_nih, TRUE, FALSE, NULL);
-  ProtectHandle (signal_arrived);
-
-  if (!(hwait_sig = makethread (wait_sig, NULL, 0, "sig")))
-    {
-      system_printf ("cannot create wait_sig thread, %E");
-      api_fatal ("terminating");
-    }
-
-  ProtectHandle (hwait_sig);
-
   /* sync_proc_subproc is used by proc_subproc.  It serialises
    * access to the children and zombie arrays.
    */
   new_muto (sync_proc_subproc);
 
+  /* local event signaled when main thread has been dispatched
+     to a signal handler function. */
+  signal_arrived = CreateEvent (&sec_none_nih, TRUE, FALSE, NULL);
+  ProtectHandle (signal_arrived);
+
+  hwait_sig = new cygthread (wait_sig, cygself, "sig");
+  hwait_sig->zap_h ();
+
   /* Initialize waitq structure for main thread.  A waitq structure is
    * allocated for each thread that executes a wait to allow multiple threads
    * to perform waits.  Pre-allocate a waitq structure for the main thread.
@@ -616,9 +605,18 @@ sigproc_init ()
 void __stdcall
 sigproc_terminate (void)
 {
-  HANDLE h = hwait_sig;
   hwait_sig = NULL;
 
+  if (!sig_loop_wait)
+    sigproc_printf ("sigproc handling not active");
+  else
+    {
+      sigproc_printf ("entering");
+      sig_loop_wait = 0;	// Tell wait_sig to exit when it is
+				//  finished with anything it is doing
+      sig_dispatch_pending (1);
+    }
+
   if (GetCurrentThreadId () == sigtid)
     {
       ForceCloseHandle (sigcomplete_main);
@@ -631,45 +629,6 @@ sigproc_terminate (void)
     }
   proc_terminate ();		// Terminate process handling thread
 
-  if (!sig_loop_wait)
-    sigproc_printf ("sigproc_terminate: sigproc handling not active");
-  else
-    {
-      sigproc_printf ("entering");
-      sig_loop_wait = 0;	// Tell wait_sig to exit when it is
-				//  finished with anything it is doing
-      // sig_dispatch_pending (TRUE);	// wake up and die
-      /* In case of a sigsuspend */
-      // SetEvent (signal_arrived);
-
-      /* If !hwait_sig, then the process probably hasn't even finished
-       * its initialization phase.
-       */
-      if (0 && hwait_sig)
-	{
-	  if (GetCurrentThreadId () != sigtid)
-	    WaitForSingleObject (h, 10000);
-	  ForceCloseHandle1 (h, hwait_sig);
-
-
-	  if (GetCurrentThreadId () != sigtid)
-	    {
-	      ForceCloseHandle (sigcomplete_main);
-	      ForceCloseHandle (sigcomplete_nonmain);
-	      ForceCloseHandle (sigcatch_main);
-	      ForceCloseHandle (sigcatch_nonmain);
-	      ForceCloseHandle (sigcatch_nosync);
-	    }
-	}
-      sigproc_printf ("done");
-    }
-
-#if 0
-  /* Set this so that subsequent tests will succeed. */
-  if (!myself->dwProcessId)
-    myself->dwProcessId = GetCurrentProcessId ();
-#endif
-
   return;
 }
 
@@ -697,7 +656,8 @@ sig_send (_pinfo *p, int sig, DWORD ebp, bool exception)
     {
       if (no_signals_available ())
 	goto out;		// Either exiting or not yet initializing
-      wait_for_me ();
+      if (wait_sig_inited)
+	wait_for_sigthread ();
       wait_for_completion = p != myself_nowait;
       p = myself;
     }
@@ -848,11 +808,10 @@ subproc_init (void)
    * the hchildren array.
    */
   events[0] = CreateEvent (&sec_none_nih, FALSE, FALSE, NULL);
-  if (!(hwait_subproc = makethread (wait_subproc, NULL, 0, "proc")))
-    system_printf ("cannot create wait_subproc thread, %E");
+  hwait_subproc = new cygthread (wait_subproc, NULL, "proc");
+  hwait_subproc->zap_h ();
   ProtectHandle (events[0]);
-  ProtectHandle (hwait_subproc);
-  sigproc_printf ("started wait_subproc thread %p", hwait_subproc);
+  sigproc_printf ("started wait_subproc thread");
 }
 
 /* Initialize some of the memory block passed to child processes
@@ -870,6 +829,7 @@ init_child_info (DWORD chtype, child_info *ch, pid_t pid, HANDLE subproc_ready)
   ch->subproc_ready = subproc_ready;
   ch->pppid_handle = myself->ppid_handle;
   ch->fhandler_union_cb = sizeof (fhandler_union);
+  ch->mount_h = cygwin_mount_h;
 }
 
 /* Check the state of all of our children to see if any are stopped or
@@ -935,7 +895,7 @@ getsem (_pinfo *p, const char *str, int init, int max)
       sigproc_printf ("pid %d, ppid %d, wait %d, initializing %x", p->pid, p->ppid, wait,
 		  ISSTATE (p, PID_INITIALIZING));
       for (int i = 0; ISSTATE (p, PID_INITIALIZING) && i < wait; i++)
-	Sleep (1);
+	low_priority_sleep (1);
     }
 
   SetLastError (0);
@@ -1063,6 +1023,15 @@ stopped_or_terminated (waitq *parent_w, _pinfo *child)
   return -potential_match;
 }
 
+static void
+talktome ()
+{
+  winpids pids;
+  for (unsigned i = 0; i < pids.npids; i++)
+    if (pids[i]->hello_pid == myself->pid)
+      pids[i]->commune_recv ();
+}
+
 /* Process signals by waiting for a semaphore to become signaled.
  * Then scan an in-memory array representing queued signals.
  * Executes in a separate thread.
@@ -1072,10 +1041,10 @@ stopped_or_terminated (waitq *parent_w, _pinfo *child)
  * has been handled, as per POSIX.
  */
 static DWORD WINAPI
-wait_sig (VOID *)
+wait_sig (VOID *self)
 {
   /* Initialization */
-  (void) SetThreadPriority (hwait_sig, WAIT_SIG_PRIORITY);
+  (void) SetThreadPriority (GetCurrentThread (), WAIT_SIG_PRIORITY);
 
   /* sigcatch_nosync       - semaphore incremented by sig_dispatch_pending and
    *			     by foreign processes to force an examination of
@@ -1094,6 +1063,7 @@ wait_sig (VOID *)
   sigcomplete_nonmain = CreateSemaphore (&sec_none_nih, 0, MAXLONG, NULL);
   sigcomplete_main = CreateEvent (&sec_none_nih, FALSE, FALSE, NULL);
   sigproc_printf ("sigcatch_nonmain %p, sigcatch_main %p", sigcatch_nonmain, sigcatch_main);
+  sigCONT = CreateEvent (&sec_none_nih, FALSE, FALSE, NULL);
 
   /* Setting dwProcessId flags that this process is now capable of receiving
    * signals.  Prior to this, dwProcessId was set to the windows pid of
@@ -1119,7 +1089,7 @@ wait_sig (VOID *)
       debug_printf ("subproc_ready %p", child_proc_info->subproc_ready);
       if (!SetEvent (child_proc_info->subproc_ready))
 	system_printf ("SetEvent (subproc_ready) failed, %E");
-      ForceCloseHandle (child_proc_info->subproc_ready);
+      ForceCloseHandle1 (child_proc_info->subproc_ready, subproc_ready);
       /* Initialize an "indirect" pid block so that if someone looks up this
 	 process via its Windows PID it will be redirected to the appropriate
 	 Cygwin PID shared memory block. */
@@ -1135,6 +1105,7 @@ wait_sig (VOID *)
   for (;;)
     {
       DWORD rc = WaitForMultipleObjects (3, catchem, FALSE, sig_loop_wait);
+      (void) SetThreadPriority (GetCurrentThread (), WAIT_SIG_PRIORITY);
 
       /* sigproc_terminate sets sig_loop_wait to zero to indicate that
        * this thread should terminate.
@@ -1162,7 +1133,6 @@ wait_sig (VOID *)
       pending_signals = -1;
       int saw_pending_signals = 0;
       int saw_sigchld = 0;
-      int dispatched_sigchld = 0;
       for (int sig = -__SIGOFFSET; sig < NSIG; sig++)
 	{
 	  while (InterlockedDecrement (myself->getsigtodo (sig)) >= 0)
@@ -1192,12 +1162,14 @@ wait_sig (VOID *)
 		  strace.hello ();
 		  break;
 
+		case __SIGCOMMUNE:
+		  talktome ();
+		  break;
+
 		/* A normal UNIX signal */
 		default:
 		  sigproc_printf ("Got signal %d", sig);
-		  int wasdispatched = sig_handle (sig, rc != 2);
-		  if (sig == SIGCHLD && wasdispatched)
-		    dispatched_sigchld = 1;
+		  sig_handle (sig, rc != 2);
 		  /* Need to decrement again to offset increment below since
 		     we really do want to decrement in this case. */
 		  InterlockedDecrement (myself->getsigtodo (sig));
@@ -1237,7 +1209,7 @@ wait_sig (VOID *)
     }
 
   sigproc_printf ("done");
-  return 0;
+  ExitThread (0);
 }
 
 /* Wait for subprocesses to terminate. Executes in a separate thread. */
@@ -1311,7 +1283,7 @@ wait_subproc (VOID *)
   ForceCloseHandle (events[0]);
   events[0] = NULL;
   sigproc_printf ("done");
-  return 0;
+  ExitThread (0);
 }
 
 extern "C" {
