@@ -21,12 +21,10 @@ details. */
 #include <ctype.h>
 #include "cygerrno.h"
 #include <sys/cygwin.h>
-#include "perprocess.h"
 #include "security.h"
 #include "fhandler.h"
 #include "path.h"
 #include "dtable.h"
-#include "sync.h"
 #include "sigproc.h"
 #include "cygheap.h"
 #include "child_info.h"
@@ -36,6 +34,7 @@ details. */
 #include "perthread.h"
 #include "registry.h"
 #include "environ.h"
+#include "cygthread.h"
 
 #define LINE_BUF_CHUNK (MAX_PATH * 2)
 
@@ -89,7 +88,7 @@ find_exec (const char *name, path_conv& buf, const char *mywinenv,
 {
   const char *suffix = "";
   debug_printf ("find_exec (%s)", name);
-  char *retval = buf;
+  const char *retval = buf;
   char tmp[MAX_PATH];
   const char *posix = (opt & FE_NATIVE) ? NULL : name;
   bool has_slash = strchr (name, '/');
@@ -166,6 +165,8 @@ find_exec (const char *name, path_conv& buf, const char *mywinenv,
     retval = NULL;
   else if (opt & FE_NATIVE)
     buf.check (name);
+  else
+    retval = name;
 
  out:
   if (posix)
@@ -179,7 +180,7 @@ find_exec (const char *name, path_conv& buf, const char *mywinenv,
 /* Utility for spawn_guts.  */
 
 static HANDLE
-handle (int n, int direction)
+handle (int n)
 {
   fhandler_base *fh = cygheap->fdtab[n];
 
@@ -187,9 +188,7 @@ handle (int n, int direction)
     return INVALID_HANDLE_VALUE;
   if (fh->get_close_on_exec ())
     return INVALID_HANDLE_VALUE;
-  if (direction == 0)
-    return fh->get_handle ();
-  return fh->get_output_handle ();
+  return n ? fh->get_output_handle () : fh->get_handle ();
 }
 
 int
@@ -210,7 +209,7 @@ class linebuf
   char *buf;
   size_t alloced;
   linebuf () : ix (0), buf (NULL), alloced (0) {}
-  ~linebuf () {/* if (buf) free (buf);*/}
+  ~linebuf () {if (buf) free (buf);}
   void add (const char *what, int len);
   void add (const char *what) {add (what, strlen (what));}
   void prepend (const char *what, int len);
@@ -254,17 +253,22 @@ class av
   char **argv;
   int calloced;
  public:
+  int error;
   int argc;
-  av (int ac, const char * const *av) : calloced (0), argc (ac)
+  av (int ac, const char * const *av) : calloced (0), error (false), argc (ac)
   {
     argv = (char **) cmalloc (HEAP_1_ARGV, (argc + 5) * sizeof (char *));
     memcpy (argv, av, (argc + 1) * sizeof (char *));
   }
   ~av ()
   {
-    for (int i = 0; i < calloced; i++)
-      cfree (argv[i]);
-    cfree (argv);
+    if (argv)
+      {
+	for (int i = 0; i < calloced; i++)
+	  if (argv[i])
+	    cfree (argv[i]);
+	cfree (argv);
+      }
   }
   int unshift (const char *what, int conv = 0);
   operator char **() {return argv;}
@@ -272,21 +276,23 @@ class av
   void replace0_maybe (const char *arg0)
   {
     /* Note: Assumes that argv array has not yet been "unshifted" */
-    if (!calloced)
-      {
-	argv[0] = cstrdup1 (arg0);
-	calloced = 1;
-      }
+    if (!calloced
+	&& (argv[0] = cstrdup1 (arg0)))
+      calloced = true;
+    else
+      error = errno;
   }
   void dup_maybe (int i)
   {
-    if (i >= calloced)
-      argv[i] = cstrdup1 (argv[i]);
+    if (i >= calloced
+	&& !(argv[i] = cstrdup1 (argv[i])))
+      error = errno;
   }
   void dup_all ()
   {
     for (int i = calloced; i < argc; i++)
-      argv[i] = cstrdup1 (argv[i]);
+      if (!(argv[i] = cstrdup1 (argv[i])))
+	error = errno;
   }
 };
 
@@ -309,14 +315,15 @@ av::unshift (const char *what, int conv)
 	*p = '\0';
       what = buf;
     }
-  *argv = cstrdup1 (what);
+  if (!(*argv = cstrdup1 (what)))
+    error = errno;
   argc++;
   calloced++;
   return 1;
 }
 
 static int __stdcall
-spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
+spawn_guts (const char * prog_arg, const char *const *argv,
 	    const char *const envp[], int mode)
 {
   BOOL rc;
@@ -338,7 +345,7 @@ spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
     {
       syscall_printf ("argv is NULL");
       set_errno (EINVAL);
-      return (-1);
+      return -1;
     }
 
   path_conv real_path;
@@ -357,17 +364,17 @@ spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
   else
     chtype = PROC_EXEC;
 
-  HANDLE spr;
+  HANDLE subproc_ready;
   if (chtype != PROC_EXEC)
-    spr = NULL;
+    subproc_ready = NULL;
   else
     {
-      spr = CreateEvent (&sec_all, TRUE, FALSE, NULL);
-      ProtectHandle (spr);
+      subproc_ready = CreateEvent (&sec_all, TRUE, FALSE, NULL);
+      ProtectHandleINH (subproc_ready);
     }
 
   init_child_info (chtype, &ciresrv, (mode == _P_OVERLAY) ? myself->pid : 1,
-		   spr);
+		   subproc_ready);
   if (!DuplicateHandle (hMainProc, hMainProc, hMainProc, &ciresrv.parent, 0, 1,
 			DUPLICATE_SAME_ACCESS))
      {
@@ -420,13 +427,10 @@ spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
      that it is NOT a script file */
   while (*ext == '\0')
     {
-      HANDLE hnd = CreateFileA (real_path,
-				GENERIC_READ,
-				FILE_SHARE_READ | FILE_SHARE_WRITE,
-				&sec_none_nih,
-				OPEN_EXISTING,
-				FILE_ATTRIBUTE_NORMAL,
-				0);
+      HANDLE hnd = CreateFile (real_path, GENERIC_READ,
+			       FILE_SHARE_READ | FILE_SHARE_WRITE,
+			       &sec_none_nih, OPEN_EXISTING,
+			       FILE_ATTRIBUTE_NORMAL, 0);
       if (hnd == INVALID_HANDLE_VALUE)
 	{
 	  __seterrno ();
@@ -562,18 +566,18 @@ spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
       MALLOC_CHECK;
     }
 
+  char *envblock;
   newargv.all_calloced ();
+  if (newargv.error)
+    {
+      set_errno (newargv.error);
+      return -1;
+    }
+
   ciresrv.moreinfo->argc = newargv.argc;
   ciresrv.moreinfo->argv = newargv;
-
-  ciresrv.moreinfo->envc = envsize (envp, 1);
-  ciresrv.moreinfo->envp = (char **) cmalloc (HEAP_1_ARGV, ciresrv.moreinfo->envc);
   ciresrv.hexec_proc = hexec_proc;
-  char **c;
-  const char * const *e;
-  for (c = ciresrv.moreinfo->envp, e = envp; *e;)
-    *c++ = cstrdup1 (*e++);
-  *c = NULL;
+
   if (mode != _P_OVERLAY ||
       !DuplicateHandle (hMainProc, myself.shared_handle (), hMainProc,
 			&ciresrv.moreinfo->myself_pinfo, 0,
@@ -585,15 +589,18 @@ spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
   si.lpReserved = NULL;
   si.lpDesktop = NULL;
   si.dwFlags = STARTF_USESTDHANDLES;
-  si.hStdInput = handle (0, 0); /* Get input handle */
-  si.hStdOutput = handle (1, 1); /* Get output handle */
-  si.hStdError = handle (2, 1); /* Get output handle */
+  si.hStdInput = handle (0);	/* Get input handle */
+  si.hStdOutput = handle (1);	/* Get output handle */
+  si.hStdError = handle (2);	/* Get output handle */
   si.cb = sizeof (si);
 
   int flags = CREATE_DEFAULT_ERROR_MODE | GetPriorityClass (hMainProc);
 
   if (mode == _P_DETACH || !set_console_state_for_spawn ())
-    flags |= DETACHED_PROCESS;
+    {
+      debug_printf ("mode %c= _P_DETACH", mode == _P_DETACH ? '=' : '!');
+      flags |= DETACHED_PROCESS;
+    }
   if (mode != _P_OVERLAY)
     flags |= CREATE_SUSPENDED;
 
@@ -605,70 +612,42 @@ spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
     flags |= CREATE_SUSPENDED;
 
 
-  /* Build windows style environment list */
-  char *envblock;
-  if (real_path.iscygexec ())
-    envblock = NULL;
-  else
-    envblock = winenv (envp, 0);
+  const char *runpath = null_app_name ? NULL : (const char *) real_path;
 
+  syscall_printf ("null_app_name %d (%s, %.132s)", null_app_name, runpath, one_line.buf);
+
+  void *newheap;
   /* Preallocated buffer for `sec_user' call */
   char sa_buf[1024];
 
-  if (!hToken && cygheap->user.impersonated
-      && cygheap->user.token != INVALID_HANDLE_VALUE)
-    hToken = cygheap->user.token;
-
-  const char *runpath = null_app_name ? NULL : (const char *) real_path;
-
-  syscall_printf ("spawn_guts null_app_name %d (%s, %.132s)", null_app_name, runpath, one_line.buf);
-
-  void *newheap;
   cygbench ("spawn-guts");
-  if (!hToken)
+
+  cygheap->fdtab.set_file_pointers_for_exec ();
+  if (!cygheap->user.issetuid ())
     {
-      ciresrv.moreinfo->uid = getuid ();
-      /* FIXME: This leaks a handle in the CreateProcessAsUser case since the
-	 child process doesn't know about cygwin_mount_h. */
-      ciresrv.mount_h = cygwin_mount_h;
+      PSECURITY_ATTRIBUTES sec_attribs = sec_user_nih (sa_buf);
+      ciresrv.moreinfo->envp = build_env (envp, envblock, ciresrv.moreinfo->envc,
+					  real_path.iscygexec ());
       newheap = cygheap_setup_for_child (&ciresrv, cygheap->fdtab.need_fixup_before ());
       rc = CreateProcess (runpath,	/* image name - with full path */
 			  one_line.buf,	/* what was passed to exec */
-					  /* process security attrs */
-			  sec_user_nih (sa_buf),
-					  /* thread security attrs */
-			  sec_user_nih (sa_buf),
-			  TRUE,	/* inherit handles from parent */
+			  sec_attribs,	/* process security attrs */
+			  sec_attribs,	/* thread security attrs */
+			  TRUE,		/* inherit handles from parent */
 			  flags,
-			  envblock,/* environment */
-			  0,	/* use current drive/directory */
+			  envblock,	/* environment */
+			  0,		/* use current drive/directory */
 			  &si,
 			  &pi);
     }
   else
     {
-      cygsid sid;
-      DWORD ret_len;
-      if (!GetTokenInformation (hToken, TokenUser, &sid, sizeof sid, &ret_len))
-	{
-	  sid = NO_SID;
-	  system_printf ("GetTokenInformation: %E");
-	}
-      /* Retrieve security attributes before setting psid to NULL
-	 since it's value is needed by `sec_user'. */
+      PSID sid = cygheap->user.sid ();
+
+      /* Set security attributes with sid */
       PSECURITY_ATTRIBUTES sec_attribs = sec_user_nih (sa_buf, sid);
 
-      /* Remove impersonation */
-      if (cygheap->user.impersonated
-	  && cygheap->user.token != INVALID_HANDLE_VALUE)
-	RevertToSelf ();
-
-      static BOOL first_time = TRUE;
-      if (first_time)
-	{
-	  set_process_privilege (SE_RESTORE_NAME);
-	  first_time = FALSE;
-	}
+      RevertToSelf ();
 
       /* Load users registry hive. */
       load_registry_hive (sid);
@@ -680,7 +659,6 @@ spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
       char wstname[1024];
       char dskname[1024];
 
-      ciresrv.moreinfo->uid = ILLEGAL_UID;
       hwst = GetProcessWindowStation ();
       SetUserObjectSecurity (hwst, &dsi, get_null_sd ());
       GetUserObjectInformation (hwst, UOI_NAME, wstname, 1024, &n);
@@ -691,23 +669,23 @@ spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
       strcat (wstname, dskname);
       si.lpDesktop = wstname;
 
+      ciresrv.moreinfo->envp = build_env (envp, envblock, ciresrv.moreinfo->envc,
+					  real_path.iscygexec ());
       newheap = cygheap_setup_for_child (&ciresrv, cygheap->fdtab.need_fixup_before ());
-      rc = CreateProcessAsUser (hToken,
+      rc = CreateProcessAsUser (cygheap->user.token,
 		       runpath,		/* image name - with full path */
 		       one_line.buf,	/* what was passed to exec */
 		       sec_attribs,     /* process security attrs */
 		       sec_attribs,     /* thread security attrs */
-		       TRUE,	/* inherit handles from parent */
+		       TRUE,		/* inherit handles from parent */
 		       flags,
-		       envblock,/* environment */
-		       0,	/* use current drive/directory */
+		       envblock,	/* environment */
+		       0,		/* use current drive/directory */
 		       &si,
 		       &pi);
       /* Restore impersonation. In case of _P_OVERLAY this isn't
 	 allowed since it would overwrite child data. */
-      if (mode != _P_OVERLAY
-	  && cygheap->user.impersonated
-	  && cygheap->user.token != INVALID_HANDLE_VALUE)
+      if (mode != _P_OVERLAY)
 	ImpersonateLoggedOnUser (cygheap->user.token);
     }
 
@@ -723,8 +701,8 @@ spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
     {
       __seterrno ();
       syscall_printf ("CreateProcess failed, %E");
-      if (spr)
-	ForceCloseHandle (spr);
+      if (subproc_ready)
+	ForceCloseHandle (subproc_ready);
       cygheap_setup_for_child_cleanup (newheap, &ciresrv, 0);
       return -1;
     }
@@ -738,13 +716,16 @@ spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
       cygheap->fdtab.fixup_before_exec (pi.dwProcessId);
       cygheap_setup_for_child_cleanup (newheap, &ciresrv, 1);
       if (mode == _P_OVERLAY)
-	ResumeThread (pi.hThread);
+	{
+	  ResumeThread (pi.hThread);
+	  cygthread::terminate ();
+	}
     }
 
-  if (mode == _P_OVERLAY)
-    cygpid = myself->pid;
-  else
+  if (mode != _P_OVERLAY)
     cygpid = cygwin_pid (pi.dwProcessId);
+  else
+    cygpid = myself->pid;
 
   /* We print the original program name here so the user can see that too.  */
   syscall_printf ("%d = spawn_guts (%s, %.132s)",
@@ -793,9 +774,6 @@ spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
 
   sigproc_printf ("spawned windows pid %d", pi.dwProcessId);
 
-  if (hToken && hToken != cygheap->user.token)
-    CloseHandle (hToken);
-
   DWORD res;
   BOOL exited;
 
@@ -805,7 +783,7 @@ spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
   if (mode == _P_OVERLAY)
     {
       int nwait = 3;
-      HANDLE waitbuf[3] = {pi.hProcess, signal_arrived, spr};
+      HANDLE waitbuf[3] = {pi.hProcess, signal_arrived, subproc_ready};
       for (int i = 0; i < 100; i++)
 	{
 	  switch (WaitForMultipleObjects (nwait, waitbuf, FALSE, INFINITE))
@@ -848,7 +826,7 @@ spawn_guts (HANDLE hToken, const char * prog_arg, const char *const *argv,
 	  break;
 	}
 
-      ForceCloseHandle (spr);
+      ForceCloseHandle (subproc_ready);
 
       sigproc_printf ("res = %x", res);
 
@@ -916,8 +894,8 @@ cwait (int *result, int pid, int)
  */
 
 extern "C" int
-_spawnve (HANDLE hToken, int mode, const char *path, const char *const *argv,
-	  const char *const *envp)
+spawnve (int mode, const char *path, const char *const *argv,
+	 const char *const *envp)
 {
   int ret;
   vfork_save *vf = vfork_storage.val ();
@@ -927,14 +905,14 @@ _spawnve (HANDLE hToken, int mode, const char *path, const char *const *argv,
   else
     vf = NULL;
 
-  syscall_printf ("_spawnve (%s, %s, %x)", path, argv[0], envp);
+  syscall_printf ("spawnve (%s, %s, %x)", path, argv[0], envp);
 
   switch (mode)
     {
     case _P_OVERLAY:
       /* We do not pass _P_SEARCH_PATH here. execve doesn't search PATH.*/
       /* Just act as an exec if _P_OVERLAY set. */
-      spawn_guts (hToken, path, argv, envp, mode);
+      spawn_guts (path, argv, envp, mode);
       /* Errno should be set by spawn_guts.  */
       ret = -1;
       break;
@@ -944,12 +922,14 @@ _spawnve (HANDLE hToken, int mode, const char *path, const char *const *argv,
     case _P_WAIT:
     case _P_DETACH:
       subproc_init ();
-      ret = spawn_guts (hToken, path, argv, envp, mode);
-      if (vf && ret > 0)
+      ret = spawn_guts (path, argv, envp, mode);
+      if (vf)
 	{
 	  debug_printf ("longjmping due to vfork");
-	  vf->pid = ret;
-	  longjmp (vf->j, 1);
+	  if (ret < 0)
+	    vf->restore_exit (ret);
+	  else
+	    vf->restore_pid (ret);
 	}
       break;
     default:
@@ -982,7 +962,7 @@ spawnl (int mode, const char *path, const char *arg0, ...)
 
   va_end (args);
 
-  return _spawnve (NULL, mode, path, (char * const  *) argv, cur_environ ());
+  return spawnve (mode, path, (char * const  *) argv, cur_environ ());
 }
 
 extern "C" int
@@ -1004,8 +984,7 @@ spawnle (int mode, const char *path, const char *arg0, ...)
   envp = va_arg (args, const char * const *);
   va_end (args);
 
-  return _spawnve (NULL, mode, path, (char * const *) argv,
-		   (char * const *) envp);
+  return spawnve (mode, path, (char * const *) argv, (char * const *) envp);
 }
 
 extern "C" int
@@ -1053,14 +1032,7 @@ spawnlpe (int mode, const char *path, const char *arg0, ...)
 extern "C" int
 spawnv (int mode, const char *path, const char * const *argv)
 {
-  return _spawnve (NULL, mode, path, argv, cur_environ ());
-}
-
-extern "C" int
-spawnve (int mode, const char *path, char * const *argv,
-					     const char * const *envp)
-{
-  return _spawnve (NULL, mode, path, argv, envp);
+  return spawnve (mode, path, argv, cur_environ ());
 }
 
 extern "C" int
@@ -1074,5 +1046,5 @@ spawnvpe (int mode, const char *file, const char * const *argv,
 					     const char * const *envp)
 {
   path_conv buf;
-  return _spawnve (NULL, mode, find_exec (file, buf), argv, envp);
+  return spawnve (mode, find_exec (file, buf), argv, envp);
 }
