@@ -1,6 +1,6 @@
 /* cygcheck.cc
 
-   Copyright 1998, 1999, 2000, 2001 Red Hat, Inc.
+   Copyright 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
 
    This file is part of Cygwin.
 
@@ -11,12 +11,14 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/time.h>
+#include <ctype.h>
+#include <io.h>
 #include <windows.h>
+#include "path.h"
 #include "cygwin/include/sys/cygwin.h"
 #include "cygwin/include/mntent.h"
 #include "cygwin/include/getopt.h"
-#include "time.h"
-#include "ctype.h"
 
 int verbose = 0;
 int registry = 0;
@@ -33,7 +35,9 @@ typedef __int64 longlong;
 
 void dump_setup (int, char **, bool);
 
-const char *known_env_vars[] = {
+static const char version[] = "$Revision: 1.32 $";
+
+static const char *known_env_vars[] = {
   "c_include_path",
   "compiler_path",
   "cxx_include_path",
@@ -63,35 +67,41 @@ struct
   const char *name;
   int missing_is_good;
 }
-common_apps[] =
-{
+static common_apps[] = {
+  {"awk", 0},
   {"bash", 0},
   {"cat", 0},
+  {"cp", 0},
   {"cpp", 1},
   {"find", 0},
   {"gcc", 0},
   {"gdb", 0},
+  {"grep", 0},
   {"ld", 0},
   {"ls", 0},
   {"make", 0},
+  {"mv", 0},
+  {"rm", 0},
+  {"sed", 0},
   {"sh", 0},
+  {"tar", 0},
   {0, 0}
 };
 
-int num_paths = 0, max_paths = 0;
-char **paths = 0;
+static int num_paths = 0, max_paths = 0;
+static char **paths = 0;
 
 /*
  * keyeprint() is used to report failure modes
  */
-int
+static int
 keyeprint (const char *name)
 {
   fprintf (stderr, "cygcheck: %s failed: %lu\n", name, GetLastError ());
   return 1;
 }
 
-void
+static void
 add_path (char *s, int maxlen)
 {
   if (num_paths >= max_paths)
@@ -119,7 +129,7 @@ add_path (char *s, int maxlen)
   num_paths++;
 }
 
-void
+static void
 init_paths ()
 {
   char tmp[4000], *sl;
@@ -156,14 +166,14 @@ init_paths ()
     printf ("WARNING: PATH is not set at all!\n");
 }
 
-char *
+static char *
 find_on_path (char *file, char *default_extension,
 	      int showall = 0, int search_sysdirs = 0)
 {
   static char rv[4000];
   char tmp[4000], *ptr = rv;
 
-  if (file == NULL)
+  if (!file)
     {
       keyeprint ("find_on_path: NULL pointer for file");
       return 0;
@@ -176,7 +186,7 @@ find_on_path (char *file, char *default_extension,
     }
 
   if (strchr (file, ':') || strchr (file, '\\') || strchr (file, '/'))
-    return file;
+    return cygpath (file, NULL);
 
   if (strchr (file, '.'))
     default_extension = (char *) "";
@@ -215,9 +225,9 @@ struct Did
   char *file;
   int state;
 };
-Did *did = 0;
+static Did *did = 0;
 
-Did *
+static Did *
 already_did (char *file)
 {
   Did *d;
@@ -232,7 +242,7 @@ already_did (char *file)
   return d;
 }
 
-int
+static int
 get_word (HANDLE fh, int offset)
 {
   short rv;
@@ -242,13 +252,13 @@ get_word (HANDLE fh, int offset)
       && GetLastError () != NO_ERROR)
     keyeprint ("get_word: SetFilePointer()");
 
-  if (!ReadFile (fh, &rv, 2, (DWORD *) & r, 0))
+  if (!ReadFile (fh, &rv, 2, (DWORD *) &r, 0))
     keyeprint ("get_word: Readfile()");
 
   return rv;
 }
 
-int
+static int
 get_dword (HANDLE fh, int offset)
 {
   int rv;
@@ -258,7 +268,7 @@ get_dword (HANDLE fh, int offset)
       && GetLastError () != NO_ERROR)
     keyeprint ("get_word: SetFilePointer()");
 
-  if (!ReadFile (fh, &rv, 4, (DWORD *) & r, 0))
+  if (!ReadFile (fh, &rv, 4, (DWORD *) &r, 0))
     keyeprint ("get_dword: Readfile()");
 
   return rv;
@@ -273,7 +283,7 @@ struct Section
   int pointer_to_raw_data;
 };
 
-int
+static int
 rva_to_offset (int rva, char *sections, int nsections, int *sz)
 {
   int i;
@@ -322,7 +332,7 @@ struct ImpDirectory
 };
 
 
-void track_down (char *file, char *suffix, int lvl);
+static void track_down (char *file, char *suffix, int lvl);
 
 #define CYGPREFIX (sizeof ("%%% Cygwin ") - 1)
 static void
@@ -380,7 +390,8 @@ cygwin_info (HANDLE h)
 	    int maj = atoi (dll_major);
 	    dll_major[1] = c;
 	    int min = atoi (dll_major + 1);
-	    sprintf (pbuf, "DLL version: %d.%d.%.*s", maj, min, len - 11, buf + 11);
+	    sprintf (pbuf, "DLL version: %d.%d.%.*s", maj, min, len - 11,
+		     buf + 11);
 	    len = strlen (s = pbuf);
 	  }
 	if (strncmp (s, "dll", 3) == 0)
@@ -400,7 +411,7 @@ cygwin_info (HANDLE h)
   return;
 }
 
-void
+static void
 dll_info (const char *path, HANDLE fh, int lvl, int recurse)
 {
   DWORD junk;
@@ -415,8 +426,8 @@ dll_info (const char *path, HANDLE fh, int lvl, int recurse)
       return;
     }
 
-  if (SetFilePointer (fh, opthdr_ofs + 40, 0, FILE_BEGIN) == INVALID_SET_FILE_POINTER
-      && GetLastError () != NO_ERROR)
+  if (SetFilePointer (fh, opthdr_ofs + 40, 0, FILE_BEGIN) ==
+      INVALID_SET_FILE_POINTER && GetLastError () != NO_ERROR)
     keyeprint ("dll_info: SetFilePointer()");
 
   if (!ReadFile (fh, &v, sizeof (v), &junk, 0))
@@ -453,8 +464,8 @@ dll_info (const char *path, HANDLE fh, int lvl, int recurse)
 
       if (expbase)
 	{
-	  if (SetFilePointer (fh, expbase, 0, FILE_BEGIN) == INVALID_SET_FILE_POINTER
-	      && GetLastError () != NO_ERROR)
+	  if (SetFilePointer (fh, expbase, 0, FILE_BEGIN) ==
+	      INVALID_SET_FILE_POINTER && GetLastError () != NO_ERROR)
 	    keyeprint ("dll_info: SetFilePointer()");
 
 	  unsigned char *exp = (unsigned char *) malloc (expsz);
@@ -484,8 +495,8 @@ dll_info (const char *path, HANDLE fh, int lvl, int recurse)
       int impbase = rva_to_offset (import_rva, sections, nsections, &impsz);
       if (impbase)
 	{
-	  if (SetFilePointer (fh, impbase, 0, FILE_BEGIN) == INVALID_SET_FILE_POINTER
-	      && GetLastError () != NO_ERROR)
+	  if (SetFilePointer (fh, impbase, 0, FILE_BEGIN) ==
+	      INVALID_SET_FILE_POINTER && GetLastError () != NO_ERROR)
 	    keyeprint ("dll_info: SetFilePointer()");
 
 	  unsigned char *imp = (unsigned char *) malloc (impsz);
@@ -511,7 +522,7 @@ dll_info (const char *path, HANDLE fh, int lvl, int recurse)
     cygwin_info (fh);
 }
 
-void
+static void
 track_down (char *file, char *suffix, int lvl)
 {
   if (file == NULL)
@@ -588,7 +599,7 @@ track_down (char *file, char *suffix, int lvl)
     keyeprint ("track_down: CloseHandle()");
 }
 
-void
+static void
 ls (char *f)
 {
   HANDLE h = CreateFile (f, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE,
@@ -610,7 +621,7 @@ ls (char *f)
     keyeprint ("ls: CloseHandle()");
 }
 
-void
+static void
 cygcheck (char *app)
 {
   char *papp = find_on_path (app, (char *) ".exe", 1, 0);
@@ -645,7 +656,7 @@ struct RegInfo
   HKEY key;
 };
 
-void
+static void
 show_reg (RegInfo * ri, int nest)
 {
   if (!ri)
@@ -657,7 +668,7 @@ show_reg (RegInfo * ri, int nest)
     printf ("%s\n", ri->name);
 }
 
-void
+static void
 scan_registry (RegInfo * prev, HKEY hKey, char *name, int cygnus)
 {
   RegInfo ri;
@@ -751,6 +762,74 @@ scan_registry (RegInfo * prev, HKEY hKey, char *name, int cygnus)
 }
 
 void
+pretty_id (const char *s, char *cygwin, size_t cyglen)
+{
+  char *groups[16384];
+
+  strcpy (cygwin + cyglen++, " ");
+  strcpy (cygwin + cyglen, s);
+  putenv (cygwin);
+
+  char *id = cygpath ("/bin/id.exe", NULL);
+  for (char *p = id; p = strchr (p, '/'); p++)
+    *p = '\\';
+
+  if (access (id, X_OK))
+    {
+      fprintf (stderr, "`id' program not found\n");
+      exit (1);
+    }
+
+  FILE *f = popen (id, "rt");
+
+  char buf[16384];
+  fgets (buf, sizeof (buf), f);
+  char *uid = strtok (buf, " ") + sizeof ("uid=") - 1;
+  char *gid = strtok (NULL, " ") + sizeof ("gid=") - 1;
+  char **ng;
+  size_t sz = 0;
+  for (ng = groups; (*ng = strtok (NULL, ",")); ng++)
+    {
+      char *p = strchr (*ng, '\n');
+      if (p)
+	*p = '\0';
+      if (ng == groups)
+	*ng += sizeof ("groups=") - 1;
+      size_t len = strlen (*ng);
+      if (sz < len)
+	sz = len;
+    }
+
+  printf ("\n%s output (%s)\n", id, s);
+  int szmaybe = sizeof ("UID: ") + strlen (uid) - 1;
+  if (sz < szmaybe)
+    sz = szmaybe;
+  sz += 1;
+  int n = 80 / (int) sz;
+  sz = -sz;
+  ng[0] += sizeof ("groups=") - 1;
+  printf ("UID: %*s GID: %s\n", sz + (sizeof ("UID: ") - 1), uid, gid);
+  int i = 0;
+  for (char **g = groups; g < ng; g++)
+    {
+      if (i < n)
+	i++;
+      else
+	{
+	  i = 0;
+	  puts ("");
+	}
+      if (++i <= n && g != (ng - 1))
+	printf ("%*s ", sz, *g);
+      else
+	{
+	  printf ("%s\n", *g);
+	  i = 0;
+	}
+    }
+}
+
+static void
 dump_sysinfo ()
 {
   int i, j;
@@ -776,13 +855,13 @@ dump_sysinfo ()
       switch (osversion.dwMinorVersion)
 	{
 	case 0:
-	  if (strchr(osversion.szCSDVersion, 'C'))
+	  if (strchr (osversion.szCSDVersion, 'C'))
 	    osname = (char *) "95 OSR2";
 	  else
 	    osname = (char *) "95";
 	  break;
 	case 10:
-	  if (strchr(osversion.szCSDVersion, 'A'))
+	  if (strchr (osversion.szCSDVersion, 'A'))
 	    osname = (char *) "98 SE";
 	  else
 	    osname = (char *) "98";
@@ -801,44 +880,46 @@ dump_sysinfo ()
 	  BOOL more_info = FALSE;
 	  OSVERSIONINFOEX osversionex;
 	  osversionex.dwOSVersionInfoSize = sizeof (osversionex);
-	  if (GetVersionEx ((OSVERSIONINFO *)&osversionex))
+	  if (GetVersionEx ((OSVERSIONINFO *) &osversionex))
 	    more_info = TRUE;
 	  if (osversion.dwMinorVersion == 0)
 	    {
 	      if (!more_info)
 		osname = (char *) "2000";
 	      else if (osversionex.wProductType == VER_NT_SERVER
-		       || osversionex.wProductType == VER_NT_DOMAIN_CONTROLLER)
-	        {
-		  if (osversionex.wSuiteMask & VER_SUITE_DATACENTER)
+		       || osversionex.wProductType ==
+		       VER_NT_DOMAIN_CONTROLLER)
+		{
+		  if (osversionex.wSuiteMask &VER_SUITE_DATACENTER)
 		    osname = (char *) "2000 Datacenter Server";
 		  else if (osversionex.wSuiteMask & VER_SUITE_ENTERPRISE)
 		    osname = (char *) "2000 Advanced Server";
 		  else
 		    osname = (char *) "2000 Server";
-	        }
+		}
 	      else
-	        osname = (char *) "2000 Professional";
+		osname = (char *) "2000 Professional";
 	    }
 	  else
 	    {
 	      if (!more_info)
 		osname = (char *) "XP";
 	      else if (osversionex.wProductType == VER_NT_SERVER
-		       || osversionex.wProductType == VER_NT_DOMAIN_CONTROLLER)
-	        {
+		       || osversionex.wProductType ==
+		       VER_NT_DOMAIN_CONTROLLER)
+		{
 		  if (osversionex.wSuiteMask & VER_SUITE_ENTERPRISE)
 		    osname = (char *) ".NET Enterprise Server";
 		  else
 		    osname = (char *) ".NET Server";
 		}
 	      else if (osversionex.wSuiteMask & VER_SUITE_PERSONAL)
-	        osname = (char *) "XP Home Edition";
+		osname = (char *) "XP Home Edition";
 	      else
-	        osname = (char *) "XP Professional";
-		
+		osname = (char *) "XP Professional";
+
 	    }
-        }
+	}
       else
 	osname = (char *) "NT";
       break;
@@ -849,9 +930,9 @@ dump_sysinfo ()
   printf ("Windows %s Ver %lu.%lu Build %lu %s\n\n", osname,
 	  osversion.dwMajorVersion, osversion.dwMinorVersion,
 	  osversion.dwPlatformId == VER_PLATFORM_WIN32_NT ?
-	    osversion.dwBuildNumber : (osversion.dwBuildNumber & 0xffff),
+	  osversion.dwBuildNumber : (osversion.dwBuildNumber & 0xffff),
 	  osversion.dwPlatformId == VER_PLATFORM_WIN32_NT ?
-	    osversion.szCSDVersion : "");
+	  osversion.szCSDVersion : "");
 
   printf ("Path:");
   char *s = getenv ("PATH"), *e;
@@ -867,6 +948,16 @@ dump_sysinfo ()
       s = e + 1;
     }
 
+  fflush (stdout);
+
+  char *cygwin = getenv ("CYGWIN") ?: const_cast <char *> ("CYGWIN=");
+  size_t cyglen = strlen (cygwin);
+  cygwin = strcpy ((char *) malloc (cyglen + sizeof (" nontsec")), cygwin);
+  pretty_id ("nontsec", cygwin, cyglen);
+  pretty_id ("ntsec", cygwin, cyglen);
+  cygwin[cyglen] = 0;
+  putenv (cygwin);
+
   if (!GetSystemDirectory (tmp, 4000))
     keyeprint ("dump_sysinfo: GetSystemDirectory()");
   printf ("\nSysDir: %s\n", tmp);
@@ -968,10 +1059,10 @@ dump_sysinfo ()
       sprintf (drive, "%c:\\", i + 'a');
       /* Report all errors, except if the Volume is ERROR_NOT_READY. 
          ERROR_NOT_READY is returned when removeable media drives are empty
-	 (CD, floppy, etc.) */
-      if (!GetVolumeInformation (drive, name, sizeof (name), &serno, &maxnamelen, &flags,
-				 fsname, sizeof (fsname))
-	  && GetLastError () != ERROR_NOT_READY)
+         (CD, floppy, etc.) */
+      if (!GetVolumeInformation
+	  (drive, name, sizeof (name), &serno, &maxnamelen, &flags, fsname,
+	   sizeof (fsname)) && GetLastError () != ERROR_NOT_READY)
 	keyeprint ("dump_sysinfo: GetVolumeInformation()");
 
       int dtype = GetDriveType (drive);
@@ -1131,7 +1222,7 @@ dump_sysinfo ()
     }
 }
 
-int
+static int
 check_keys ()
 {
   HANDLE h = CreateFileA ("CONIN$", GENERIC_READ | GENERIC_WRITE,
@@ -1215,18 +1306,20 @@ check_keys ()
   return 0;
 }
 
-void
-usage ()
+static void
+usage (FILE * stream, int status)
 {
-  fprintf (stderr, "Usage: cygcheck [OPTIONS] [program ...]\n");
-  fprintf (stderr, "  -c, --check-setup = check packages installed via setup.exe\n");
-  fprintf (stderr, "  -s, --sysinfo     = system information (not with -k)\n");
-  fprintf (stderr, "  -v, --verbose     = verbose output (indented) (for -s or programs)\n");
-  fprintf (stderr, "  -r, --registry    = registry search (requires -s)\n");
-  fprintf (stderr, "  -k, --keycheck    = perform a keyboard check session (not with -s)\n");
-  fprintf (stderr, "  -h, --help        = give help about the info (not with -c)\n");
-  fprintf (stderr, "You must at least give either -s or -k or a program name\n");
-  exit (1);
+  fprintf (stream, "\
+Usage: cygcheck [OPTIONS] [program ...]\n\
+ -c, --check-setup  check packages installed via setup.exe\n\
+ -s, --sysinfo      system information (not with -k)\n\
+ -v, --verbose      verbose output (indented) (for -s or programs)\n\
+ -r, --registry     registry search (requires -s)\n\
+ -k, --keycheck     perform a keyboard check session (not with -s)\n\
+ -h, --help         give help about the info (not with -c)\n\
+ -V, --version      output version information and exit\n\
+You must at least give either -s or -k or a program name\n");
+  exit (status);
 }
 
 struct option longopts[] = {
@@ -1236,10 +1329,34 @@ struct option longopts[] = {
   {"verbose", no_argument, NULL, 'v'},
   {"keycheck", no_argument, NULL, 'k'},
   {"help", no_argument, NULL, 'h'},
+  {"version", no_argument, 0, 'V'},
   {0, no_argument, NULL, 0}
 };
 
-char opts[] = "srvkhc";
+static char opts[] = "chkrsvV";
+
+static void
+print_version ()
+{
+  const char *v = strchr (version, ':');
+  int len;
+  if (!v)
+    {
+      v = "?";
+      len = 1;
+    }
+  else
+    {
+      v += 2;
+      len = strchr (v, ' ') - v;
+    }
+  printf ("\
+cygcheck version %.*s\n\
+System Checker for Cygwin\n\
+Copyright 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.\n\
+Compiled on %s\n\
+", len, v, __DATE__);
+}
 
 int
 main (int argc, char **argv)
@@ -1267,17 +1384,23 @@ main (int argc, char **argv)
       case 'h':
 	givehelp = 1;
 	break;
+      case 'V':
+	print_version ();
+	exit (0);
       default:
-	usage ();
+	usage (stderr, 1);
        /*NOTREACHED*/}
   argc -= optind;
   argv += optind;
 
   if (argc == 0 && !sysinfo && !keycheck && !check_setup)
-    usage ();
+    if (givehelp)
+      usage (stdout, 0);
+    else
+      usage (stderr, 1);
 
   if ((check_setup || sysinfo) && keycheck)
-    usage ();
+    usage (stderr, 1);
 
   if (keycheck)
     return check_keys ();
