@@ -1,6 +1,6 @@
 /* strace.cc
 
-   Copyright 2000, 2001, 2002 Red Hat Inc.
+   Copyright 2000, 2001, 2002, 2003 Red Hat Inc.
 
    Written by Chris Faylor <cgf@redhat.com>
 
@@ -23,6 +23,7 @@ details. */
 #include <errno.h>
 #include "cygwin/include/sys/strace.h"
 #include "cygwin/include/sys/cygwin.h"
+#include "path.h"
 #undef cygwin_internal
 
 /*  GCC runtime library's C++ EH code unfortunately pulls in stdio, and we
@@ -35,10 +36,10 @@ int _impure_ptr;
 #define alloca __builtin_alloca
 
 // Version string.
-static const char *version_string = "@(#)strace V1.0, Copyright (C) 2001, 2002 Red Hat Inc., " __DATE__ "\n";
+static const char version[] = "$Revision: 1.21 $";
 
 static const char *pgm;
-static int forkdebug = 0;
+static int forkdebug = 1;
 static int numerror = 1;
 static int usecs = 1;
 static int delta = 1;
@@ -102,7 +103,7 @@ error (int geterrno, const char *fmt, ...)
       fputs (buf, stderr);
       fputs ("\n", stderr);
     }
-  ExitProcess (1);
+  exit (1);
 }
 
 DWORD lastid = 0;
@@ -114,7 +115,7 @@ static void
 add_child (DWORD id, HANDLE hproc)
 {
   child_list *c = children.next;
-  children.next = new (child_list);
+  children.next = (child_list *) calloc (1, sizeof (child_list));
   children.next->next = c;
   lastid = children.next->id = id;
   lasth = children.next->hproc = hproc;
@@ -142,7 +143,7 @@ remove_child (DWORD id)
       {
 	child_list *c1 = c->next;
 	c->next = c1->next;
-	delete c1;
+	free (c1);
 	return;
       }
 
@@ -286,12 +287,14 @@ attach_process (pid_t pid)
   load_cygwin ();
   child_pid = (DWORD) cygwin_internal (CW_CYGWIN_PID_TO_WINPID, pid);
   if (!child_pid)
-    error (0, "no such pid - %d", pid);
+    {
+      warn (0, "no such cygwin pid - %d", pid);
+      child_pid = pid;
+    }
 
   if (!DebugActiveProcess (child_pid))
     error (0, "couldn't attach to pid %d<%d> for debugging", pid, child_pid);
 
-  (void) cygwin_internal (CW_STRACE_TOGGLE, pid);
   printf ("Attached to pid %d (windows pid %u)\n", pid, (unsigned) child_pid);
   return;
 }
@@ -307,17 +310,14 @@ create_child (char **argv)
   BOOL ret;
   DWORD flags;
 
-  if (!*argv)
-    error (0, "no program argument specified");
-
+  *argv = cygpath (*argv, NULL);
   memset (&si, 0, sizeof (si));
   si.cb = sizeof (si);
 
-  /* cygwin32_conv_to_win32_path (exec_file, real_path); */
-
-  flags = forkdebug ? 0 : DEBUG_ONLY_THIS_PROCESS;
-  flags |= CREATE_DEFAULT_ERROR_MODE | DEBUG_PROCESS;
-  flags |= (new_window ? CREATE_NEW_CONSOLE | CREATE_NEW_PROCESS_GROUP : 0);
+  flags = CREATE_DEFAULT_ERROR_MODE
+	  | (forkdebug ? DEBUG_PROCESS : DEBUG_ONLY_THIS_PROCESS);
+  if (new_window)
+    flags |= CREATE_NEW_CONSOLE | CREATE_NEW_PROCESS_GROUP;
 
   make_command_line (one_line, argv);
 
@@ -572,7 +572,7 @@ handle_output_debug_string (DWORD id, LPVOID p, unsigned mask, FILE *ofile)
 }
 
 static void
-proc_child (unsigned mask, FILE *ofile)
+proc_child (unsigned mask, FILE *ofile, pid_t pid)
 {
   DEBUG_EVENT ev;
   int processes = 0;
@@ -595,6 +595,12 @@ proc_child (unsigned mask, FILE *ofile)
       if (!debug_event)
 	continue;
 
+      if (pid)
+	{
+	  (void) cygwin_internal (CW_STRACE_TOGGLE, pid);
+	  pid = 0;
+	}
+
       switch (ev.dwDebugEventCode)
 	{
 	case CREATE_PROCESS_DEBUG_EVENT:
@@ -642,16 +648,30 @@ proc_child (unsigned mask, FILE *ofile)
     }
 }
 
+static void
+dotoggle (pid_t pid)
+{
+  load_cygwin ();
+  child_pid = (DWORD) cygwin_internal (CW_CYGWIN_PID_TO_WINPID, pid);
+  if (!child_pid)
+    {
+      warn (0, "no such cygwin pid - %d", pid);
+      child_pid = pid;
+    }
+  if (cygwin_internal (CW_STRACE_TOGGLE, child_pid))
+    error (0, "failed to toggle tracing for process %d<%d>", pid, child_pid);
+
+  return;
+}
+
 static void
 dostrace (unsigned mask, FILE *ofile, pid_t pid, char **argv)
 {
-  if (*argv && pid)
-    error (0, "can't use -p with program argument");
   if (!pid)
     create_child (argv);
   else
     attach_process (pid);
-  proc_child (mask, ofile);
+  proc_child (mask, ofile, pid);
 
   return;
 }
@@ -775,25 +795,30 @@ parse_mask (const char *ms, char **endptr)
 }
 
 static void
-usage ()
+usage (FILE *where = stderr)
 {
-  fprintf (stderr, "\
-Usage: strace [OPTIONS] <command-line>\n\
+  fprintf (where, "\
+Usage: %s [OPTIONS] <command-line>\n\
+Usage: %s [OPTIONS] -p <pid>\n\
   -b, --buffer-size=SIZE       set size of output file buffer\n\
   -d, --no-delta               don't display the delta-t microsecond timestamp\n\
-  -f, --trace-children         also trace forked child processes\n\
-  -h, --help                   display this help info\n\
+  -f, --trace-children         trace child processes (toggle - default true)\n\
+  -h, --help                   output usage information and exit\n\
   -m, --mask=MASK              set message filter mask\n\
-  -o, --output=FILENAME        set output file to FILENAME\n\
-  -p, --pid=n                  attach to executing program with cygwin pid n\n\
   -n, --crack-error-numbers    output descriptive text instead of error\n\
                                numbers for Windows errors\n\
+  -o, --output=FILENAME        set output file to FILENAME\n\
+  -p, --pid=n                  attach to executing program with cygwin pid n\n\
   -S, --flush-period=PERIOD    flush buffered strace output every PERIOD secs\n\
-  -t, --timestamp              use an absolute hh:mm:ss timestamp insted of the\n\
-                               default microsecond timestamp.  Implies -d\n\
-  -v, --version                display version info\n\
+  -t, --timestamp              use an absolute hh:mm:ss timestamp insted of \n\
+                               the default microsecond timestamp.  Implies -d\n\
+  -T, --toggle                 toggle tracing in a process already being\n\
+                               traced. Requires -p <pid>\n\
+  -v, --version                output version information and exit\n\
   -w, --new-window             spawn program under test in a new window\n\
-\n\
+\n", pgm, pgm);
+  if ( where == stdout)
+    fprintf (stdout, "\
     MASK can be any combination of the following mnemonics and/or hex values\n\
     (0x is optional).  Combine masks with '+' or ',' like so:\n\
 \n\
@@ -815,17 +840,14 @@ Usage: strace [OPTIONS] <command-line>\n\
     sigp     0x00800 (_STRACE_SIGP)     Trace signal and process handling.\n\
     minimal  0x01000 (_STRACE_MINIMAL)  Very minimal strace output.\n\
     exitdump 0x04000 (_STRACE_EXITDUMP) Dump strace cache on exit.\n\
-    system   0x08000 (_STRACE_SYSTEM)   Serious error which goes to console and log.\n\
+    system   0x08000 (_STRACE_SYSTEM)   Serious error; goes to console and log.\n\
     nomutex  0x10000 (_STRACE_NOMUTEX)  Don't use mutex for synchronization.\n\
     malloc   0x20000 (_STRACE_MALLOC)   Trace malloc calls.\n\
     thread   0x40000 (_STRACE_THREAD)   Thread-locking calls.\n\
 ");
-}
-
-static void
-version ()
-{
-  fputs (version_string + 4, stderr);
+  if (where == stderr)
+    fprintf (stderr, "Try '%s --help' for more information.\n", pgm);
+  exit (where == stderr ? 1 : 0 );
 }
 
 struct option longopts[] = {
@@ -838,6 +860,7 @@ struct option longopts[] = {
   {"no-delta", no_argument, NULL, 'd'},
   {"pid", required_argument, NULL, 'p'},
   {"timestamp", no_argument, NULL, 't'},
+  {"toggle", no_argument, NULL, 'T'},
   {"trace-children", no_argument, NULL, 'f'},
   {"translate-error-numbers", no_argument, NULL, 'n'},
   {"usecs", no_argument, NULL, 'u'},
@@ -845,15 +868,39 @@ struct option longopts[] = {
   {NULL, 0, NULL, 0}
 };
 
-static const char *const opts = "b:dhfm:no:p:S:tuvw";
+static const char *const opts = "b:dhfm:no:p:S:tTuvw";
+
+static void
+print_version ()
+{
+  const char *v = strchr (version, ':');
+  int len;
+  if (!v)
+    {
+      v = "?";
+      len = 1;
+    }
+  else
+    {
+      v += 2;
+      len = strchr (v, ' ') - v;
+    }
+  printf ("\
+%s (cygwin) %.*s\n\
+System Trace\n\
+Copyright 2000, 2001, 2002, 2003 Red Hat, Inc.\n\
+Compiled on %s\n\
+", pgm, len, v, __DATE__);
+}
 
 int
 main (int argc, char **argv)
 {
   unsigned mask = 0;
   FILE *ofile = NULL;
-  pid_t attach_pid = 0;
+  pid_t pid = 0;
   int opt;
+  int toggle = 0;
 
   if (!(pgm = strrchr (*argv, '\\')) && !(pgm = strrchr (*argv, '/')))
     pgm = *argv;
@@ -874,8 +921,7 @@ main (int argc, char **argv)
 	break;
       case 'h':
 	// Print help and exit
-	usage ();
-	return 1;
+	usage (stdout);
 	break;
       case 'm':
 	{
@@ -887,41 +933,55 @@ main (int argc, char **argv)
 	      error (0, "syntax error in mask expression \"%s\" near \
 character #%d.\n", optarg, (int) (endptr - optarg), endptr);
 	    }
-	break;
+	  break;
 	}
       case 'n':
 	numerror ^= 1;
 	break;
       case 'o':
-	if ((ofile = fopen (optarg, "w")) == NULL)
+	if ((ofile = fopen (cygpath (optarg, NULL), "w")) == NULL)
 	  error (1, "can't open %s", optarg);
 #ifdef F_SETFD
 	(void) fcntl (fileno (ofile), F_SETFD, 0);
 #endif
 	break;
       case 'p':
-	attach_pid = strtol (optarg, NULL, 10);
+	pid = strtoul (optarg, NULL, 10);
 	break;
       case 'S':
-	flush_period = strtol (optarg, NULL, 10);
+	flush_period = strtoul (optarg, NULL, 10);
 	break;
       case 't':
 	hhmmss ^= 1;
 	break;
+      case 'T':
+	toggle ^= 1;
+	break;
       case 'u':
 	// FIXME: currently unimplemented
 	usecs ^= 1;
 	break;
       case 'v':
 	// Print version info and exit
-	version ();
-	return 1;
-	break;
+	print_version ();
+	return 0;
       case 'w':
 	new_window ^= 1;
 	break;
+      case '?':
+	fprintf (stderr, "Try '%s --help' for more information.\n", pgm);
+	exit (1);
       }
 
+  if (pid && argv[optind])
+    error (0, "cannot provide both a command line and a process id");
+
+  if (!pid && !argv[optind])
+    error (0, "must provide either a command line or a process id");
+
+  if (toggle && !pid)
+    error (0, "must provide a process id to toggle tracing");
+
   if (!mask)
     mask = 1;
 
@@ -931,7 +991,10 @@ character #%d.\n", optarg, (int) (endptr - optarg), endptr);
   if (!ofile)
     ofile = stdout;
 
-  dostrace (mask, ofile, attach_pid, argv + optind);
+  if (toggle)
+    dotoggle (pid);
+  else
+    dostrace (mask, ofile, pid, argv + optind);
 }
 
 #undef CloseHandle
